Package: PGRdup
Title: Discover Probable Duplicates in Plant Genetic Resources
        Collections
Version: 0.3.0
Authors@R: c(
    person(given = "J.", family = "Aravind",
           email = "j.aravind@icar.org.in",
           role = c("aut", "cre"),
		       comment = c(ORCID = "0000-0002-4791-442X")),
    person(given = "J.", family = "Radhamani",
           email = "jalli.radhamani@icar.gov.in",
		       role = "aut"),
    person(family = c("Kalyani Srinivasan"),
           email = "kalyani.srinivasan@icar.gov.in",
		       role = "aut"),
    person(given = "B.", family = c("Ananda Subhash"),
           email = "anandasubhash@gmail.com",
		       role = "aut"),
    person(given = c("Rishi", "Kumar"), family = "Tyagi",
           email = "rishi.tyagi@icar.gov.in",
		       role = "aut"),
    person("ICAR-NBGPR",
           role = "cph",
		       comment = c(ROR = "https://ror.org/00scbd467",
		                   url = "https://nbpgr.org.in/")),
    person(given = "Maurice", family = "Aubrey",
           email = "maurice@hevanet.com",
           comment = "Double Metaphone",
           role = "ctb"),
    person(given = "Kevin", family = "Atkinson",
           email = "kevina@users.sourceforge.net",
           comment = "Double Metaphone",
           role = "ctb"),
    person(given = "Lawrence", family = "Philips",
           comment = "Double Metaphone",
           role = "ctb"))
Description: Provides functions to aid the identification of probable/possible
    duplicates in Plant Genetic Resources (PGR) collections using
    'passport databases' comprising of information records of each constituent
    sample. These include methods for cleaning the data, creation of a
    searchable Key Word in Context (KWIC) index of keywords associated with
    sample records and the identification of nearly identical records with
    similar information by fuzzy, phonetic and semantic matching of keywords.
Depends: R (>= 4.1.0)
Imports: data.table (>= 1.9.3), igraph, stringdist (>= 0.9.4), stringi,
        ggplot2, grid, gridExtra, methods, utils, stats
Suggests: diagram, wordcloud, microbenchmark, XML, httr, RCurl, knitr,
        rmarkdown, pander
Copyright: 2014-2026, ICAR-NBPGR
License: GPL-2 | GPL-3
Encoding: UTF-8
LazyData: true
VignetteBuilder: knitr
RoxygenNote: 7.3.3
URL: https://cran.r-project.org/package=PGRdup,
        https://github.com/aravind-j/PGRdup,
        https://doi.org/10.5281/zenodo.841963,
        https://aravind-j.github.io/PGRdup/,
        https://www.rdocumentation.org/packages/PGRdup
BugReports: https://github.com/aravind-j/PGRdup/issues
NeedsCompilation: yes
Packaged: 2026-04-21 10:51:52 UTC; Aravind-DGC
Author: J. Aravind [aut, cre] (ORCID: <https://orcid.org/0000-0002-4791-442X>),
  J. Radhamani [aut],
  Kalyani Srinivasan [aut],
  B. Ananda Subhash [aut],
  Rishi Kumar Tyagi [aut],
  ICAR-NBGPR [cph] (ROR: <https://ror.org/00scbd467>, url:
    https://nbpgr.org.in/),
  Maurice Aubrey [ctb] (Double Metaphone),
  Kevin Atkinson [ctb] (Double Metaphone),
  Lawrence Philips [ctb] (Double Metaphone)
Maintainer: J. Aravind <j.aravind@icar.org.in>
Repository: CRAN
Date/Publication: 2026-04-21 22:22:29 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-26 09:23:13 UTC; unix
Archs: PGRdup.so.dSYM

Package: PGRdup
Title: Discover Probable Duplicates in Plant Genetic Resources
        Collections
Version: 0.3.0
Authors@R: c(
    person(given = "J.", family = "Aravind",
           email = "j.aravind@icar.org.in",
           role = c("aut", "cre"),
		       comment = c(ORCID = "0000-0002-4791-442X")),
    person(given = "J.", family = "Radhamani",
           email = "jalli.radhamani@icar.gov.in",
		       role = "aut"),
    person(family = c("Kalyani Srinivasan"),
           email = "kalyani.srinivasan@icar.gov.in",
		       role = "aut"),
    person(given = "B.", family = c("Ananda Subhash"),
           email = "anandasubhash@gmail.com",
		       role = "aut"),
    person(given = c("Rishi", "Kumar"), family = "Tyagi",
           email = "rishi.tyagi@icar.gov.in",
		       role = "aut"),
    person("ICAR-NBGPR",
           role = "cph",
		       comment = c(ROR = "https://ror.org/00scbd467",
		                   url = "https://nbpgr.org.in/")),
    person(given = "Maurice", family = "Aubrey",
           email = "maurice@hevanet.com",
           comment = "Double Metaphone",
           role = "ctb"),
    person(given = "Kevin", family = "Atkinson",
           email = "kevina@users.sourceforge.net",
           comment = "Double Metaphone",
           role = "ctb"),
    person(given = "Lawrence", family = "Philips",
           comment = "Double Metaphone",
           role = "ctb"))
Description: Provides functions to aid the identification of probable/possible
    duplicates in Plant Genetic Resources (PGR) collections using
    'passport databases' comprising of information records of each constituent
    sample. These include methods for cleaning the data, creation of a
    searchable Key Word in Context (KWIC) index of keywords associated with
    sample records and the identification of nearly identical records with
    similar information by fuzzy, phonetic and semantic matching of keywords.
Depends: R (>= 4.1.0)
Imports: data.table (>= 1.9.3), igraph, stringdist (>= 0.9.4), stringi,
        ggplot2, grid, gridExtra, methods, utils, stats
Suggests: diagram, wordcloud, microbenchmark, XML, httr, RCurl, knitr,
        rmarkdown, pander
Copyright: 2014-2026, ICAR-NBPGR
License: GPL-2 | GPL-3
Encoding: UTF-8
LazyData: true
VignetteBuilder: knitr
RoxygenNote: 7.3.3
URL: https://cran.r-project.org/package=PGRdup,
        https://github.com/aravind-j/PGRdup,
        https://doi.org/10.5281/zenodo.841963,
        https://aravind-j.github.io/PGRdup/,
        https://www.rdocumentation.org/packages/PGRdup
BugReports: https://github.com/aravind-j/PGRdup/issues
NeedsCompilation: yes
Packaged: 2026-04-21 10:51:52 UTC; Aravind-DGC
Author: J. Aravind [aut, cre] (ORCID: <https://orcid.org/0000-0002-4791-442X>),
  J. Radhamani [aut],
  Kalyani Srinivasan [aut],
  B. Ananda Subhash [aut],
  Rishi Kumar Tyagi [aut],
  ICAR-NBGPR [cph] (ROR: <https://ror.org/00scbd467>, url:
    https://nbpgr.org.in/),
  Maurice Aubrey [ctb] (Double Metaphone),
  Kevin Atkinson [ctb] (Double Metaphone),
  Lawrence Philips [ctb] (Double Metaphone)
Maintainer: J. Aravind <j.aravind@icar.org.in>
Repository: CRAN
Date/Publication: 2026-04-21 22:22:29 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-28 08:18:04 UTC; unix
Archs: PGRdup.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(print,KWIC)
S3method(print,ProbDup)
export(AddProbDup)
export(DataClean)
export(DisProbDup)
export(DoubleMetaphone)
export(KWCounts)
export(KWIC)
export(MergeKW)
export(MergePrefix)
export(MergeProbDup)
export(MergeSuffix)
export(ParseProbDup)
export(ProbDup)
export(ReconstructProbDup)
export(ReviewProbDup)
export(SplitProbDup)
export(ValidatePrimKey)
export(ViewProbDup)
export(print.KWIC)
export(print.ProbDup)
export(read.genesys)
import(data.table)
import(igraph)
import(stringdist)
importFrom(ggplot2,aes)
importFrom(ggplot2,coord_flip)
importFrom(ggplot2,coord_polar)
importFrom(ggplot2,element_blank)
importFrom(ggplot2,element_line)
importFrom(ggplot2,element_rect)
importFrom(ggplot2,element_text)
importFrom(ggplot2,geom_bar)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_rect)
importFrom(ggplot2,geom_text)
importFrom(ggplot2,geom_tile)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,ggplotGrob)
importFrom(ggplot2,guide_legend)
importFrom(ggplot2,guides)
importFrom(ggplot2,labs)
importFrom(ggplot2,margin)
importFrom(ggplot2,scale_y_reverse)
importFrom(ggplot2,theme)
importFrom(ggplot2,xlab)
importFrom(ggplot2,xlim)
importFrom(ggplot2,ylab)
importFrom(grid,textGrob)
importFrom(grid,unit)
importFrom(gridExtra,arrangeGrob)
importFrom(methods,is)
importFrom(stats,embed)
importFrom(stringi,stri_count)
importFrom(stringi,stri_count_fixed)
importFrom(stringi,stri_count_regex)
importFrom(stringi,stri_split_fixed)
importFrom(utils,capture.output)
importFrom(utils,head)
importFrom(utils,setTxtProgressBar)
importFrom(utils,stack)
importFrom(utils,txtProgressBar)
importFrom(utils,unzip)
useDynLib(PGRdup, .registration = TRUE)

<img src="https://raw.githubusercontent.com/aravind-j/PGRdup/master/inst/extdata/PGRdup.png" align="center" alt="logo" width="750" height = "85" style = "border: solid white 10px;">

###### Version : [0.3.0](https://aravind-j.github.io/PGRdup/articles/Introduction.html#version-history); Copyright (C) 2014-2026: [ICAR-NBPGR](https://nbpgr.org.in/); License: [GPL-2\|GPL-3](https://www.r-project.org/Licenses/)

##### *Aravind, J.<sup>1</sup>, Radhamani, J.<sup>1</sup>, Kalyani Srinivasan<sup>1</sup>, Ananda Subhash, B.<sup>2</sup>, and Tyagi, R. K.<sup>1</sup>*

1.  ICAR-National Bureau of Plant Genetic Resources, New Delhi, India
2.  Centre for Development of Advanced Computing, Thiruvananthapuram,
    Kerala, India

------------------------------------------------------------------------

[![minimal R
version](https://img.shields.io/badge/R%3E%3D-3.0.2-6666ff.svg?logo=R)](https://cran.r-project.org/)
[![License: GPL
v3](https://img.shields.io/badge/License-GPL%20v3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version-last-release/PGRdup)](https://cran.r-project.org/package=PGRdup)
[![Dependencies](https://tinyverse.netlify.app/status/PGRdup)](https://cran.r-project.org/package=PGRdup)
[![rstudio mirror
downloads](https://cranlogs.r-pkg.org/badges/grand-total/PGRdup?color=green)](https://CRAN.R-project.org/package=PGRdup)
[![develVersion](https://img.shields.io/badge/devel%20version-0.3.0-orange.svg)](https://github.com/aravind-j/PGRdup)
[![Github Code
Size](https://img.shields.io/github/languages/code-size/aravind-j/PGRdup.svg)](https://github.com/aravind-j/PGRdup)
[![R-CMD-check](https://github.com/aravind-j/PGRdup/workflows/R-CMD-check/badge.svg)](https://github.com/aravind-j/PGRdup/actions)
[![Project Status:
Inactive](https://www.repostatus.org/badges/latest/inactive.svg)](https://www.repostatus.org/#inactive)
[![lifecycle](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
[![Last-changedate](https://img.shields.io/badge/last%20change-2026--04--21-yellowgreen.svg)](https://github.com/aravind-j/PGRdup)
[![Rdoc](https://api.rdocumentation.org/badges/version/PGRdup)](https://www.rdocumentation.org/packages/PGRdup/)
[![Zenodo
DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.841963.svg)](https://doi.org/10.5281/zenodo.841963)
[![Website -
pkgdown](https://img.shields.io/website-up-down-green-red/https/aravind-j.github.io/PGRdup.svg)](https://aravind-j.github.io/PGRdup/)
<!-- [![.](https://pro-pulsar-193905.appspot.com/UA-112827210-2/welcome-page)](https://github.com/aravind-j/google-analytics-beacon) -->
[![GoatCounter](https://pgrdup-gh.goatcounter.com/count?p=/test)](https://pgrdup-gh.goatcounter.com/)
<!-- [![packageversion](https://img.shields.io/badge/Package%20version-0.2.3.3-orange.svg)](https://github.com/aravind-j/PGRdup) -->
<!-- [![GitHub Download Count](https://github-basic-badges.herokuapp.com/downloads/aravind-j/PGRdup/total.svg)] -->

------------------------------------------------------------------------

## Introduction <img src="https://raw.githubusercontent.com/aravind-j/PGRdup/master/inst/extdata/PGRdup_v2.png" align="right" alt="logo" width="173" height = "200" style = "padding: 10px; border: none; float: right;">

The `R` package `PGRdup` was developed as a tool to aid genebank
managers in the identification of probable duplicate accessions from
plant genetic resources (PGR) passport databases.

This package primarily implements a workflow designed to fetch groups or
sets of germplasm accessions with similar passport data particularly in
fields associated with accession names within or across PGR passport
databases.

The functions in this package are primarily built using the following R
packages:

- [`data.table`](https://cran.r-project.org/package=data.table)
- [`igraph`](https://cran.r-project.org/package=igraph)
- [`stringdist`](https://cran.r-project.org/package=stringdist)
- [`stringi`](https://cran.r-project.org/package=stringi)

## Installation

The package can be installed from CRAN as follows:

``` r
# Install from CRAN
install.packages('PGRdup', dependencies=TRUE)

The development version can be installed from github as follows:
# Install development version from Github
devtools::install_github("aravind-j/PGRdup")
```

## Workflow

The series of steps involve in the workflow along with the associated
functions are are illustrated below:

#### Step 1

**Function(s) :**

- `DataClean`
- `MergeKW`
- `MergePrefix`
- `MergeSuffix`

Use these functions for the appropriate data standardisation of the
relevant fields in the passport databases to harmonize punctuation,
leading zeros, prefixes, suffixes etc. associated with accession names.

#### Step 2

**Function(s) :**

- `KWIC`

Use this function to extract the information in the relevant fields as
keywords or text strings in the form of a searchable Keyword in Context
(KWIC) index.

#### Step 3

**Function(s) :**

- `ProbDup`

Execute fuzzy, phonetic and semantic matching of keywords to identify
probable duplicate sets either within a single KWIC index or between two
indexes using this function. For fuzzy matching the levenshtein edit
distance is used, while for phonetic matching, the double metaphone
algorithm is used. For semantic matching, synonym sets or ‘synsets’ of
accession names can be supplied as an input and the text strings in such
sets will be treated as being identical for matching. Various options to
tweak the matching strategies used are also available in this function.

#### Step 4

**Function(s) :**

- `DisProbDup`
- `ReviewProbDup`
- `ReconstructProbDup`

Inspect, revise and improve the retrieved sets using these functions. If
considerable intersections exist between the initially identified sets,
then `DisProbDup` may be used to get the disjoint sets. The identified
sets may be subjected to clerical review after transforming them into an
appropriate spreadsheet format which contains the raw data from the
original database(s) using `ReviewProbDup` and subsequently converted
back using `ReconstructProbDup`.

#### Adjuncts

**Function(s) :**

- `ValidatePrimKey`
- `DoubleMetaphone`
- `ParseProbDup`
- `AddProbDup`
- `SplitProbDup`
- `MergeProbDup`
- `ViewProbDup`
- `KWCounts`
- `read.genesys`

Use these helper functions if needed. `ValidatePrimKey` can be used to
check whether a column chosen in a data.frame as the primary primary
key/ID confirms to the constraints of absence of duplicates and NULL
values.

`DoubleMetaphone` is an implementation of the Double Metaphone phonetic
algorithm in `R` and is used for phonetic matching.

`ParseProbDup` and `AddProbDup` work with objects of class `ProbDup`.
The former can be used to parse the probable duplicate sets in a
`ProbDup` object to a `data.frame` while the latter can be used to add
these sets data fields to the passport databases. `SplitProbDup` can be
used to split an object of class `ProbDup` according to set counts.
`MergeProbDup` can be used to merge together two objects of class
`ProbDup`. `ViewProbDup` can be used to plot the summary visualizations
of probable duplicate sets retrieved in an object of class `ProbDup`.

`KWCounts` can be used to compute keyword counts from PGR passport
database fields(columns), which can give a rough indication of the
completeness of the data.

`read.genesys` can be used to import PGR data in a Darwin Core -
germplasm zip archive downloaded from genesys database into the R
environment.

## Tips

- Use [`fread`](https://rdrr.io/cran/data.table/man/fread.html) to
  rapidly read large files instead of `read.csv` or `read.table` in
  `base`.
- In case the PGR passport data is in any DBMS, use the appropriate
  [`R`-database interface
  packages](https://www.burns-stat.com/r-database-interfaces/) to get
  the required table as a `data.frame` in `R`.

## Note

- The `ProbDup` function can be memory hungry with large passport
  databases. In such cases, ensure that the system has sufficient memory
  for smooth functioning (See `?ProbDup`).

## Detailed tutorial

For a detailed tutorial (vignette) on how to used this package type:

``` r
browseVignettes(package = 'PGRdup')
```

The vignette for the latest version is also available
[online](https://aravind-j.github.io/PGRdup/articles/Introduction.html).

## What’s new

To know whats new in this version type:

``` r
news(package='PGRdup')
```

## Links

[CRAN page](https://cran.r-project.org/package=PGRdup)

[Github page](https://github.com/aravind-j/PGRdup)

[Documentation website](https://aravind-j.github.io/PGRdup/)

[Zenodo DOI](https://doi.org/10.5281/zenodo.841963)

## CRAN checks

[![Linux](https://img.shields.io/badge/Linux-FCC624?style=for-the-badge&logo=linux&logoColor=black)](https://cran.r-project.org/web/checks/check_results_germinationmetrics.html)

|  |  |
|:---|----|
| r-devel-linux-x86_64-debian-clang | [![CRAN check - r-devel-linux-x86_64-debian-clang](https://badges.cranchecks.info/flavor/r-devel-linux-x86_64-debian-clang/PGRdup.svg)](https://cran.r-project.org/web/checks/check_results_PGRdup.html) |
| r-devel-linux-x86_64-debian-gcc | [![CRAN check - r-devel-linux-x86_64-debian-gcc](https://badges.cranchecks.info/flavor/r-devel-linux-x86_64-debian-gcc/PGRdup.svg)](https://cran.r-project.org/web/checks/check_results_PGRdup.html) |
| r-devel-linux-x86_64-fedora-clang | [![CRAN check - r-devel-linux-x86_64-fedora-clang](https://badges.cranchecks.info/flavor/r-devel-linux-x86_64-fedora-clang/PGRdup.svg)](https://cran.r-project.org/web/checks/check_results_PGRdup.html) |
| r-devel-linux-x86_64-fedora-gcc | [![CRAN check - r-devel-linux-x86_64-fedora-gcc](https://badges.cranchecks.info/flavor/r-devel-linux-x86_64-fedora-gcc/PGRdup.svg)](https://cran.r-project.org/web/checks/check_results_PGRdup.html) |
| r-patched-linux-x86_64 | [![CRAN check - r-patched-linux-x86_64](https://badges.cranchecks.info/flavor/r-patched-linux-x86_64/PGRdup.svg)](https://cran.r-project.org/web/checks/check_results_PGRdup.html) |
| r-release-linux-x86_64 | [![CRAN check - r-release-linux-x86_64](https://badges.cranchecks.info/flavor/r-release-linux-x86_64/PGRdup.svg)](https://cran.r-project.org/web/checks/check_results_PGRdup.html) |

[![Windows](https://img.shields.io/badge/Windows-0078D6?style=for-the-badge&logo=windows&logoColor=white)](https://cran.r-project.org/web/checks/check_results_germinationmetrics.html)

|  |  |
|:---|----|
| r-devel-windows-x86_64 | [![CRAN check - r-devel-windows-x86_64](https://badges.cranchecks.info/flavor/r-devel-windows-x86_64/PGRdup.svg)](https://cran.r-project.org/web/checks/check_results_PGRdup.html) |
| r-release-windows-x86_64 | [![CRAN check - r-release-windows-x86_64](https://badges.cranchecks.info/flavor/r-release-windows-x86_64/PGRdup.svg)](https://cran.r-project.org/web/checks/check_results_PGRdup.html) |
| r-oldrel-windows-x86_64 | [![CRAN check - r-oldrel-windows-x86_64](https://badges.cranchecks.info/flavor/r-oldrel-windows-x86_64/PGRdup.svg)](https://cran.r-project.org/web/checks/check_results_PGRdup.html) |

[![MacOS](https://img.shields.io/badge/mac%20os-000000?style=for-the-badge&logo=apple&logoColor=white)](https://cran.r-project.org/web/checks/check_results_germinationmetrics.html)

|  |  |
|:---|----|
| r-release-macos-x86_64 | [![CRAN check - r-release-macos-x86_64](https://badges.cranchecks.info/flavor/r-release-macos-x86_64/PGRdup.svg)](https://cran.r-project.org/web/checks/check_results_PGRdup.html) |
| r-oldrel-macos-x86_64 | [![CRAN check - r-oldrel-macos-x86_64](https://badges.cranchecks.info/flavor/r-oldrel-macos-x86_64/PGRdup.svg)](https://cran.r-project.org/web/checks/check_results_PGRdup.html) |

## Citing `PGRdup`

To cite the methods in the package use:

``` r
citation("PGRdup")
```

    To cite the R package 'PGRdup' in publications use:

      Aravind, J., Radhamani, J., Kalyani Srinivasan, Ananda Subhash, B., and Tyagi, R.
      K.  (2026).  PGRdup: Discover Probable Duplicates in Plant Genetic Resources
      Collections. R package version 0.3.0,
      https://github.com/aravind-j/PGRdup,https://cran.r-project.org/package=PGRdup.

    A BibTeX entry for LaTeX users is

      @Manual{,
        title = {PGRdup: Discover Probable Duplicates in Plant Genetic Resources Collections},
        author = {J. Aravind and J. Radhamani and {Kalyani Srinivasan} and B. {Ananda Subhash} and Rishi Kumar Tyagi},
        note = {R package version 0.3.0 https://github.com/aravind-j/PGRdup, https://cran.r-project.org/package=PGRdup},
        year = {2026},
      }

    This free and open-source software implements academic research by the authors and
    co-workers. If you use it, please support the project by citing the package.

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/



#' Retrieve probable duplicate set information from PGR passport database for
#' review
#'
#' \code{ReviewProbDup} retrieves information associated with the probable
#' duplicate sets from the original PGR passport database(s) from which they
#' were identified in order to facilitate manual clerical review.
#'
#' This function helps to retrieve PGR passport information associated with
#' fuzzy, phonetic or semantic probable duplicate sets in an object of class
#' \code{ProbDup} from the original databases(s) from which they were
#' identified. The original information of accessions comprising a set, which
#' have not been subjected to data standardization can be compared under manual
#' clerical review for the validation of the set.
#'
#' By default only the fields(columns) which were used initially for creation of
#' the KWIC indexes using the \code{\link[PGRdup]{KWIC}} function are retrieved.
#' Additional fields(columns) if necessary can be specified using the
#' \code{extra.db1} and \code{extra.db2} arguments.
#'
#' The output data frame can be subjected to clerical review either after
#' exporting into an external spreadsheet using
#' \code{\link[utils:write.table]{write.csv}} function or by using the
#' \code{\link[utils]{edit}} function.
#'
#' The column \code{DEL} can be used to indicate whether a record has to be
#' deleted from a set or not. \code{Y} indicates "Yes", and the default \code{N}
#' indicates "No".
#'
#' The column \code{SPLIT} similarly can be used to indicate whether a record in
#' a set has to be branched into a new set. A set of identical integers in this
#' column other than the default \code{0} can be used to indicate that they are
#' to be removed and assembled into a new set.
#'
#' @param pdup An object of class \code{ProbDup}.
#' @param db1 A data frame of the PGR passport database.
#' @param db2 A data frame of the PGR passport database. Required when
#'   \code{pdup} was created using more than one KWIC Index.
#' @param extra.db1 A character vector of extra \code{db1} column names to be
#'   retrieved.
#' @param extra.db2 A character vector of extra \code{db2} column names to be
#'   retrieved.
#' @param max.count The maximum count of probable duplicate sets whose
#'   information is to be retrieved.
#' @param insert.blanks logical. If \code{TRUE}, inserts a row of \code{NAs}
#'   after each set.
#' @return A data frame of the long/narrow form of the probable duplicate sets
#'   data along with associated fields from the original database(s). The core
#'   columns in the resulting data frame are as follows: \tabular{ll}{
#'   \code{SET_NO} \tab The set number. \cr \code{TYPE} \tab The type of
#'   probable duplicate set. 'F' for fuzzy, 'P' for phonetic and 'S' for
#'   semantic matching sets. \cr \code{K[*]} \tab The KWIC index or database of
#'   origin of the record. The \code{method} is specified within the square
#'   brackets in the column name.  \cr \code{PRIM_ID} \tab The primary ID of the
#'   accession record from which the set could be identified. \cr \code{IDKW}
#'   \tab The 'matching' keywords along with the IDs. \cr \code{DEL} \tab Column
#'   to indicate whether record has to be deleted or not. \cr \code{SPLIT} \tab
#'   Column to indicate whether record has to be branched and assembled into new
#'   set. \cr \code{COUNT} \tab The number of elements in a set. \cr } For the
#'   retrieved columns(fields) the prefix \code{K*} indicates the KWIC index of
#'   origin.
#' @note When any primary ID/key records in the fuzzy, phonetic or semantic
#'   duplicate sets are found to be missing from the original databases
#'   \code{db1} and \code{db2}, then they are ignored and only the matching
#'   records are considered for retrieving the information with a warning.
#'
#'   This may be due to data standardization of the primary ID/key field using
#'   the function \code{\link[PGRdup]{DataClean}} before creation of the KWIC
#'   index and subsequent identification of probable duplicate sets. In such a
#'   case, it is recommended to use an identical data standardization operation
#'   on the databases \code{db1} and \code{db2} before running this function.
#'
#'   With \code{R} <= v3.0.2, due to copying of named objects by \code{list()},
#'   \code{Invalid .internal.selfref detected and fixed...} warning can appear,
#'   which may be safely ignored.
#' @seealso \code{\link[PGRdup]{DataClean}}, \code{\link[PGRdup]{KWIC}},
#'   \code{\link[PGRdup]{ProbDup}}
#' @examples
#' 
#' \dontshow{
#' threads_dt <- data.table::getDTthreads()
#' threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
#' data.table::setDTthreads(2)
#' 
#' data.table::setDTthreads(2)
#' Sys.setenv(`OMP_THREAD_LIMIT` = 2)
#' }
#' 
#' \dontrun{
#'
#' # Load PGR passport database
#' GN <- GN1000
#'
#' # Specify as a vector the database fields to be used
#' GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")
#'
#' # Clean the data
#' GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
#' y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
#' c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
#' c("Mota", "Company"))
#' y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
#' y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
#'         "Bunch", "Peanut")
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))
#'
#' # Generate KWIC index
#' GNKWIC <- KWIC(GN, GNfields)
#'
#' # Specify the exceptions as a vector
#' exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE",
#'          "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT",
#'          "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE",
#'          "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R",
#'          "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE",
#'          "U", "VALENCIA", "VIRGINIA", "WHITE")
#'
#' # Specify the synsets as a list
#' syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))
#'
#' # Fetch probable duplicate sets
#' GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, fuzzy = TRUE,
#'                  phonetic = TRUE, encoding = "primary",
#'                  semantic = TRUE, syn = syn)
#'
#' # Get disjoint probable duplicate sets of each kind
#' disGNdup <- DisProbDup(GNdup, combine = NULL)
#'
#' # Get the data frame for reviewing the duplicate sets identified
#' RevGNdup <- ReviewProbDup(pdup = disGNdup, db1 = GN1000,
#'                           extra.db1 = c("SourceCountry", "TransferYear"),
#'                           max.count = 30, insert.blanks = TRUE)
#' # Examine and review the duplicate sets using edit function
#' RevGNdup <- edit(RevGNdup)
#'
#' # OR examine and review the duplicate sets after exporting them as a csv file
#' write.csv(file="Duplicate sets for review.csv", x=RevGNdup)
#'
#' }
#' 
#' \dontshow{
#' data.table::setDTthreads(threads_dt)
#' Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
#' }
#' 
#' @import data.table
#' @importFrom methods is
#' @export
ReviewProbDup <- function (pdup, db1, db2 = NULL,
                           extra.db1 = NULL, extra.db2 = NULL,
                           max.count = 30, insert.blanks  = TRUE) {
  if (!is(pdup, "ProbDup")) {
    stop('"pdup" is not of class ProbDup')
  }
  method <- attributes(pdup)$method
  fields <- attributes(pdup)$fields
  if (method == "c" | method == "b") {
    if (is.null(db2)) {
      stop(paste("argument 'db2' is missing, with no default.",
                 "\nSecond database is to be specified as method ",
                 method, " was used to generate 'pdup'", sep = ""))
    }
  }
  if (is.element(FALSE, fields[[1]] %in% colnames(db1))) {
    # Check if fields are present in db1 and stop if not
    stop("One or more kwic1 fields are missing in 'db1'")
  }
  if (!is.null(extra.db1) && !is.vector(extra.db1, mode = "character")) {
    stop('"extra.db1" is not a character vector')
  }
  if (is.element(FALSE, extra.db1 %in% colnames(db1))) {
    # Check if extra fields are present in db1 and stop if not
    warning(paste("One or more extra kwic1 fields are missing in 'db1'.",
                  "Only the default kwic1 fields will be retrieved"))
  }
  fields[[1]] <- union(fields[[1]], extra.db1)
  #setDT(db1)
  db1 <- as.data.table(db1)
  if (!identical(setdiff(colnames(db1), fields[[1]]), character(0))){
    db1[, setdiff(colnames(db1), fields[[1]]) := NULL]
  }
  db1[, K1_PRIM_ID := get(fields[[1]][1])]
  setcolorder(db1, neworder = union("K1_PRIM_ID",
                                    setdiff(colnames(db1), "K1_PRIM_ID")))
  db1[, K1_PRIM_ID := as.character(K1_PRIM_ID)]
  setkey(db1, "K1_PRIM_ID")
  if (method == "c" | method == "b") {
    if (is.element(FALSE, fields[[2]] %in% colnames(db2)) == TRUE) {
      # Check if fields are present in db2 and stop if not
      stop("One or more kwic2 fields are missing in 'db2'")
    }
    if (!is.null(extra.db2) && !is.vector(extra.db2, mode = "character")) {
      stop("'extra.db2' is not a character vector")
    }
    if (is.element(FALSE, extra.db2 %in% colnames(db2)) == TRUE) {
      # Check if extra fields are present in db2 and stop if not
      warning(paste("One or more extra kwic2 fields are missing in 'db2'.",
                    "Only the default kwic2 fields will be retrieved"))
    }
    fields[[2]] <- union(fields[[2]], extra.db2)
    #setDT(db2)
    db2 <- as.data.table(db2)
    if (!identical(setdiff(colnames(db2), fields[[1]]), character(0))) {
      db2[, setdiff(colnames(db2), fields[[2]]) := NULL]
    }
    db2[, K2_PRIM_ID := get(fields[[2]][1])]
    setcolorder(db2, neworder = union("K2_PRIM_ID",
                                      setdiff(colnames(db2), "K2_PRIM_ID")))
    db2[, K2_PRIM_ID := as.character(K2_PRIM_ID)]
    setkey(db2, "K2_PRIM_ID")
  }
  N <- length(seq_along(pdup))
  types <- c("F", "P", "S", "D")
  types2 <- c("Fuzzy", "Phonetic", "Semantic", "Disjoint")
  for (i in 1:N) {
    if (!is.null(pdup[[i]])) {
      #setDT(pdup[[i]])
      pdup[[i]] <- as.data.table(pdup[[i]])
      pdup[[i]] <- subset(pdup[[i]], COUNT <= max.count)
      pdup[[i]] <- unique(pdup[[i]])
      # Reset SET_NO to take into account deleted sets with coutn > max.count
      pdup[[i]][, Seq := 1:.N]
      # Cast ID and IDKW by SET_NO
      pdup[[i]] <- pdup[[i]][, .(unlist(strsplit(IDKW, ", ", TRUE))),
                             by = list(SET_NO, TYPE, COUNT)][,
                               .(IDKW = toString(V1)), .(SET_NO, TYPE, COUNT,
                                                         PRIM_ID = gsub(":.*", "", V1))]
      # Create K1_PRIM_ID column for merging with db1
      pdup[[i]][, K1_PRIM_ID := gsub("\\[K2\\].*", "", PRIM_ID, perl = TRUE)]
      pdup[[i]][, K1_PRIM_ID := gsub("\\[K1\\]", "", K1_PRIM_ID, perl = TRUE)]
      pdup[[i]][, K1_PRIM_ID := as.character(K1_PRIM_ID)]
      setkey(pdup[[i]], K1_PRIM_ID)
      # Check for IDs in pdup missing in db1
      if (length(setdiff(pdup[[i]]$K1_PRIM_ID[pdup[[i]]$K1_PRIM_ID != ""],
                         db1$K1_PRIM_ID)) != 0) {
        warning(paste("Encountered primary ID records in",
                      types2[i], "probable duplicate sets missing from 'db1'.",
                      "\nOnly matching records are merged"))
      }
      # Merge with db1
      pdup[[i]] <- merge(pdup[[i]], db1, all.x = TRUE)
      pdup[[i]][, K1_PRIM_ID := NULL]
      setkey(pdup[[i]], "SET_NO")
      # Set prefix to merged fields
      setnames(pdup[[i]], old = fields[[1]],
               new =  paste("K1", fields[[1]], sep = "_"))
      if (method == "c" | method == "b") {
        # Create K2_PRIM_ID column for merging with db2
        pdup[[i]][, K2_PRIM_ID := gsub("\\[K1\\].*", "", PRIM_ID, perl = TRUE)]
        pdup[[i]][, K2_PRIM_ID := gsub("\\[K2\\]", "", K2_PRIM_ID, perl = TRUE)]
        pdup[[i]][, K2_PRIM_ID := as.character(K2_PRIM_ID)]
        setkey(pdup[[i]], K2_PRIM_ID)
        # Check for IDs in pdup missing in db1
        if (length(setdiff(pdup[[i]]$K2_PRIM_ID[pdup[[i]]$K2_PRIM_ID != ""],
                           db2$K2_PRIM_ID)) != 0) {
          warning(paste("Encountered primary ID records in",
                        types2[i],
                        "probable duplicate sets missing from 'db2'.",
                        "\nOnly matching records are merged"))
        }
        # Merge with db1
        pdup[[i]] <- merge(pdup[[i]], db2, all.x = TRUE)
        pdup[[i]][, K2_PRIM_ID := NULL]
        setkey(pdup[[i]], "SET_NO")
        setnames(pdup[[i]], old = fields[[2]],
                 new =  paste("K2", fields[[2]], sep = "_"))
      }
    }
  }
  # rbind pdup list
  pdup <- rbindlist(pdup)
  # Split K* from PRIM_ID column
  pdup[, PRIM_ID := gsub("\\]", "\\]_", PRIM_ID, perl = TRUE)]
  pdup[, K := gsub("_.*", "", PRIM_ID, perl = TRUE)]
  pdup[, PRIM_ID := gsub("\\[K1\\]_", "", PRIM_ID, perl = TRUE)]
  pdup[, PRIM_ID := gsub("\\[K2\\]_", "", PRIM_ID, perl = TRUE)]
  # Add review columns
  pdup[, DEL := "N" ]
  pdup[, SPLIT := 0 ]
  # Reset column order
  nameslist <- union(c("SET_NO", "TYPE", "K", "PRIM_ID",
                       "IDKW", "DEL", "SPLIT", "COUNT"),
                    colnames(pdup))
  setcolorder(x = pdup, neworder = nameslist)
  # Add prefix to additional fields
  if (!is.null(extra.db1) && !is.element(FALSE,
                                        paste("K1", extra.db1, sep = "_") %in% colnames(pdup))) {
    setnames(pdup, old = paste("K1", extra.db1, sep = "_"),
             new = paste("K1X", extra.db1, sep = "_"))
  }
  if (method == "c" | method == "b") {
    if (!is.null(extra.db2) && !is.element(FALSE,
                                          paste("K2", extra.db2,
                                                sep = "_") %in% colnames(pdup))) {
      setnames(pdup, old = paste("K2", extra.db2, sep = "_"),
               new = paste("K2X", extra.db2, sep = "_"))
    }
  }
  # Insert blanks
  setkey(pdup, SET_NO)
  setkey(pdup, NULL)
  if (insert.blanks  == TRUE) {
    pdup[, TEMP := as.factor(SET_NO)]
    pdup[, TEMP := interaction(pdup$TEMP, as.factor(pdup$TYPE), drop = TRUE)]
    setattr(pdup$TEMP, "levels", seq(from = 1, to = length(levels(pdup$TEMP))))
    pdup[, TEMP := as.numeric(TEMP)]
    pdup <- setDT(pdup)[pdup[, c(.I, NA), TEMP]$V1][!.N]
    pdup[, TEMP := NULL]
    pdup[is.na(TYPE), TYPE := ""]
  }
  setnames(pdup, old = "K", new = paste("K[", method, "]", sep = ""))
  setDF(pdup)
  return(pdup)
}

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/



#' Identify probable duplicates of accessions
#'
#' \code{ProbDup} identifies probable duplicates of germplasm accessions in KWIC
#' indexes created from PGR passport databases using fuzzy, phonetic and
#' semantic matching strategies.
#'
#' This function performs fuzzy, phonetic and semantic matching of keywords in
#' KWIC indexes of PGR passport databases (created using
#' \code{\link[PGRdup]{KWIC}} function) to identify probable duplicates of
#' germplasm accessions. The function can execute matching according to either
#' of the following three methods as specified by the \code{method} argument.
#'
#' \describe{ \item{Method \code{a}:}{Perform string matching of keywords in a
#' single KWIC index to identify probable duplicates of accessions in a single
#' PGR passport database.} \item{Method \code{b}:}{Perform string matching of
#' keywords in the first KWIC index (query) with that of the keywords in the
#' second index (source) to identify probable duplicates of accessions of the
#' first PGR passport database among the accessions in the second database.}
#' \item{Method \code{c}:}{Perform string matching of keywords in two different
#' KWIC indexes jointly to identify probable duplicates of accessions from among
#' two PGR passport databases.}}
#'
#' \strong{Fuzzy matching} or approximate string matching of keywords is carried
#' out by computing the generalized levenshtein (edit) distance between them.
#' This distance measure  counts the number of deletions, insertions and
#' substitutions necessary to turn one string to the another. A distance of up
#' to \code{max.dist} are considered for a match.
#'
#' Exact matching will be enforced when the argument \code{force.exact} is
#' \code{TRUE}. It can be used to avoid fuzzy matching when the number of
#' alphabet characters in keywords is lesser than a critical value
#' (\code{max.alpha}). Similarly, the value of \code{max.digit} can also be set
#' according to the requirements. The default value of \code{Inf} avoids fuzzy
#' matching and enforces exact matching for all keywords having any numerical
#' characters. If \code{max.digit} and \code{max.alpha} are both set to
#' \code{Inf}, exact matching will be enforced for all the keywords.
#'
#' When exact matching is enforced, for keywords having both alphabet and
#' numeric characters and with the number of alphabet characters greater than
#' \code{max.digit}, matching will be carried out separately for alphabet and
#' numeric characters present.
#'
#' \strong{Phonetic matching} of keywords is carried out using the Double
#' Metaphone phonetic algorithm (\code{\link[PGRdup]{DoubleMetaphone}}) to
#' identify keywords that have the similar pronunciation. Either the
#' \code{primary} or \code{alternate} encodings can be used by specifying the
#' \code{encoding} argument. The argument \code{phon.min.alpha} sets the limits
#' for the number of alphabet characters to be present in a string for executing
#' phonetic matching. Similarly \code{min.enc} sets the limits for the number of
#' characters to be present in the encoding of a keyword for phonetic matching.
#'
#' \strong{Semantic matching} matches keywords based on a list of accession name
#' synonyms supplied as list with character vectors of synonym sets (synsets) to
#' the \code{syn} argument. Synonyms in this context refers to interchangeable
#' identifiers or names by which an accession is recognized. Multiple keywords
#' specified as members of the same synset in \code{syn} are merged together. To
#' facilitate accurate identification of synonyms from the KWIC index, identical
#' data standardization operations using the \code{\link[PGRdup]{MergeKW}} and
#' \code{\link[PGRdup]{DataClean}} functions for both the original database
#' fields and the synset list are recommended.
#'
#' The probable duplicate sets identified initially here may be intersecting
#' with other sets. To get the disjoint sets after the union of all the
#' intersecting sets use the \code{\link[PGRdup]{DisProbDup}} function.
#'
#' The function \code{\link[PGRdup]{AddProbDup}} can be used to add the
#' information associated with the identified sets in an object of class
#' \code{ProbDup} as fields(columns) to the original PGR passport database.
#'
#' All of the string matching operations here are executed through the
#' \code{\link[stringdist]{stringdist-package}} functions.
#'
#' @param kwic1 An object of class \code{KWIC}.
#' @param kwic2 An object of class \code{KWIC}. Required for \code{method}
#'   \code{"b"} and \code{"c"} only (see \strong{Details}).
#' @param method The method to be followed for identification of probable
#'   duplicates. Either \code{"a"}, \code{"b"} or \code{"c"}. (see
#'   \strong{Details}).
#' @param chunksize A value indicating the size of KWIC index keyword block to
#'   be used for searching for matches at a time in case of large number of
#'   keywords(see \strong{Note}).
#' @param useBytes logical. If \code{TRUE}, performs byte-wise comparison
#'   instead of character-wise comparison (see \strong{Note}).
#' @param excep A vector of the keywords in KWIC not to be used for probable
#'   duplicate search (see \strong{Details}).
#' @param fuzzy logical. If \code{TRUE} identifies probable duplicates based on
#'   fuzzy matching.
#' @param max.dist The maximum levenshtein distance between keyword strings
#'   allowed for a match. Default is 3 (see \strong{Details}).
#' @param force.exact logical. If \code{TRUE}, enforces exact matching instead
#'   of fuzzy matching for keyword strings which match the criteria specified in
#'   arguments \code{max.alpha} and \code{max.digit} (see \strong{Details}).
#' @param max.alpha Maximum number of alphabet characters present in a keyword
#'   string up to which exact matching is enforced rather than fuzzy matching.
#'   Default is 4 (see \strong{Details}).
#' @param max.digit Maximum number of numeric characters present in a keyword
#'   string up to which exact matching is enforced rather than fuzzy matching.
#'   Default is Inf (see \strong{Details}).
#' @param phonetic logical. If \code{TRUE} identifies probable duplicates based
#'   on phonetic matching.
#' @param encoding Double metaphone encoding for phonetic matching. The default
#'   is \code{"primary"} (see \strong{Details}).
#' @param phon.min.alpha Minimum number of alphabet characters to be present in
#'   a keyword string for phonetic matching (see \strong{Details}).
#' @param min.enc Minimum number of characters to be be present in double
#'   metaphone encoding of a keyword string for phonetic matching (see
#'   \strong{Details}).
#' @param semantic logical. If \code{TRUE} identifies probable duplicates based
#'   on semantic matching.
#' @param syn A list with character vectors of synsets (see \strong{Details}).
#' @note As the number of keywords in the KWIC indexes increases, the memory
#'   consumption by the function also increases. For string matching, this
#'   function relies upon creation of a \eqn{n}*\eqn{m} matrix of all possible
#'   keyword pairs for comparison, where \eqn{n} and \eqn{m} are the number of
#'   keywords in the query and source indexes respectively. This can lead to
#'   \code{cannot allocate vector of size} errors in case very large KWIC
#'   indexes where the comparison matrix is too large to reside in memory. In
#'   such a case, try to adjust the \code{chunksize} argument to get the
#'   appropriate size of the KWIC index keyword block to be used for searching
#'   for matches at a time. However a smaller chunksize may lead to longer
#'   computation time due to the memory-time trade-off.
#'
#'   The progress of matching is displayed in the console as number of blocks
#'   completed out of total (e.g. 6 / 30), the percentage of achievement (e.g.
#'   30\%) and a text-based progress bar.
#'
#'   In case of multi-byte characters in keywords, the matching speed is further
#'   dependent upon the \code{useBytes} argument as described in
#'   \strong{Encoding issues} for the \code{\link[stringdist]{stringdist}}
#'   function, which is made use of here for string matching.
#' @references van der Loo, M. P. J. 2014. "The Stringdist Package for
#'   Approximate String Matching." \emph{R Journal} 6 (1):111-22.
#'   \url{https://journal.r-project.org/articles/RJ-2014-011/index.html}.
#' @encoding UTF-8
#' @examples
#' 
#' \dontshow{
#' threads_dt <- data.table::getDTthreads()
#' threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
#' data.table::setDTthreads(2)
#' 
#' data.table::setDTthreads(2)
#' Sys.setenv(`OMP_THREAD_LIMIT` = 2)
#' }
#' 
#' \dontrun{
#'
#' # Method "a"
#' #===========
#'
#' # Load PGR passport database
#' GN <- GN1000
#'
#' # Specify as a vector the database fields to be used
#' GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")
#'
#' # Clean the data
#' GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
#' y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
#' c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
#' c("Mota", "Company"))
#' y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
#' y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
#'         "Bunch", "Peanut")
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))
#'
#' # Generate KWIC index
#' GNKWIC <- KWIC(GN, GNfields)
#'
#' # Specify the exceptions as a vector
#' exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE",
#'          "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT",
#'          "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE",
#'          "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R",
#'          "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE",
#'          "U", "VALENCIA", "VIRGINIA", "WHITE")
#'
#' # Specify the synsets as a list
#' syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))
#'
#' # Fetch probable duplicate sets
#' GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, fuzzy = TRUE,
#'                  phonetic = TRUE, encoding = "primary",
#'                  semantic = TRUE, syn = syn)
#' GNdup
#'
#' # Method "b and c"
#' #=================
#'
#' # Load PGR passport databases
#' GN1 <- GN1000[!grepl("^ICG", GN1000$DonorID), ]
#' GN1$DonorID <- NULL
#' GN2 <- GN1000[grepl("^ICG", GN1000$DonorID), ]
#' GN2 <- GN2[!grepl("S", GN2$DonorID), ]
#' GN2$NationalID <- NULL
#'
#' # Specify as a vector the database fields to be used
#' GN1fields <- c("NationalID", "CollNo", "OtherID1", "OtherID2")
#' GN2fields <- c("DonorID", "CollNo", "OtherID1", "OtherID2")
#'
#' # Clean the data
#' GN1[GN1fields] <- lapply(GN1[GN1fields], function(x) DataClean(x))
#' GN2[GN2fields] <- lapply(GN2[GN2fields], function(x) DataClean(x))
#' y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
#' c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
#' c("Mota", "Company"))
#' y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
#' y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
#'         "Bunch", "Peanut")
#' GN1[GN1fields] <- lapply(GN1[GN1fields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
#' GN1[GN1fields] <- lapply(GN1[GN1fields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
#' GN1[GN1fields] <- lapply(GN1[GN1fields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))
#' GN2[GN2fields] <- lapply(GN2[GN2fields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
#' GN2[GN2fields] <- lapply(GN2[GN2fields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
#' GN2[GN2fields] <- lapply(GN2[GN2fields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))
#'
#' # Remove duplicated DonorID records in GN2
#' GN2 <- GN2[!duplicated(GN2$DonorID), ]
#'
#' # Generate KWIC index
#' GN1KWIC <- KWIC(GN1, GN1fields)
#' GN2KWIC <- KWIC(GN2, GN2fields)
#'
#' # Specify the exceptions as a vector
#' exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE",
#'          "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT",
#'          "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE",
#'          "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R",
#'          "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE",
#'          "U", "VALENCIA", "VIRGINIA", "WHITE")
#'
#' # Specify the synsets as a list
#' syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))
#'
#' # Fetch probable duplicate sets
#' GNdupb <- ProbDup(kwic1 = GN1KWIC, kwic2 = GN2KWIC, method = "b",
#'                   excep = exep, fuzzy = TRUE, phonetic = TRUE,
#'                   encoding = "primary", semantic = TRUE, syn = syn)
#' GNdupb
#'
#' GNdupc <- ProbDup(kwic1 = GN1KWIC, kwic2 = GN2KWIC, method = "c",
#'                   excep = exep, fuzzy = TRUE, phonetic = TRUE,
#'                   encoding = "primary", semantic = TRUE, syn = syn)
#' GNdupc
#'
#' }
#' 
#' \dontshow{
#' data.table::setDTthreads(threads_dt)
#' Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
#' }
#' 
#' @seealso \code{\link[PGRdup]{KWIC}}, \code{\link[PGRdup]{DoubleMetaphone}}
#'   \code{\link[stringdist:stringdist]{stringdistmatrix}},
#'   \code{\link[utils]{adist}}, \code{\link[PGRdup]{print.ProbDup}}
#' @return A list of class \code{ProbDup} containing the following data frames
#'   of probable duplicate sets identified along with the corresponding keywords
#'   and set counts: \enumerate{ \item \code{FuzzyDuplicates} \item
#'   \code{PhoneticDuplicates} \item \code{SemanticDuplicates} } Each data frame
#'   has the following columns: \tabular{ll}{ \code{SET_NO} \tab The set number.
#'   \cr \code{TYPE} \tab The type of probable duplicate set. 'F' for fuzzy, 'P'
#'   for phonetic and 'S' for semantic matching sets. \cr \code{ID} \tab The
#'   primary IDs of records of accessions comprising a set. \cr \code{ID:KW}
#'   \tab The 'matching' keywords along with the IDs. \cr \code{COUNT} \tab The
#'   number of elements in a set. \cr }
#'
#'   The prefix \code{[K*]} indicates the KWIC index of origin of the KEYWORD or
#'   PRIM_ID.
#' @import igraph
#' @import stringdist
#' @import data.table
#' @importFrom stringi stri_count_fixed
#' @importFrom stringi stri_count_regex
#' @importFrom methods is
#' @importFrom utils stack
#' @importFrom utils capture.output
#' @importFrom utils setTxtProgressBar
#' @importFrom utils txtProgressBar
#' @importFrom stats embed
#' @export ProbDup
#' @export print.ProbDup
ProbDup <- function (kwic1, kwic2 = NULL, method = c("a", "b", "c"),
                     excep = NULL, chunksize = 1000, useBytes = TRUE,
                     fuzzy = TRUE, max.dist = 3, force.exact = TRUE,
                     max.alpha = 4, max.digit = Inf,
                     phonetic = TRUE, encoding = c("primary", "alternate"),
                     phon.min.alpha = 5, min.enc = 3,
                     semantic = FALSE, syn = NULL) {
  # Preliminary Checks ----
  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  # Check method argument
  method <- match.arg(method)
  fields <- list(k1 = NULL, k2 = NULL)
  # Check excep argument
  if (!is.null(excep) && is.vector(excep, mode = "character") == FALSE) {
    stop('"excep" is not a character vector')
  }
  if (!is.null(excep)) {
    excep <- toupper(excep)
  } else {
    excep <- ""
  }
  # Check if kwic 1 is present
  if (is.null(kwic1)) {
    stop('"kwic1" is missing')
  }
  # Check if kwic1 is of class KWIC
  if (is(kwic1, "KWIC")) {
    fields[[1]] <- kwic1[[3]]
    kwic1 <- kwic1[[1]][!(kwic1[[1]]$KEYWORD %in% excep), c(1, 3)]
  } else {
    stop('"kwic1" is not of class KWIC')
  }
  # Assign and check query and source kwic indexes according to methods
  if (method == "a") {
    kwic2 <- NULL
  }
  if (method == "b" | method == "c") {
    if (is.null(kwic2)) {
      stop('"kwic2" is missing')
    }
    if (is(kwic2, "KWIC")) {
      fields[[2]] <- kwic2[[3]]
      kwic2 <- kwic2[[1]][!(kwic2[[1]]$KEYWORD %in% excep), c(1, 3)]
    } else {
      stop('"kwic2" is not of class KWIC')
    }
  }
  
  # Check encoding argument
  if (phonetic) {
    encoding <- match.arg(encoding)
  }
  # Check syn argument
  if (semantic) {
    if (is.null(syn)) {
      stop('"syn" is missing')
    }
    if (!is.list(syn)) {
      stop('"syn" is not a list')
    }
    if (is.element(FALSE, as.logical(lapply(syn,
                                            function(x) is.character(x))))) {
      warning('list "syn" had non character vectors; coerced to character')
      syn <-  as.logical(lapply(syn, function(x) as.character(x)))
    }
    nsyn1 <- length(syn)
    syn <- fix.syn(syn)
    nsyn2 <- length(syn)
    if (nsyn1 != nsyn2) {
      warning('synsets in list "syn" with common strings were merged')
    }
    rm(nsyn1, nsyn2)
  }
  # Create the output list
  out <- list(FuzzyDuplicates = NULL, PhoneticDuplicates = NULL,
              SemanticDuplicates = NULL)
  attr(out, "method") <- method
  attr(out, "fields") <- fields
  rm(fields)
  # Prepare the indexes ----
  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  kwicQ <- as.data.table(kwic1)
  rm(kwic1)
  kwicQ[, PRIM_ID := gsub("([[:space:]])\\1*", "", PRIM_ID)]
  kwicQ[, PRIM_ID := paste("[K1]", PRIM_ID, sep = "")]
  kwicQ[, IDKW := paste(PRIM_ID, KEYWORD, sep = ":")]
  
  if (method == "a" | method == "b") {
    kq_prim_id_agg <- collapse_dedup(kwicQ, by_col = "KEYWORD", val_col = "PRIM_ID")
    kq_idkw_agg    <- collapse_dedup(kwicQ, by_col = "KEYWORD", val_col = "IDKW")
    
    kwicQ <- kq_prim_id_agg[kq_idkw_agg, on = "KEYWORD"]
    
    # Id the chunks in kwic1
    M <- nrow(kwicQ)
    if (M > chunksize) {
      chunksize <- chunksize
    } else {
      chunksize <- M
    }
    kwicQ[, iter := rep(1:M, each = chunksize, length.out = M)]
    setcolorder(kwicQ, neworder = c("PRIM_ID", "KEYWORD", "IDKW", "iter"))
    setDF(kwicQ)
  }
  
  if (method == "a") {
    kwicS <- kwicQ
    #N <- M
  } else {
    kwicS <- as.data.table(kwic2)
    rm(kwic2)
    kwicS[, PRIM_ID := gsub("([[:space:]])\\1*", "", PRIM_ID)]
    kwicS[, PRIM_ID := paste("[K2]", PRIM_ID, sep = "")]
    kwicS[, IDKW := paste(PRIM_ID, KEYWORD, sep = ":")]
    if (method == "b") {
      ks_prim_id_agg <- collapse_dedup(kwicS, by_col = "KEYWORD", 
                                       val_col = "PRIM_ID")
      ks_idkw_agg    <- collapse_dedup(kwicS, by_col = "KEYWORD", 
                                       val_col = "IDKW")
      
      kwicS <- ks_prim_id_agg[ks_idkw_agg, on = "KEYWORD"]
    }
    
    N <- nrow(kwicS)
    
    #kwicS[, iter:= rep(1:M, each = chunksize, length.out = N)]
    setcolorder(kwicS, neworder = c("PRIM_ID", "KEYWORD", "IDKW"))
    setDF(kwicS)
    
    if (method == "c") {
      kwicQ <- as.data.table(rbind(kwicQ, kwicS))
      
      kq_prim_id_agg <- collapse_dedup(kwicQ, by_col = "KEYWORD", 
                                       val_col = "PRIM_ID")
      kq_idkw_agg    <- collapse_dedup(kwicQ, by_col = "KEYWORD", 
                                       val_col = "IDKW")
      
      kwicQ <- kq_prim_id_agg[kq_idkw_agg, on = "KEYWORD"]
  
      M <- nrow(kwicQ)
      
      if (M > chunksize) {
        chunksize <- chunksize
      } else {
        chunksize <- M
      }
      
      kwicQ[, iter := rep(1:M, each = chunksize, length.out = M)]
      setcolorder(kwicQ, neworder = c("PRIM_ID", "KEYWORD", "IDKW", "iter"))
      kwicS <- kwicQ
      setDF(kwicQ)
      setDF(kwicS)
    }
  }
  
  # Fuzzy Matching ----
  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  if (fuzzy) {
    # Coerce to integer
    if (max.alpha != Inf) {
      max.alpha <- as.integer(max.alpha)
    }
    if (max.digit != Inf) {
      max.digit <- as.integer(max.digit)
    }
    out[[1]] <- FuzzyDup(kwic1 = kwicQ, kwic2 = kwicS,
                         method = method, useBytes = useBytes,
                         max.dist = max.dist, force.exact = force.exact,
                         max.alpha = max.alpha, max.digit = max.digit)
  }
  # Phonetic Matching ----
  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  if (phonetic) {
    # Coerce to integer
    if (phon.min.alpha != Inf) {
      phon.min.alpha <- as.integer(phon.min.alpha)
    }
    if (min.enc != Inf) {
      min.enc <- as.integer(min.enc)
    }
    out[[2]] <- PhoneticDup(kwic1 = kwicQ, kwic2 = kwicS,
                            method = method, useBytes = useBytes,
                            encoding = encoding,
                            phon.min.alpha = phon.min.alpha, min.enc = min.enc)
  }
  # Semantic Matching ----
  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  if (semantic) {
    out[[3]] <- SemanticDup(kwic1 = kwicQ, kwic2 = kwicS,
                            method = method, syn = syn,
                            useBytes = useBytes)
  }
  class(out) <- "ProbDup"
  # Convert to null if no sets of a type are retrieved
  for (i in 1:3) {
    if (is.null(out[[i]])) {
      out[i] <- list(NULL)
    } else {
      if (dim(out[[i]])[1] == 0) {
        out[i] <- list(NULL)
      }
    }
  }
  return(out)
}

FuzzyDup <- function(kwic1, kwic2, max.dist, useBytes,
                     force.exact, max.alpha, max.digit, method) {
  kwic1 <- as.data.table(kwic1)
  kwic2 <- as.data.table(kwic2)
  M <- nrow(kwic1)
  N <- nrow(kwic2)
  
  # Identify strings to be exactly matched
  ind_exactQ <- logical(length = M)
  if (force.exact) {
    cond1 <- grepl("[[:digit:]]", kwic1$KEYWORD,
                   ignore.case = TRUE) == TRUE & 
      stri_count_regex(kwic1$KEYWORD,
                       "[[:digit:]]") <= max.digit
    cond2 <- grepl("[[:alpha:]]", kwic1$KEYWORD,
                   ignore.case = TRUE) == TRUE & 
      stri_count_regex(kwic1$KEYWORD,
                       "[[:alpha:]]") <= max.alpha
    ind_exactQ <- cond1 | cond2
    rm(cond1, cond2)
  }
  ind_exactS <- logical(length = N)
  if (force.exact) {
    cond1 <- grepl("[[:digit:]]", kwic2$KEYWORD,
                   ignore.case = TRUE) == TRUE & 
      stri_count_regex(kwic2$KEYWORD,
                       "[[:digit:]]") <= max.digit
    cond2 <- grepl("[[:alpha:]]", kwic2$KEYWORD,
                   ignore.case = TRUE) == TRUE & 
      stri_count_regex(kwic2$KEYWORD,
                       "[[:alpha:]]") <= max.alpha
    ind_exactS <- cond1 | cond2
    rm(cond1, cond2)
  }
  
  # Identify strings with mixed characters(alpha + digit)
  ind_mixedQ <- logical(length = M)
  cond1 <- grepl("[[:digit:]]", kwic1$KEYWORD,
                 ignore.case = TRUE) == TRUE & 
    grepl("[[:alpha:]]",
          kwic1$KEYWORD,
          ignore.case = TRUE) == TRUE
  cond2 <- stri_count_regex(kwic1$KEYWORD, "[[:alpha:]]") > max.alpha
  ind_mixedQ <- cond1 & cond2
  rm(cond1, cond2)
  ind_mixedS <- logical(length = N)
  cond1 <- grepl("[[:digit:]]", kwic2$KEYWORD,
                 ignore.case = TRUE) == TRUE & 
    grepl("[[:alpha:]]",
          kwic2$KEYWORD,
          ignore.case = TRUE) == TRUE
  cond2 <- stri_count_regex(kwic2$KEYWORD, "[[:alpha:]]") > max.alpha
  ind_mixedS <- cond1 & cond2
  rm(cond1, cond2)
  
  # Prepare mixed string vectors
  mixed_alphaQ <- ifelse(ind_mixedQ == TRUE, kwic1$KEYWORD, "")
  mixed_digitQ <- mixed_alphaQ
  mixed_alphaQ <- gsub(pattern = "[[:digit:]]", replacement = "",
                       x = mixed_alphaQ)
  mixed_digitQ <- gsub(pattern = "[[:alpha:]]", replacement = "",
                       x = mixed_digitQ)
  mixed_alphaS <- ifelse(ind_mixedS == TRUE, kwic2$KEYWORD, "")
  mixed_digitS <- mixed_alphaS
  mixed_alphaS <- gsub(pattern = "[[:digit:]]", replacement = "",
                       x = mixed_alphaS)
  mixed_digitS <- gsub(pattern = "[[:alpha:]]", replacement = "",
                       x = mixed_digitS)
  ind_mixed_digit_exact <- stri_count_regex(mixed_digitQ,
                                            "[[:digit:]]") <= max.digit
  
  # Pre-build lookup for exact
  if (method == "b") {
    exact_lookup <- data.table(
      keyword = kwic2$KEYWORD[ind_exactS],
      idkw    = kwic2$IDKW[ind_exactS]
    )
    setkey(exact_lookup, keyword)
    
    # Collapse multiple IDKWs per keyword into one string
    exact_lookup <- 
      exact_lookup[, .(FuzzydupIDKW = paste(idkw, collapse = ", ")),
                   by = keyword]
    setkey(exact_lookup, keyword)
  }
  
  # Create progress bar
  invisible(capture.output(pb <- txtProgressBar(min = 0, max = max(kwic1$iter),
                                                style = 3)))
  message("Fuzzy matching")
  for (i in unique(kwic1$iter)) {
    in_iter <- (kwic1$iter == i)
    
    # Single join for exact
    if (method == "b" && sum(in_iter & ind_exactQ) > 0 && 
        nrow(exact_lookup) > 0) {
      
      rows_exact <- which(in_iter & ind_exactQ)
      
      kwic1[rows_exact,
            FuzzydupIDKW := exact_lookup[.(kwic1$KEYWORD[rows_exact]),
                                         FuzzydupIDKW,
                                         nomatch = NA]]
    }
    
    # Compute the chunk stringdist results (fuzzy and mixed)
    keywords_Q_fuzzy <- kwic1$KEYWORD[in_iter & !ind_exactQ]
    keywords_S_fuzzy <- kwic2$KEYWORD[!ind_exactS]
    idkw_S_fuzzy     <- kwic2$IDKW[!ind_exactS]
    
    fuzzy_results <- 
      vapply(keywords_Q_fuzzy, 
             function(kw) {
               dists <- stringdist(kw, keywords_S_fuzzy, method = "lv", 
                                   useBytes = useBytes)
               paste(idkw_S_fuzzy[dists <= max.dist], collapse = ", ")
             }, character(1))
    
    mixed <- NULL
    # if (sum(in_iter & ind_mixedQ) > 0 & sum(ind_mixedS) > 0) {
    if (sum(in_iter & ind_mixedQ) > 0) {
      alpha_S  <- mixed_alphaS[ind_mixedS]
      digit_S  <- mixed_digitS[ind_mixedS]
      idkw_S   <- kwic2$IDKW[ind_mixedS]
      
      rows_mixed        <- which(in_iter & ind_mixedQ)
      rows_digit_exact  <- which(in_iter & ind_mixedQ &  ind_mixed_digit_exact)
      rows_digit_fuzzy  <- which(in_iter & ind_mixedQ & !ind_mixed_digit_exact)
      
      mixed_results <- vapply(rows_mixed, function(row_idx) {
        
        d_alpha <- stringdist(mixed_alphaQ[row_idx], alpha_S,
                              method = "lv", useBytes = useBytes)
        d_alpha[d_alpha > max.dist] <- Inf
        
        if (row_idx %in% rows_digit_exact) {
          d_digit <- stringdist(mixed_digitQ[row_idx], digit_S,
                                method = "lv", useBytes = useBytes)
          d_digit[d_digit != 0] <- Inf
          
        } else if (row_idx %in% rows_digit_fuzzy) {
          d_digit <- stringdist(mixed_digitQ[row_idx], digit_S,
                                method = "lv", useBytes = useBytes)
          d_digit[d_digit > max.dist] <- Inf
          
        } else {
          d_digit <- rep(0, length(digit_S))
        }
        
        d_combined <- d_alpha + d_digit
        
        paste(idkw_S[d_combined <= max.dist], collapse = ", ")
        
      }, character(1))
      
    }
    
    # Fetch duplicates to a new column
    if (sum(in_iter & !ind_exactQ) > 0 & sum(!ind_exactS) > 0) {
      kwic1[in_iter & !ind_exactQ, FuzzydupIDKW := fuzzy_results]
    }
    if (sum(in_iter & ind_mixedQ) > 0 & sum(ind_mixedS) > 0) {
      kwic1[in_iter & ind_mixedQ, FuzzydupIDKW2 := mixed_results]
    }
    
    # update progress bar
    setTxtProgressBar(pb, i)
    cat("\rBlock", i, "/", max(kwic1$iter), "|")
  }
  close(pb)
  
  ind_exactQ2 <- stri_count_fixed(kwic1$PRIM_ID, ",") != 0 & ind_exactQ
  
  rm(ind_exactQ, ind_mixedQ, ind_exactS, ind_mixedS,
     pb, mixed_digitS, mixed_alphaS, mixed_alphaQ, mixed_digitQ,
     ind_mixed_digit_exact, in_iter)
  
  cols <- setdiff(colnames(kwic1), c("KEYWORD", "PRIM_ID", "iter"))
  
  for (j in cols) {
    set(kwic1, which(is.na(kwic1[[j]])), j, "")
  }
  
  if (method == "a" | method == "c") {
    kwic1[ind_exactQ2, FuzzydupIDKW := IDKW]
  }
  
  rm(ind_exactQ2)
  
  if ("FuzzydupIDKW2" %in% colnames(kwic1)) {
    kwic1[, FuzzydupIDKW := toString(unique(c(strsplit(FuzzydupIDKW,
                                                       split = ", ")[[1]],
                                              strsplit(FuzzydupIDKW2,
                                                       split = ", ")[[1]]))),
          by = IDKW]
    kwic1[, FuzzydupIDKW2 := NULL]
  }
  
  kwic1[, FuzzydupID := gsub(":\\S+\\b", "", FuzzydupIDKW)]
  kwic1 <- dupsets(kwic1, "F", method = method)
  
  return(kwic1)
}

PhoneticDup <- function(kwic1, kwic2, encoding, useBytes,
                        phon.min.alpha, min.enc, method) {
  kwic1 <- as.data.table(kwic1)
  kwic2 <- as.data.table(kwic2)
  M <- nrow(kwic1)
  N <- nrow(kwic2)
  
  # Fetch keyword phonetic encodings
  if (encoding == "primary") {
    DMQ <- DoubleMetaphone(kwic1$KEYWORD)[[1]]
    DMS <- DoubleMetaphone(kwic2$KEYWORD)[[1]]
  }
  if (encoding == "alternate") {
    DMQ <- DoubleMetaphone(kwic1$KEYWORD)[[2]]
    DMS <- DoubleMetaphone(kwic2$KEYWORD)[[2]]
  }
  
  # Prepare keyword phonetic encodings
  DMQ <- ifelse(stri_count_regex(kwic1$KEYWORD, "[[:alpha:]]") < phon.min.alpha,
                "", DMQ)
  DMS <- ifelse(stri_count_regex(kwic2$KEYWORD, "[[:alpha:]]") < phon.min.alpha,
                "", DMS)
  DMQ <- ifelse(nchar(DMQ) < min.enc, "", DMQ)
  DMS <- ifelse(nchar(DMS) < min.enc, "", DMS)
  
  # Identify strings with phonetic encodings
  ind_phon <- logical(length = M)
  ind_phon <- DMQ != ""
  ind_phonS <- logical(length = N)
  ind_phonS <- DMS != ""
  
  # Identify strings with phonetic encodings having digits
  ind_phon_digitQ <- grepl("[[:digit:]]", kwic1$KEYWORD,
                           ignore.case = TRUE) == TRUE & ind_phon
  ind_phon_digitS <- grepl("[[:digit:]]", kwic2$KEYWORD,
                           ignore.case = TRUE) == TRUE & ind_phonS
  
  # Prepared strings with phonetic encodings having digits
  phon_digitQ <- ifelse(ind_phon_digitQ == TRUE, kwic1$KEYWORD, "")
  phon_digitQ <- gsub(pattern = "[[:alpha:]]", replacement = "",
                      x = phon_digitQ)
  phon_digitS <- ifelse(ind_phon_digitS == TRUE, kwic2$KEYWORD, "")
  phon_digitS <- gsub(pattern = "[[:alpha:]]", replacement = "",
                      x = phon_digitS)
  
  # Phonetic-only candidates (no digits)
  phon_lookup <- data.table(
    enc    = DMS[ind_phonS & !ind_phon_digitS],
    idkw   = kwic2$IDKW[ind_phonS & !ind_phon_digitS]
  )
  
  # Phonetic + digit candidates
  phon_digit_lookup <- data.table(
    enc    = DMS[ind_phonS & ind_phon_digitS],
    digit  = phon_digitS[ind_phon_digitS],
    idkw   = kwic2$IDKW[ind_phonS & ind_phon_digitS]
  )
  setkey(phon_lookup, enc)
  setkey(phon_digit_lookup, enc, digit)
  
  # Create progress bar
  invisible(capture.output(pb <- txtProgressBar(min = 0, max = max(kwic1$iter),
                                                style = 3)))
  message("Phonetic matching")
  for (i in unique(kwic1$iter)) {
    in_iter <- (kwic1$iter == i)
    
    if (sum(in_iter & ind_phon) == 0) next # skip to next loop iteration
    
    rows_phon       <- which(in_iter & ind_phon)
    rows_phon_digit <- which(in_iter & ind_phon_digitQ)
    rows_phon_only  <- setdiff(rows_phon, rows_phon_digit)
    
    results <- vapply(rows_phon, function(row_idx) {
      
      enc_q <- DMQ[row_idx]
      
      if (row_idx %in% rows_phon_digit) {
        # Must match BOTH phonetic encoding AND digit part exactly
        dig_q <- phon_digitQ[row_idx]
        hits  <- phon_digit_lookup[.(enc_q, dig_q), idkw, nomatch = NULL]
        
      } else {
        # Match phonetic encoding only
        hits <- phon_lookup[.(enc_q), idkw, nomatch = NULL]
      }
      
      paste(hits, collapse = ", ")
      
    }, character(1))
    
    # Fetch duplicates to a new column
    kwic1[rows_phon, PhoneticdupIDKW := results]
    
    rm(in_iter)
    
    # update progress bar
    setTxtProgressBar(pb, i)
    cat("\rBlock", i, "/", max(kwic1$iter), "|")
  }
  close(pb)
  
  rm(ind_phon, pb, ind_phonS, ind_phon_digitQ, phon_digitQ, phon_digitS)
  
  for (j in c("IDKW", "PhoneticdupIDKW")) {
    set(kwic1, which(is.na(kwic1[[j]])), j, "")
  }
  
  kwic1[, PhoneticdupID := gsub(":\\S+\\b", "", PhoneticdupIDKW)]
  kwic1 <- dupsets(kwic1, "P", method = method)
  
  return(kwic1)
}

SemanticDup <- function(kwic1, kwic2, syn, useBytes, method) {
  kwic1 <- as.data.table(kwic1)
  kwic2 <- as.data.table(kwic2)
  M <- nrow(kwic1)
  N <- nrow(kwic2)
  
  # Identify keyword strings associated with synsets
  SMQ <- as.character(with(stack(syn), ind[match(kwic1$KEYWORD, values)]))
  SMS <- as.character(with(stack(syn), ind[match(kwic2$KEYWORD, values)]))
  SMQ[is.na(SMQ)]   <- ""
  SMS[is.na(SMS)]   <- ""
  
  # Identify strings for semantic matching
  ind_sem <- SMQ != ""
  ind_semS <- SMS != ""
  
  # Pre-build lookup ONCE before the loop
  sem_lookup <- data.table(
    synset = SMS[ind_semS],
    idkw   = kwic2$IDKW[ind_semS]
  )
  setkey(sem_lookup, synset)
  
  # Create progress bar
  invisible(capture.output(pb <- txtProgressBar(min = 0, max = max(kwic1$iter),
                                                style = 3)))
  message("Semantic matching")
  
  for (i in unique(kwic1$iter)) {
    in_iter <- (kwic1$iter == i)
    
    if (sum(in_iter & ind_sem) == 0) next # skip to next loop iteration
    
    rows_sem <- which(in_iter & ind_sem)
    
    results <- vapply(rows_sem, function(row_idx) {
      hits <- sem_lookup[.(SMQ[row_idx]), idkw, nomatch = NULL]
      paste(hits, collapse = ", ")
    }, character(1))
    
    # Fetch duplicates to a new column
    if (sum(in_iter & ind_sem) > 0 & sum(ind_semS) > 0) {
      kwic1[rows_sem, SemanticdupIDKW := results]
    }
    
    rm(in_iter)
    
    # update progress bar
    setTxtProgressBar(pb, i)
    cat("\rBlock", i, "/", max(kwic1$iter), "|")
  }
  close(pb)
  
  rm(ind_sem, ind_semS)
  
  for (j in c("IDKW", "SemanticdupIDKW")) {
    set(kwic1, which(is.na(kwic1[[j]])), j, "")
  }
  
  kwic1[, SemanticdupID := gsub(":\\S+\\b", "", SemanticdupIDKW)]
  kwic1 <- dupsets(kwic1, "S", method = method)
  
  # Remove synsets with single/unique members
  kwic1 <- kwic1[!unlist(lapply(strsplit(gsub("*?\\[\\S+:", "", kwic1$IDKW),
                                         split = ", "),
                                function(x) length(unique(x)))) == 1, ]
  return(kwic1)
}

dupsets <- function(kwicout, type, method) {
  if (dim(kwicout)[2] == 6) {
    kwicout <- as.data.table(subset(kwicout, get(names(kwicout)[6]) != ""))
    setkey(kwicout, PRIM_ID)
    kwicout[, c("iter", "KEYWORD") := NULL]
    if (method == "b") {
      kwicout[,  (3) := paste(get(names(kwicout)[2]),
                              get(names(kwicout)[3]), sep = ", ")]
      kwicout[,  (4) := paste(get(names(kwicout)[1]),
                              get(names(kwicout)[4]), sep = ", ")]
    }
    kwicout[, IDKW := NULL]
    if (method != "b") {
      kwicout[, Ndup := stri_count_fixed(get(colnames(kwicout)[3]), ",")]
      kwicout <- subset(kwicout, Ndup != 0)
      kwicout[, Ndup := NULL]
    }
    # Merge by PRIM_ID, then by ID, then add TYPE
    # group by PRIM_ID
    id_agg   <- collapse_dedup(kwicout, by_col = "PRIM_ID", 
                               val_col = names(kwicout)[3])
    idkw_agg <- collapse_dedup(kwicout, by_col = "PRIM_ID", 
                               val_col = names(kwicout)[2])
    
    kwicout <- id_agg[idkw_agg, on = "PRIM_ID"]
    setnames(kwicout, names(kwicout)[2:3], c("ID", "IDKW"))
    
    # group by ID
    idkw_agg2 <- collapse_dedup(kwicout, by_col = "ID", 
                                val_col = "IDKW")
    
    kwicout <- idkw_agg2[, TYPE := type]
    setkey(kwicout, NULL)
    kwicout <- unique(kwicout)
    
    # Add SET_NO
    setkey(kwicout, "ID")
    kwicout[, SET_NO := as.factor(ID)]
    kwicout[, SET_NO := as.numeric(SET_NO)]
    setkey(kwicout, "SET_NO")
    
    # Add count
    kwicout[, COUNT := stri_count_fixed(ID, ",") + 1]
    kwicout <- subset(kwicout, COUNT != 1)
    
    # Finalise output
    setcolorder(kwicout, c("SET_NO", "TYPE", "ID", "IDKW", "COUNT"))
    setkey(kwicout, "SET_NO")
    setDF(kwicout)
    return(kwicout)
  }
}


fix.syn <- function(syn) {
  names(syn) <- NULL
  names(syn) <- paste0("SM", seq_along(syn))
  syn <- lapply(syn, sort)
  syn <- subset(syn, duplicated(syn) == FALSE)
  if (is.element(TRUE, sapply(syn, function(x) length(x) == 1))) {
    syn <- syn[!sapply(syn, function(x) length(x) == 1)]
    warning("synsets encountered in list 'syn' with length 1")
  }
  syncomb <- do.call("rbind", lapply(syn, embed, 2))
  gg <- graph.edgelist(syncomb, directed = F)
  x <- split(V(gg)$name, clusters(gg)$membership)
  x <- lapply(x, sort)
  x <- subset(x, duplicated(x) == FALSE)
  names(x) <- paste0("SM", seq_along(x))
  return(x)
}

collapse_dedup <- function(dt, by_col, val_col) {
  max_tokens <- max(stringi::stri_count_fixed(dt[[val_col]], ",") + 1L,
                    na.rm = TRUE)
  
  split_cols <- paste0(val_col, "_", seq_len(max_tokens))
  dt <- copy(dt)  # avoid modifying in place
  dt[, (split_cols) := tstrsplit(get(val_col), ", ", fixed = TRUE,
                                 fill = NA, type.convert = FALSE)]
  
  long <- melt(dt[, c(by_col, split_cols), with = FALSE],
               id.vars      = by_col,
               measure.vars = split_cols,
               value.name   = val_col,
               variable.name = "token_no",
               na.rm        = TRUE)
  
  # Drop empty strings 
  long <- long[get(val_col) != "" & !is.na(get(val_col))]
  
  long[, .(out = paste(sort(unique(get(val_col))), collapse = ", ")),
       by = by_col] |>
    setnames("out", val_col)
}

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/

#' Generate keyword counts
#' 
#' \code{KWCounts} generates keyword counts from PGR passport database 
#' fields(columns).
#' 
#' This function computes the keyword counts from PGR passport database 
#' fields(columns) specified in the \code{fields} argument. The first field is 
#' considered as the primary key or identifier and is not used for counting the 
#' keywords. Any strings given in the \code{excep} argument are ignored for 
#' generating the counts.
#' 
#' The keyword counts can give a rough indication of the completeness of the 
#' data in the database fields being used for identification of probable 
#' duplicates.
#' 
#' @param x A data frame.
#' @param fields A character vector with the names of fields(columns) of the 
#'   data frame from which KWIC index is to be generated. The first field is 
#'   considered as the primary key or identifier (see \strong{Details}).
#' @param excep A vector of the keywords not to be considered for the counts
#'   (see \strong{Details}).
#' @note For large number of exceptions and/or large data.frame computation of
#' keyword counts may take some time.
#' @examples
#' 
#' \dontshow{
#' threads_dt <- data.table::getDTthreads()
#' threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
#' data.table::setDTthreads(2)
#' 
#' data.table::setDTthreads(2)
#' Sys.setenv(`OMP_THREAD_LIMIT` = 2)
#' }
#' 
#' # Load PGR passport database
#' GN <- GN1000
#'
#' # Specify database fields to use as a vector
#' GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")
#' 
#' # Specify the exceptions as a vector
#' exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE", 
#'          "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT", 
#'          "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE", 
#'          "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R", 
#'          "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE", 
#'          "U", "VALENCIA", "VIRGINIA", "WHITE")
#'          
#' # Compute the keyword counts
#' GNKWCouts <- KWCounts(GN, GNfields, exep)
#' 
#' # Plot the keyword counts
#' bp <- barplot(table(GNKWCouts$COUNT),
#'               xlab = "Word count", ylab = "Frequency", col = "#CD5555")
#' text(bp, 0, table(GNKWCouts$COUNT),cex=1,pos=3)
#' 
#' \dontshow{
#' data.table::setDTthreads(threads_dt)
#' Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
#' }
#' 
#' @seealso \code{\link[stringi]{stri_count}}
#' @return A data frame with the keyword counts for each record.
#' @import data.table
#' @importFrom stringi stri_count
#' @export KWCounts
KWCounts <- function(x, fields, excep) {
  if (is.data.frame(x) == FALSE) {
    # Check if x is a data.frame and stop if not
    stop("x is not a data.frame")
  }
  if (is.vector(fields) == FALSE) {
    # Check if fields is a vector or not
    stop("fields is not a vector")
  }
  if (length(fields) == 1) {
    # Check if more than one field is given as input and stop if not
    stop("Only one field given as input")
  }
  if (is.element(FALSE, fields %in% colnames(x)) == TRUE) {
    # Check if fields are present in x and stop if not
    stop("One or more fields are missing in x")
  }
  # Check excep argument
  if (!is.null(excep) && is.vector(excep, mode = "character") == FALSE) {
    stop("'excep' is not a character vector")
  }
  if (!is.null(excep)) {
    excep <- toupper(excep)
  } else {
    excep <- ""
  }
  #setDT(x)
  x <- as.data.table(x[fields])
  # Convert the fields in x to character
  for (col in fields) set(x, j = col, value = as.character(x[[col]]))
  # Convert NAs to empty strings
  for (j in fields) set(x, which(is.na(x[[j]])), j, "")
  setDF(x)
  if (is.element("", x[fields[1]]) | is.element(TRUE,
                                                duplicated(x[fields[1]]))) {
    # Check primary key/ID is unique and not NULL
    stop("Primary key/ID field should be unique and not NULL\n Use PGRdup::ValidatePrimKey() to identify and rectify the aberrant records first")
  }
  #setDT(x)
  x <- as.data.table(x)
  # Remove exceptions
  if (length(excep) <= 100) {
    x[, `:=`(fields[-1], lapply(.SD, function(x) gsub(paste0(excep,
                                                             collapse = "|"), "", x))), .SDcols = fields[-1]]
  } else {
    iter <- rep(seq_len(length(excep)), each = 100, length.out = seq_len(length(excep)))
    for (i in unique(iter)) {
      in_iter <- iter == i
      x[, `:=`(fields[-1], lapply(.SD, function(x) gsub(paste0(excep[in_iter],
                                                               collapse = "|"), "", x))), .SDcols = fields[-1]]
    }
  }
  # Get the word counts
  x[, COUNT := stri_count(do.call(paste, c(.SD, sep = " ")),
                          regex = "\\S+"), .SDcols = fields[-1]]
  # Prepare output
  x[, fields[-1] := NULL]
  setnames(x, names(x), c("PRIM_ID", "COUNT"))
  setDF(x)
  return(x)
}

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/



#' Clean PGR passport data
#' 
#' \code{DataClean} cleans the data in a character vector according to the 
#' conditions in the arguments.
#' 
#' This function aids in standardization and preparation of the PGR passport 
#' data for creation of a KWIC index with \code{\link[PGRdup]{KWIC}} function 
#' and the identification of probable duplicate accessions by the 
#' \code{\link[PGRdup]{ProbDup}} function. It cleans the character strings in 
#' passport data fields(columns) specified as the input character vector 
#' \code{x} according to the conditions in the arguments in the same order. If 
#' the input vector \code{x} is not of type character, it is coerced to a 
#' character vector.
#' 
#' This function is designed particularly for use with fields corresponding to 
#' accession names such as accession ids, collection numbers, accession names 
#' etc. It is essentially a wrapper around the \code{\link[base:grep]{gsub}} base 
#' function with \code{\link[base]{regex}} arguments. It also converts all 
#' strings to upper case and removes leading and trailing spaces.
#' 
#' Commas, semicolons and colons which are sometimes used to separate multiple 
#' strings or names within the same field can be replaced with a single space 
#' using the logical arguments \code{fix.comma}, \code{fix.semcol} and 
#' \code{fix.col} respectively.
#' 
#' Similarly the logical argument \code{fix.bracket} can be used to replace all 
#' brackets including parenthesis, square brackets and curly brackets with
#' space.
#' 
#' The logical argument \code{fix.punct} can be used to remove all punctuation 
#' from the data.
#' 
#' \code{fix.space} can be used to convert all space characters such as tab, 
#' newline, vertical tab, form feed and carriage return to spaces and finally 
#' convert multiple spaces to single space.
#' 
#' \code{fix.sep} can be used to merge together accession identifiers 
#' composed of alphabetic characters separated from as series of digits by a 
#' space character. For example IR 64, PUSA 256 etc.
#' 
#' \code{fix.leadzero} can be used to remove leading zeros from accession name 
#' fields to facilitate matching to identify probable duplicates. e.g. IR0064 -> 
#' IR64
#' 
#' @param x A character vector. If not, coerced to character by 
#'   \code{as.character}.
#' @param fix.comma logical. If \code{TRUE}, all the commas are replaced by 
#'   space (see \strong{Details}).
#' @param fix.semcol logical. If \code{TRUE}, all the semicolons are replaced by
#'   space (see \strong{Details}).
#' @param fix.col logical. If \code{TRUE}, all the colons are replaced by space 
#'   (see \strong{Details}).
#' @param fix.bracket logical. If \code{TRUE}, all the brackets are replaced by 
#'   space (see \strong{Details}).
#' @param fix.punct logical. If \code{TRUE}, all punctuation characters are 
#'   removed (see \strong{Details}).
#' @param fix.space logical. If \code{TRUE}, all space characters are replaced 
#'   by space and multiple spaces are converted to single space (see 
#'   \strong{Details}).
#' @param fix.sep logical. If \code{TRUE}, space between alphabetic characters 
#'   followed by digits is removed (see \strong{Details}).
#' @param fix.leadzero logical. If \code{TRUE}, leading zeros are removed (see 
#'   \strong{Details}).
#' @return A character vector with the cleaned data converted to upper case. 
#'   \code{NAs} if any are converted to blank strings.
#' @seealso \code{\link[base:grep]{gsub}}, \code{\link[base]{regex}}, 
#'   \code{\link[PGRdup]{MergeKW}}, \code{\link[PGRdup]{KWIC}}, 
#'   \code{\link[PGRdup]{ProbDup}}
#' @examples
#' names <- c("S7-12-6", "ICG-3505", "U 4-47-18;EC 21127", "AH 6481", "RS   1",
#'            "AK 12-24", "2-5 (NRCG-4053)", "T78, Mwitunde", "ICG 3410",
#'            "#648-4 (Gwalior)", "TG4;U/4/47/13", "EC0021003")
#' DataClean(names)
#' @export
DataClean <- function(x, fix.comma = TRUE, fix.semcol = TRUE, fix.col = TRUE,
                       fix.bracket = TRUE, fix.punct = TRUE, fix.space = TRUE,
                       fix.sep = TRUE, fix.leadzero = TRUE) {
  if (!is.character(x)) {
    warning("x is not of type character; coerced to character")
    x <- as.character(x)
  }
    x[is.na(x)]   <- ""  # Convert NAs to empty strings
    # Convert all strings to upper case
    x <- vapply(x, FUN = toupper, FUN.VALUE = "character")
    if (fix.comma) {
    x <- gsub(pattern = ",", replacement = " ", x)  # Replace "," by space
    }
    if (fix.semcol) {
    x <- gsub(pattern = ";", replacement = " ", x)  # Replace ";" by space
    }
    if (fix.col) {
    x <- gsub(pattern = ":", replacement = " ", x)  # Replace ":" by space
    }
    # Replace brackets by space
    if (fix.bracket) {
    x <- gsub(pattern = "[](){}[]", replacement = " ", x)
    }
    if (fix.punct) {
    # Remove all punctuations
    x <- gsub(pattern = "[[:punct:]]", replacement = "", x, perl = TRUE)
    }
    if (fix.space) {
    # replace all space characters with space (" ")
    x <- gsub(pattern = "[[:space:]]", replacement = " ", x)
    # replace multiple spaces with single space
    x <- gsub(pattern = "([[:space:]])\\1+", replacement = "\\1", x)
    }
    if (fix.sep) {
    # Remove space when alphabetic characters followed by digits
    x <- gsub(pattern = "([a-zA-Z]) ([0-9])", replacement = "\\1\\2", x,
              perl = TRUE)
    }
    if (fix.leadzero) {
    # Remove leading zeros
    x <- gsub("(?<![0-9])0+", "", x, perl = TRUE)
    }
    # Remove leading and trailing spaces if any.
    x <- gsub("^\\s+|\\s+$", "", x)
    # Remove names attribute
    names(x) <- NULL
return(x)
}

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/



#' Functions to facilitate genebank managers in the identification of probable 
#' duplicate accessions from plant genetic resources (PGR) passport databases.
#' 
#' @title The PGRdup Package
#' @name PGRdup-package
#' @aliases PGRdup
#' @author J Aravind \email{j.aravind@@icar.org.in} \cr J Radhamani 
#'   \email{jalli.radhamani@@icar.gov.in} \cr Kalyani Srinivasan 
#'   \email{kalyani.srinivasan@@icar.gov.in} \cr B Ananda Subhash 
#'   \email{anandasubhash@@gmail.com} \cr RK Tyagi
#'   \email{rishi.tyagi@@icar.gov.in}
#' @import data.table
#' @import igraph
#' @import stringdist
#' @importFrom stringi stri_split_fixed
#' @keywords internal
"_PACKAGE"

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/


#' Create a KWIC index
#'
#' \code{KWIC} creates a Keyword in Context index from PGR passport database
#' fields.
#'
#' The function generates a Keyword in Context index from a data frame of a PGR
#' passport database based on the fields(columns) stated in the arguments, using
#' \code{\link[data.table]{data.table}} package.
#'
#' The first element of vector \code{fields} is considered as the primary key or
#' identifier which uniquely identifies all rows in the data frame.
#'
#' Cleaning of the data the input fields(columns) using the
#' \code{\link[PGRdup]{DataClean}} function with appropriate arguments is
#' suggested before running this function.
#'
#' @param x A data frame from which KWIC index is to be generated.
#' @param fields A character vector with the names of fields(columns) of the
#'   data frame from which KWIC index is to be generated. The first field is
#'   considered as the primary key or identifier (see \strong{Details}).
#' @param min.freq Frequency of keywords are not computed if below
#'   \code{min.freq}. Default is 10.
#' @return A list of class \code{KWIC} containing the following components:
#'   \tabular{ll}{ \code{KWIC} \tab The KWIC index in the form of a data frame.
#'   \cr \code{KeywordFreq} \tab A data frame of the keywords detected with
#'   frequency greater than \code{min.freq}. \cr \code{Fields} \tab A character
#'   vector with the names of the PGR database fields from which the keywords
#'   were extracted. \cr }
#' @seealso \code{\link[data.table]{data.table}},
#'   \code{\link[PGRdup]{print.KWIC}}
#'
#' @references Kn\enc{ü}{ue}pffer, H. 1988. "The European Barley Database of the ECP/GR:
#'   An Introduction." \emph{Die Kulturpflanze} 36 (1): 135-62.
#'   \doi{10.1007/BF01986957}.
#'
#'   Kn\enc{ü}{ue}pffer, H., L. Frese, and M. W. M. Jongen. 1997. "Using Central Crop
#'   Databases: Searching for Duplicates and Gaps." In \emph{Central Crop
#'   Databases: Tools for Plant Genetic Resources Management. Report of a
#'   Workshop, Budapest, Hungary, 13-16 October 1996}, edited by E. Lipman, M.
#'   W. M. Jongen, T. J. L. van Hintum, T. Gass, and L. Maggioni, 67-77. Rome,
#'   Italy and Wageningen, The Netherlands: International Plant Genetic
#'   Resources Institute and Centre for Genetic Resources.
#'
#' @encoding UTF-8
#' @examples
#' 
#' \dontshow{
#' threads_dt <- data.table::getDTthreads()
#' threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
#' data.table::setDTthreads(2)
#' 
#' data.table::setDTthreads(2)
#' Sys.setenv(`OMP_THREAD_LIMIT` = 2)
#' }
#' 
#' # Load PGR passport database
#' GN <- GN1000
#'
#' # Specify as a vector the database fields to be used
#' GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")
#'
#' # Clean the data
#' GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
#'
#' \dontrun{
#'
#' # Generate KWIC index
#' GNKWIC <- KWIC(GN, GNfields)
#' GNKWIC
#'
#' # Retrieve the KWIC index from the KWIC object
#' KWIC <- GNKWIC[[1]]
#'
#' # Retrieve the keyword frequencies from the KWIC object
#' KeywordFreq <- GNKWIC[[2]]
#'
#' # Show error in case of duplicates and NULL values
#' # in the primary key/ID field "NationalID"
#' GN[1001:1005,] <- GN[1:5,]
#' GN[1001,3] <- ""
#' GNKWIC <- KWIC(GN, GNfields)
#' }
#' 
#' \dontshow{
#' data.table::setDTthreads(threads_dt)
#' Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
#' }
#' 
#' @import data.table
#' @importFrom stringi stri_split_fixed
#' @export KWIC
#' @export print.KWIC
#' @rdname KWIC
KWIC <- function(x, fields, min.freq = 10) {
  if (is.data.frame(x) == FALSE) {
    # Check if x is a data.frame and stop if not
    stop("x is not a data.frame")
  }
  if (is.vector(fields) == FALSE) {
    # Check if fields is a vector or not
    stop("fields is not a vector")
  }
  if (length(fields) == 1) {
    # Check if more than one field is given as input and stop if not
    stop("Only one field given as input")
  }
  if (is.element(FALSE, fields %in% colnames(x)) == TRUE) {
    # Check if fields are present in x and stop if not
    stop("One or more fields are missing in x")
  }
  #setDT(x)
  x <- as.data.table(x)
  # Convert the fields in x to character
  for (col in fields) set(x, j = col, value = as.character(x[[col]]))
  # Convert NAs to empty strings
  for (j in fields) set(x, which(is.na(x[[j]])), j, "")
  setDF(x)
  if (is.element("", x[fields[1]]) | is.element(TRUE,
                                                duplicated(x[fields[1]]))) {
    # Check primary key/ID is unique and not NULL
    stop("Primary key/ID field should be unique and not NULL\n Use PGRdup::ValidatePrimKey() to identify and rectify the aberrant records first")
  }
  #setDT(x)
  x <- as.data.table(x)
  # Create context fields
  x[, KWIC := do.call(paste, c(.SD, sep = " = ")), .SDcols = fields]
  x[, COMBINED := do.call(paste, .SD), .SDcols = fields]
  # Create KWIC index using data.table
  K <-  as.list(rep(NA, length(fields)))
  for (i in 1:(length(fields))) {
    K[[i]] <-  x[, list(KEYWORD = unlist(strsplit(get(fields[i]), " ")),
                        FIELD = fields[i]),
                 by = list(PRIM_ID = get(fields[1]), KWIC)]
    K[[i]] <- K[[i]][!is.na(K[[i]]$KEYWORD), ]
  }
  KWIC <- rbindlist(K)
  rm(K, x)
  #KWIC$KEYWORD[is.na(KWIC$KEYWORD)] <- ""
  set(KWIC, which(is.na(KWIC[["KEYWORD"]])), "KEYWORD", "")
  KWIC <- setkey(KWIC, KEYWORD)
  # Remove all '\' from KWIC
  KWIC[, KWIC := gsub(pattern = "([\\])", replacement = "", x = KWIC)]
  KWIC[, KEYWORD := gsub(pattern = "([\\])", replacement = "", x = KEYWORD)]
  # Remove records with blank keywords
  KWIC <- subset(KWIC, KEYWORD != "")
  # Remove duplicate records
  KWIC <- setkey(KWIC, NULL)
  KWIC <- unique(KWIC)
  # Add padding space in KWIC
  KWIC[, KWIC := paste(" ", KWIC, " ")]
  # Escape all Regex special characters
  KWIC[, KEYWORD := gsub(pattern = "([.|()\\^{}+$*?]|\\[|\\])",
                         replacement = "\\\\\\1", x = KEYWORD)]
  # Highlight keywords in KWIC
  KWIC[, KWIC := mapply(gsub, pattern = paste0(" ", KEYWORD, " "),
                        replacement = paste0(" <<", KEYWORD, ">> "), KWIC)]
  KWIC[, KWIC := gsub("^\\s+|\\s+$", "", KWIC)]
  # Unescape all Regex special characters
  KWIC[, KEYWORD := gsub(pattern = "\\\\(.)", replacement = "\\1", x = KEYWORD)]
  # Split KWIC
  KWIC[, c("KWIC_L", "KW1") := do.call(rbind.data.frame,
                                 stri_split_fixed(KWIC, "<<", 2))][]
  KWIC[, c("KWIC_KW", "KWIC_R") := do.call(rbind.data.frame,
                                stri_split_fixed(KW1, ">>", 2))][]
  cols <- c("KWIC_L", "KWIC_KW", "KWIC_R")
  KWIC[, (cols) := lapply(.SD, as.character), .SDcols = cols]
  KWIC[, KW1 := NULL]
  # Clean output data.frame
  KWIC <- setkey(KWIC, FIELD)
  KWIC <- setkey(KWIC, PRIM_ID)
  setcolorder(KWIC, c("PRIM_ID", "FIELD", "KEYWORD", "KWIC", "KWIC_L",
                      "KWIC_KW", "KWIC_R"))
  KWICIndex <- list(KWIC = NULL, KeywordFreq = NULL, Fields = fields)
  #KWICIndex[[1]] <- as.data.frame(KWIC)
  KWICIndex[[1]] <- setDF(KWIC)
  # Get keyword freq
  kwf <- as.data.frame(table(KWIC$KEYWORD))
  kwf <- subset(kwf, Freq > min.freq)
  kwf <- kwf[order(-kwf$Freq), ]
  rownames(kwf) <- NULL
  setnames(kwf, old = "Var1", new = "Keyword")
  KWICIndex[[2]] <- kwf
  # Set Class
  class(KWICIndex) <- "KWIC"
  return(KWICIndex)
}

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/



#' 'Double Metaphone' phonetic algorithm
#'
#' \code{DoubleMetaphone} converts strings to double metaphone phonetic codes.
#'
#' An implementation of the Double Metaphone phonetic algorithm in \code{R}. If
#' non-ASCII characters encountered in the input character vector \code{str}, a
#' warning is issued and they are transliterated so that the accented characters
#' are converted to their ASCII unaccented versions.
#'
#' @param str A character vector whose strings are to be encoded by double
#'   metaphone algorithm.
#' @return Returns a list with two character vectors of the same length as the
#'   input vector. The first character vector contains the primary double
#'   metaphone encodings, while the second character vector contains the
#'   alternate encodings.
#' @seealso \code{\link[stringdist]{phonetic}},
#'   \href{https://cran.r-project.org/package=RecordLinkage}{\code{phonetics}}
#'
#' @section Acknowledgement: The \code{C} code for the double metaphone
#'   algorithm was adapted from Maurice Aubrey's perl module hosted at the
#'   \strong{gitpan/Text-DoubleMetaphone}
#'   \href{https://github.com/gitpan/Text-DoubleMetaphone/blob/master/double_metaphone.c}{public
#'    github library} along with the corresponding
#'   \href{https://github.com/gitpan/Text-DoubleMetaphone/blob/master/README}{license
#'    information}.
#' @references Philips, Lawrence. 2000. "The Double Metaphone Search Algorithm."
#'   \emph{C/C++ Users Journal} 18 (6): 38-43.
#'   \url{https://dl.acm.org/doi/10.5555/349124.349132}.
#' @encoding UTF-8
#' @note In case of non-ASCII characters in strings, a warning is issued and
#'   accented characters are converted to their ASCII unaccented versions.
#' @examples
#' # Return the primary and secondary Double Metaphone encodings for a character vector.
#' str1 <- c("Jyothi", "Jyoti")
#' str2 <- c("POLLACHI", "BOLLACHI")
#' DoubleMetaphone(str1)
#' DoubleMetaphone(str2)
#' \dontrun{
#' # Issue a warning in case of non-ASCII characters.
#' str3 <- c("J\xf5geva", "Jogeva")
#' DoubleMetaphone(str3) }
#' @export
#' @useDynLib PGRdup, .registration = TRUE
DoubleMetaphone <- function(str) {
  if (is.character(str) == FALSE) {
    # Check if str is a character vector and stop if not
    stop("str is not a character vector")
  }
  if (any(grepl("NON_ASCII",
                iconv(str, "latin1", "ASCII", sub = "NON_ASCII"))) == TRUE) {
    str <- iconv(str, to = "ASCII//TRANSLIT")
    warning("Non-ASCII characters were encountered.")
  }
  out <- .C("fdouble_metaphone", as.character(str),
            primary =  as.character(str),
            alternate =  as.character(str),
            len = length(str), PACKAGE = "PGRdup")
  out <- out[2:3]
  return(out)
}

# character(length(str))

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/

#' Prints summary of \code{KWIC} object.
#' 
#' \code{print.KWIC} prints to console the summary of an object of class 
#' \code{KWIC} including the database fields(columns) used, the total number of
#' keywords and the number of distinct keywords in the index.
#' 
#' @param x An object of class \code{KWIC}.
#' @param ... Unused
#' @seealso \code{\link[PGRdup]{KWIC}}
#'   
#' @export
print.KWIC <- function(x, ...) {
  cat(paste("KWIC fields : ", sep = ""))
  cat(paste(x$Fields, sep = ""))
  cat(paste("\n", "Number of keywords : ", nrow(x[[1]]), sep = ""))
  cat(paste("\n", "Number of distinct keywords : ",
            length(unique(x[[1]]$KEYWORD)), sep = ""))
}

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/

#' Split an object of class \code{ProbDup}
#' 
#' \code{SplitProbDup} splits an object of class \code{ProbDup} into two on the 
#' basis of set counts.
#' 
#' @param pdup An object of class \code{ProbDup}.
#' @param splitat A vector of 3 integers indicating the set count at which 
#'   Fuzzy, Phonetic and Semantic duplicate sets in \code{pdup} are to be split.
#' @return A list with the the divided objects of class \code{ProbDup} 
#'   (\code{pdup1} and \code{pdup2}) along with the corresponding lists of
#'   accessions present in each (\code{list1} and \code{list2}).
#' @seealso \code{\link[PGRdup]{ProbDup}}, \code{\link[PGRdup]{MergeProbDup}}
#' @examples
#' 
#' \dontshow{
#' threads_dt <- data.table::getDTthreads()
#' threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
#' data.table::setDTthreads(2)
#' 
#' data.table::setDTthreads(2)
#' Sys.setenv(`OMP_THREAD_LIMIT` = 2)
#' }
#' 
#' \dontrun{
#' # Load PGR passport database
#' GN <- GN1000
#'
#' # Specify as a vector the database fields to be used
#' GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")
#'
#' # Clean the data
#' GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
#' y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
#' c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
#' c("Mota", "Company"))
#' y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
#' y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
#'         "Bunch", "Peanut")
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))
#' 
#' # Generate KWIC index
#' GNKWIC <- KWIC(GN, GNfields)
#'
#' # Specify the exceptions as a vector
#' exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE", 
#'          "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT", 
#'          "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE", 
#'          "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R", 
#'          "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE", 
#'          "U", "VALENCIA", "VIRGINIA", "WHITE")
#'           
#' # Specify the synsets as a list
#' syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))
#'
#' # Fetch probable duplicate sets
#' GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, fuzzy = TRUE,
#'                  phonetic = TRUE, encoding = "primary", 
#'                  semantic = TRUE, syn = syn)
#'                  
#' # Split the probable duplicate sets
#' GNdupSplit <- SplitProbDup(GNdup, splitat = c(10, 10, 10))
#' 
#' }
#' 
#' \dontshow{
#' data.table::setDTthreads(threads_dt)
#' Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
#' }
#' 
#' @import data.table
#' @importFrom methods is
#' @export


SplitProbDup <- function(pdup, splitat = c(30, 30, 30)) {
  if (!is(pdup, "ProbDup")) {
    stop('"pdup" is not of class ProbDup')
  }
  N <- length(seq_along(pdup))
  outnames <- attributes(pdup)$names
  out1 <- vector("list", length(outnames))
  names(out1) <- outnames
  rm(outnames)
  attr(out1, "method") <- attributes(pdup)$method
  attr(out1, "fields") <- attributes(pdup)$fields
  class(out1) <- class(pdup)
  out2 <- out1
  pdup <- lapply(pdup, function(x) if (!is.null(x)) as.data.table(x))
  list1 <- data.table(TYPE = vector(mode = "character"),
                      V1 = vector(mode = "character"))
  list2 <- data.table(TYPE = vector(mode = "character"),
                      V1 = vector(mode = "character"))
  for (i in 1:N) {
    if (!is.null(pdup[[i]])) {
      out1[[i]] <- subset(pdup[[i]], COUNT <= splitat[[i]])
      out2[[i]] <- subset(pdup[[i]], COUNT > splitat[[i]])
      out1[[i]][, SET_NO := as.numeric(as.factor(SET_NO))]
      out2[[i]][, SET_NO := as.numeric(as.factor(SET_NO))]
      if (!dim(out1[[i]])[1] == 0) {
        list1 <- rbind(list1,
                       out1[[i]][, .(unlist(strsplit(ID, ", ", TRUE))),
                                 by = list(TYPE)])
      }
      if (!dim(out2[[i]])[1] == 0) {
        list2 <- rbind(list2,
                       out2[[i]][, .(unlist(strsplit(ID, ", ", TRUE))),
                                 by = list(TYPE)])
      }
      out1[[i]] <- setDF(out1[[i]])
      out2[[i]] <- setDF(out2[[i]])
      if (dim(out1[[i]])[1] == 0) {
        out1[i] <- list(NULL)
        }
      if (dim(out2[[i]])[1] == 0) {
        out2[i] <- list(NULL)
        }
    }
  }
  if (dim(list1)[1] == 0) {
    list1 <- NULL
    }
  if (dim(list2)[1] == 0) {
    list2 <- NULL
    }
  if (!is.null(list1)) {
    list1 <- unique(list1, by = colnames(list1))
    list1[, PRIM_ID := gsub("\\[K1\\]", "", V1, perl = TRUE)]
    list1[, PRIM_ID := gsub("\\[K2\\]", "", PRIM_ID, perl = TRUE)]
    list1[, V1 := gsub("(\\]).*", "\\1", V1, perl = TRUE)]
    setnames(list1, "V1", "K")
    setDF(list1)
  }
  if (!is.null(list2)) {
    list2 <- unique(list2, by = colnames(list2))
    list2[, PRIM_ID := gsub("\\[K1\\]", "", V1, perl = TRUE)]
    list2[, PRIM_ID := gsub("\\[K2\\]", "", PRIM_ID, perl = TRUE)]
    list2[, V1 := gsub("(\\]).*", "\\1", V1, perl = TRUE)]
    setnames(list2, "V1", "K")
    setDF(list2)
  }
  out <- list(pdup1 = NULL, list1 = NULL,
              pdup2 = NULL, list2 = NULL)
  if (is.element(FALSE, lapply(out1, is.null))) {
    out[[1]] <- out1
    }
  if (!is.null(list1)) {
    out[[2]] <- list1
    }
  if (is.element(FALSE, lapply(out2, is.null))) {
    out[[3]] <- out2
    }
  if (!is.null(list2)) {
    out[[4]] <- list2
    }
  rm(out1, out2, list1, list2)
  return(out)
}

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/



#' Parse an object of class \code{ProbDup} to a data frame.
#' 
#' \code{ParseProbDup} converts an object of class \code{ProbDup} to a data frame for export.
#' 
#' @param pdup An object of class \code{ProbDup}.
#' @param max.count The maximum count of probable duplicate sets which are to be parsed to a data frame.
#' @param insert.blanks logical. If \code{TRUE}, inserts a row of \code{NAs} 
#'   after each set.
#' @return A data frame of the long/narrow form of the probable duplicate sets 
#'   data with the following core columns: \tabular{ll}{ 
#'   \code{SET_NO} \tab The set number. \cr \code{TYPE} \tab The type of 
#'   probable duplicate set. 'F' for fuzzy, 'P' for phonetic and 'S' for 
#'   semantic matching sets. \cr \code{K} \tab The KWIC index or database of 
#'   origin of the record. The \code{method} is specified within the square 
#'   brackets in the column name.  \cr \code{PRIM_ID} \tab The primary ID of the
#'   accession record from which the set could be identified. \cr \code{IDKW} 
#'   \tab The 'matching' keywords along with the IDs. \cr \code{COUNT} \tab The number of elements in a set. \cr } For the 
#'   retrieved columns(fields) the prefix \code{K*} indicates the KWIC index of 
#'   origin.
#' @examples
#' 
#' \dontshow{
#' threads_dt <- data.table::getDTthreads()
#' threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
#' data.table::setDTthreads(2)
#' 
#' data.table::setDTthreads(2)
#' Sys.setenv(`OMP_THREAD_LIMIT` = 2)
#' }
#' 
#' \dontrun{
#' 
#' #' # Load PGR passport database
#' GN <- GN1000
#'
#' # Specify as a vector the database fields to be used
#' GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")
#'
#' # Clean the data
#' GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
#' y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
#' c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
#' c("Mota", "Company"))
#' y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
#' y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
#'         "Bunch", "Peanut")
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))
#' 
#' # Generate KWIC index
#' GNKWIC <- KWIC(GN, GNfields)
#'
#' # Specify the exceptions as a vector
#' exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE", 
#'          "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT", 
#'          "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE", 
#'          "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R", 
#'          "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE", 
#'          "U", "VALENCIA", "VIRGINIA", "WHITE")
#'           
#' # Specify the synsets as a list
#' syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))
#'
#' # Fetch probable duplicate sets
#' GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, fuzzy = TRUE,
#'                  phonetic = TRUE, encoding = "primary", 
#'                  semantic = TRUE, syn = syn)
#'                  
#' # Convert to data frame of sets               
#' GNdupParsed <- ParseProbDup(GNdup)
#' 
#' }
#' 
#' \dontshow{
#' data.table::setDTthreads(threads_dt)
#' Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
#' }
#' 
#' @seealso 
#'   \code{\link[PGRdup]{ProbDup}},
#' @import data.table
#' @importFrom methods is
#' @export
ParseProbDup <- function(pdup, max.count = 30,
                          insert.blanks  = TRUE) {
  if (!is(pdup, "ProbDup")) {
    stop('"pdup" is not of class ProbDup')
  }
  N <- length(seq_along(pdup))
  types <- c("F", "P", "S", "D")
  types2 <- c("Fuzzy", "Phonetic", "Semantic", "Disjoint")
  for (i in 1:N) {
    if (!is.null(pdup[[i]])) {
      #setDT(pdup[[i]])
      pdup[[i]] <- as.data.table(pdup[[i]])
      pdup[[i]] <- subset(pdup[[i]], COUNT <= max.count)
      pdup[[i]] <- unique(pdup[[i]])
      # Reset SET_NO to take into account deleted sets with coutn > max.count
      pdup[[i]][, Seq := 1:.N]
      # Cast ID and IDKW by SET_NO
      pdup[[i]] <- pdup[[i]][, .(unlist(strsplit(IDKW, ", ", TRUE))),
                             by = list(SET_NO, TYPE, COUNT)][,
                               .(IDKW = toString(V1)), .(SET_NO, TYPE, COUNT,
                                                         PRIM_ID = gsub(":.*",
                                                                        "",
                                                                        V1))]
    }
  }
  # rbind pdup list
  pdup <- rbindlist(pdup)
  # Split K* from PRIM_ID column
  pdup[, PRIM_ID := gsub("\\]", "\\]_", PRIM_ID, perl = TRUE)]
  pdup[, K := gsub("_.*", "", PRIM_ID, perl = TRUE)]
  pdup[, PRIM_ID := gsub("\\[K1\\]_", "", PRIM_ID, perl = TRUE)]
  pdup[, PRIM_ID := gsub("\\[K2\\]_", "", PRIM_ID, perl = TRUE)]
  # Reset column order
  nameslist <- c("SET_NO", "TYPE", "K", "PRIM_ID", "IDKW", "COUNT")
  setcolorder(x = pdup, neworder = nameslist)
  # Insert blanks
  setkey(pdup, NULL)
  if (insert.blanks  == TRUE) {
    pdup[, TEMP := as.factor(SET_NO)]
    pdup[, TEMP := interaction(pdup$TEMP, as.factor(pdup$TYPE), drop = TRUE)]
    setattr(pdup$TEMP, "levels", seq(from = 1, to = length(levels(pdup$TEMP))))
    pdup[, TEMP := as.numeric(TEMP)]
    pdup <- setDT(pdup)[pdup[, c(.I, NA), TEMP]$V1][!.N]
    pdup[, TEMP := NULL]
    pdup[is.na(TYPE), TYPE := ""]
  }
  setDF(pdup)
  return(pdup)
}

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/

#' Merge two objects of class \code{ProbDup}
#' 
#' \code{MergeProbDup} merges two objects of class \code{ProbDup} into a single
#' one.
#' 
#' @param pdup1 An object of class \code{ProbDup}.
#' @param pdup2 An object of class \code{ProbDup}.
#' @return An object of class \code{ProbDup} with the merged list of  fuzzy, phonetic
#'   and semantic probable duplicate sets.
#' @seealso \code{\link[PGRdup]{ProbDup}}, \code{\link[PGRdup]{SplitProbDup}}
#' @examples
#' 
#' \dontshow{
#' threads_dt <- data.table::getDTthreads()
#' threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
#' data.table::setDTthreads(2)
#' 
#' data.table::setDTthreads(2)
#' Sys.setenv(`OMP_THREAD_LIMIT` = 2)
#' }
#' 
#' \dontrun{
#' #' # Load PGR passport database
#' GN <- GN1000
#' 
#' # Specify as a vector the database fields to be used
#' GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")
#' 
#' # Clean the data
#' GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
#' y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
#' c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
#' c("Mota", "Company"))
#' y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
#' y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
#'         "Bunch", "Peanut")
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))
#' 
#' # Generate KWIC index
#' GNKWIC <- KWIC(GN, GNfields)
#' 
#' # Specify the exceptions as a vector
#' exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE", 
#'          "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT", 
#'          "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE", 
#'          "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R", 
#'          "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE", 
#'          "U", "VALENCIA", "VIRGINIA", "WHITE")
#'           
#' # Specify the synsets as a list
#' syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))
#' 
#' # Fetch probable duplicate sets
#' GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, fuzzy = TRUE,
#'                  phonetic = TRUE, encoding = "primary", 
#'                  semantic = TRUE, syn = syn)
#'                  
#' # Split the probable duplicate sets
#' GNdupSplit <- SplitProbDup(GNdup, splitat = c(10, 10, 10))
#' 
#' # Merge the split sets
#' GNdupMerged <- MergeProbDup(GNdupSplit[[1]], GNdupSplit[[3]])
#' 
#' }
#' 
#' \dontshow{
#' data.table::setDTthreads(threads_dt)
#' Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
#' }
#' 
#' @import data.table
#' @importFrom methods is
#' @export

MergeProbDup <- function(pdup1, pdup2) {

  if (!is(pdup1, "ProbDup")) {
    stop('"pdup1" is not of class ProbDup')
  }
  if (!is(pdup2, "ProbDup")) {
    stop('"pdup2" is not of class ProbDup')
  }
  if (!identical(attributes(pdup1), attributes(pdup2))) {
    stop('Attributes of "pdup1" and "pdup2" are not identical')
  }
  N <- length(seq_along(pdup1))
  outnames <- attributes(pdup1)$names
  out <- vector("list", length(outnames))
  names(out) <- outnames
  rm(outnames)
  attr(out, "method") <- attributes(pdup1)$method
  attr(out, "fields") <- attributes(pdup1)$fields
  class(out) <- class(pdup1)
  pdup1 <- lapply(pdup1, as.data.table)
  pdup2 <- lapply(pdup2, as.data.table)
  for (i in 1:N) {
    out[[i]] <- rbind.data.frame(pdup1[[i]], pdup2[[i]])
    out[[i]] <- unique(out[[i]], by = c("TYPE", "ID", "IDKW", "COUNT"))
    out[[i]] <- setDF(out[[i]])
    if (dim(out[[i]])[1] == 0) {
      out[i] <- list(NULL)
    }
  }
  return(out)
}

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/



#' Reconstruct an object of class ProbDup
#' 
#' \code{ReconstructProbDup} reconstructs a data frame of probable duplicate 
#' sets created using the function \code{ReviewProbDup} and subjected to manual 
#' clerical review, back into an object of class \code{ProbDup}.
#' 
#' A data frame created using the function \code{\link[PGRdup]{ReviewProbDup}} 
#' from an object of class \code{ProbDup} for manual clerical review of 
#' identified probable duplicate sets can be reconstituted back to the same 
#' object after the review using this function. The instructions for modifying 
#' the sets entered in the appropriate format in the columns \code{DEL} and 
#' \code{SPLIT} during clerical review are taken into account for reconstituting
#' the probable duplicate sets.
#' 
#' Any records with \code{Y} in column \code{DEL} are deleted and records with 
#' identical integers in the column \code{SPLIT} other than the default \code{0}
#' are reassembled into a new set.
#' 
#' @param rev A data frame with the the core columns(fields) \code{SET_NO}, 
#'   \code{TYPE}, \code{K}, \code{PRIM_ID}, \code{DEL}, \code{SPLIT}, 
#'   \code{COUNT} and \code{IDKW}
#' @return An object of class \code{ProbDup} with the modified fuzzy,
#'   phonetic and semantic probable duplicate sets according to the instructions
#'   specified under clerical review.
#' @examples
#' 
#' \dontshow{
#' threads_dt <- data.table::getDTthreads()
#' threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
#' data.table::setDTthreads(2)
#' 
#' data.table::setDTthreads(2)
#' Sys.setenv(`OMP_THREAD_LIMIT` = 2)
#' }
#' 
#' \dontrun{
#'
#' # Load PGR passport database
#' GN <- GN1000
#'
#' # Specify as a vector the database fields to be used
#' GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")
#'
#' # Clean the data
#' GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
#' y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
#' c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
#' c("Mota", "Company"))
#' y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
#' y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
#'         "Bunch", "Peanut")
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))
#' 
#' # Generate KWIC index
#' GNKWIC <- KWIC(GN, GNfields)
#'
#' # Specify the exceptions as a vector
#' exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE", 
#'          "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT", 
#'          "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE", 
#'          "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R", 
#'          "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE", 
#'          "U", "VALENCIA", "VIRGINIA", "WHITE")
#'           
#' # Specify the synsets as a list
#' syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))
#'
#' # Fetch probable duplicate sets
#' GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, fuzzy = TRUE,
#'                  phonetic = TRUE, encoding = "primary", 
#'                  semantic = TRUE, syn = syn)
#' 
#' # Get disjoint probable duplicate sets of each kind
#' disGNdup <- DisProbDup(GNdup, combine = NULL)
#' 
#' # Get the data frame for reviewing the duplicate sets identified
#' RevGNdup <- ReviewProbDup(pdup = disGNdup, db1 = GN1000,
#'                           extra.db1 = c("SourceCountry", "TransferYear"), 
#'                           max.count = 30, insert.blanks = TRUE)
#' # Examine and review the duplicate sets using edit function
#' RevGNdup <- edit(RevGNdup)
#' 
#' # Examine and make changes to a set
#' subset(RevGNdup, SET_NO==12 & TYPE=="P", select= c(IDKW, DEL, SPLIT))
#' RevGNdup[c(110, 112, 114, 118, 121, 122, 124), 6] <- "Y"
#' RevGNdup[c(111, 115, 128), 7] <- 1
#' RevGNdup[c(113, 117, 120), 7] <- 2
#' RevGNdup[c(116, 119), 7] <- 3
#' RevGNdup[c(123, 125), 7] <- 4
#' RevGNdup[c(126, 127), 7] <- 5
#' subset(RevGNdup, SET_NO==12 & TYPE=="P", select= c(IDKW, DEL, SPLIT))
#' 
#' # Reconstruct ProDup object
#' GNdup2 <- ReconstructProbDup(RevGNdup)
#' lapply(disGNdup, nrow)
#' lapply(GNdup2, nrow)
#' 
#' }
#' 
#' \dontshow{
#' data.table::setDTthreads(threads_dt)
#' Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
#' }
#' 
#' @seealso \code{\link[PGRdup]{ProbDup}}, \code{\link[PGRdup]{ReviewProbDup}}
#' @import data.table
#' @importFrom stringi stri_count_fixed
#' @export
ReconstructProbDup <- function(rev) {
  # Check if rev is a data frame
  if (!is.data.frame(rev)) {
    stop("'rev' is not a data frame")
  }
  # Check if core fields are present
  core <- c("SET_NO", "TYPE", "PRIM_ID", "DEL", "SPLIT", "COUNT", "IDKW")
  core2 <- c("K[a]", "K[b]", "K[c]")
  j <- which(core2 %in% colnames(rev))
  core <- union(core, core2[j])
  if (is.element(FALSE, core %in% colnames(rev))) {
    # Check if core fields are present in rev and stop if not
    stop("One or more core fields are missing in 'rev'")
  }
  if (!is.numeric(rev$SET_NO)) {
    stop("'SET_NO' is not of class numeric or integer")
  }
  if (!is.character(rev$TYPE)) {
    stop("'TYPE' is not of class character")
  }
  if (!is.character(rev[, core2[j]])) {
    stop("'K[*]' is not of class character")
  }
  if (!is.character(rev$PRIM_ID)) {
    stop("'PRIM_ID' is not of class character")
  }
  if (!is.character(rev$IDKW)) {
    stop("'IDKW' is not of class character")
  }
  if (!is.numeric(rev$COUNT)) {
    stop("'COUNT' is not of class numeric or integer")
  }
  if (!is.character(rev$DEL)) {
    stop("'DEL' is not of class character")
  }
  if (!is.numeric(rev$SPLIT)) {
    stop("'SPLIT' is not of class numeric or integer")
  }
  # Retrieve method and fields
  method <- regmatches(core2[j], gregexpr("(?<=\\[).*?(?=\\])", core2[j],
                                          perl = TRUE))[[1]]
  fields <- list(k1 = NULL, k2 = NULL)
  fields[[1]] <- colnames(rev)[which(grepl("^K1_", colnames(rev),
                                           perl = TRUE))]
  if (method == "b" | method == "c" ) {
    fields[[2]] <- colnames(rev)[which(grepl("^K2_", colnames(rev),
                                             perl = TRUE))]
  }
  # Remove non-core fields
  #setDT(rev)
  rev <- as.data.table(rev)
  setkey(rev, SET_NO)
  #rev <- data.table(rev, key = "SET_NO")
  rev[, setdiff(colnames(rev), core) := NULL]
  rev <- rev[ TYPE != ""]
  rev[, SET_NO := as.factor(rev$SET_NO)]
  # Cleanup column DEL
  rev[, DEL := toupper(DEL)]
  rev[, DEL := as.factor(DEL)]
  setattr(rev$DEL, "levels",
          levels(rev$DEL)[levels(rev$DEL) != "Y" & levels(rev$DEL) != "N"] <- "N")
  # Delete sets according to column DEL
  rev <- rev[DEL != "Y"]
  # Cleanup column SPLIT
  rev[, SPLIT := suppressWarnings(as.integer(SPLIT))]
  invisible(rev[is.na(SPLIT), SPLIT := 0])
  rev[, SPLIT := as.factor(SPLIT)]
  # Split sets according to column SPLIT
  rev[, SET_NO := interaction(rev$SET_NO, rev$SPLIT, drop = TRUE)]
  rev[, SET_NO := factor(SET_NO, levels = sort(levels(SET_NO)))]
  setattr(rev$SET_NO, "levels", seq(from = 1, to = length(levels(rev$SET_NO))))
  rev[, SET_NO := as.numeric(SET_NO)]
  setkey(rev, key = "SET_NO")
  # Reconsturct
  rev[, c("DEL", "SPLIT", "COUNT") := NULL]
  rev[, PRIM_ID := paste(get(core2[j]), PRIM_ID, sep = "")]
  rev[, core2[j] := NULL]
  rev <- rev[, list(ID = paste0(sort(unique(PRIM_ID)), collapse = ", "),
                    IDKW = paste0(sort(unique(IDKW)), collapse = ", ")),
             by = c("TYPE", "SET_NO")]
  rev[, COUNT := stri_count_fixed(ID, ",") + 1]
  rev <- rev[COUNT != 1]
  setcolorder(rev, c("SET_NO", "TYPE", "ID", "IDKW", "COUNT"))
  rev <- unique(rev, by = c("TYPE", "ID", "IDKW", "COUNT"))
  out <- list(FuzzyDuplicates = NULL, PhoneticDuplicates = NULL,
              SemanticDuplicates = NULL, DisjointDupicates = NULL)
  attr(out, "method") <- method
  fields[[1]] <-  gsub("^K1_", "", fields[[1]])
  if (method == "b" | method == "c" ) {
    fields[[2]] <- gsub("^K2_", "", fields[[2]])
  }
  attr(out, "fields") <- fields
  rm(fields, method)
  types <- c("F", "P", "S", "D")
  # Reset the SET_NO
  setkey(rev, TYPE, ID, SET_NO)
  rev[, TYPE := as.factor(TYPE)][, SET_NO := as.integer(SET_NO)]
  rev[, SET_NO := 1:.N, by = TYPE]
  rev[, SET_NO := as.numeric(SET_NO)][, TYPE := as.character(TYPE)]
  N <- length(seq_along(out))
  for (i in 1:N) {
    out[[i]] <- setDF(subset(rev, TYPE == types[i]))
    #out[[i]] <- as.data.frame(subset(rev, TYPE == types[i]))
  }
  rev[, TYPE := as.factor(TYPE)]
  if (!"D" %in% levels(rev$TYPE)) {
    out[[4]] <- NULL
  }
  out[which(unlist(lapply(seq_along(out),
                          function(i) dim(out[[i]])[1])) == 0)] <- list(NULL)
  rm(rev)
  class(out) <- "ProbDup"
  return(out)
}

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/



#' Get disjoint probable duplicate sets
#'
#' \code{DisProbDup} finds and joins intersecting sets in an object of class
#' \code{ProbDup} to get disjoint probable duplicate sets.
#'
#' This function considers the accession primary keys/IDs for finding
#' intersecting sets and subsequently joins them to retrieve disjoint sets.
#' These operations are implemented utilizing the
#' \code{\link[igraph:aaa-igraph-package]{igraph}} package functions.
#'
#' Disjoint sets are retrieved either individually for each type of probable
#' duplicate sets or considering all type of sets simultaneously. In case of the
#' latter, the disjoint of all the type of sets alone are returned in the output
#' as an additional data frame \code{DisjointDuplicates} in an object of class
#' \code{ProbDup}
#'
#' @param pdup An object of class \code{ProbDup}.
#' @param combine A character vector indicating the type of sets to be
#'   considered together for retrieving disjoint sets. If \code{NULL}, then
#'   disjoint sets within each type are retrieved (see \strong{Details}).
#' @return Returns an object of class \code{ProbDup} with either the disjoint
#'   sets within each type - \code{FuzzyDuplicates}, \code{PhoneticDuplicates}
#'   and \code{SemanticDuplicates} when \code{combine = NULL} or the combined
#'   disjoint duplicate sets as an additional element \code{DisjointDupicates}
#'   according to the choice specified in the argument \code{combine}.
#' @examples
#' 
#' \dontshow{
#' threads_dt <- data.table::getDTthreads()
#' threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
#' data.table::setDTthreads(2)
#' 
#' data.table::setDTthreads(2)
#' Sys.setenv(`OMP_THREAD_LIMIT` = 2)
#' }
#' 
#' \dontrun{
#'
#' # Load PGR passport database
#' GN <- GN1000
#'
#' # Specify as a vector the database fields to be used
#' GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")
#'
#' # Clean the data
#' GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
#' y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
#' c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
#' c("Mota", "Company"))
#' y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
#' y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
#'         "Bunch", "Peanut")
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))
#'
#' # Generate KWIC index
#' GNKWIC <- KWIC(GN, GNfields)
#'
#' # Specify the exceptions as a vector
#' exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE",
#'          "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT",
#'          "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE",
#'          "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R",
#'          "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE",
#'          "U", "VALENCIA", "VIRGINIA", "WHITE")
#'
#' # Specify the synsets as a list
#' syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))
#'
#' # Fetch probable duplicate sets
#' GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, fuzzy = TRUE,
#'                  phonetic = TRUE, encoding = "primary",
#'                  semantic = TRUE, syn = syn)
#' lapply(GNdup, dim)
#'
#' # Get disjoint probable duplicate sets of each kind
#' disGNdup1 <- DisProbDup(GNdup, combine = NULL)
#' lapply(disGNdup1, nrow)
#'
#' # Get disjoint probable duplicate sets combining all the kinds of sets
#' disGNdup2 <- DisProbDup(GNdup, combine = c("F", "P", "S"))
#' lapply(disGNdup2, nrow)
#'
#' }
#' 
#' \dontshow{
#' data.table::setDTthreads(threads_dt)
#' Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
#' }
#' 
#' @seealso \code{\link[PGRdup]{ProbDup}}
#' @import igraph
#' @import data.table
#' @importFrom methods is
#' @importFrom stats embed
#' @importFrom utils stack
#' @importFrom stringi stri_count_fixed
#' @export
DisProbDup <- function(pdup, combine = c("F", "P", "S")) {
  # Check arguments
  if (!is(pdup, "ProbDup")) {
    stop('"pdup" is not of class ProbDup')
  }
  out <- list(FuzzyDuplicates = NULL, PhoneticDuplicates = NULL,
              SemanticDuplicates = NULL, DisjointDupicates = NULL)
  attr(out, "method") <- attr(pdup, "method")
  attr(out, "fields") <- attr(pdup, "fields")
  types <- c("F", "P", "S", "D")
  # Check if Disjoint duplicates is in list and convert combine and pdup
  if (!is.null(pdup[["DisjointDupicates"]])) {
    pdup[1:3] <- NULL
    combine <- NULL
    warning(paste("Only Disjoint duplicate sets encountered in 'pdup'",
                  "\nFurther disjoint sets, if any are returned", sep = ""))
  }
  cond <- length(!unlist(lapply(pdup[1:3],
                                is.null))[!unlist(lapply(pdup[1:3],
                                                         is.null))]) == 1
  if (is.null(pdup[["DisjointDupicates"]]) & cond) {
    combine <- NULL
    warning(paste("Only one kind of probable duplicate set encountered in 'pdup'",
                  "\nDisjoint sets within the same are returned", sep = ""))
  }
  if (!is.null(combine)) {
    combine <- match.arg(combine, c("F", "P", "S"), several.ok = TRUE)
    if (length(combine == 1) & is.element(TRUE,
                                          lapply(pdup[which(types %in% combine)],
                                                 is.null))) {
      p <- unlist(lapply(pdup[which(types %in% combine)], is.null))
      stop(paste("The following set specified in argument 'combine' is missing from 'pdup'",
                 "\n", paste("#", names(p[p]), collapse = "\n")))
    }
    # Check and report any missing sets specified in combine
    if (is.element(TRUE, lapply(pdup[which(types %in% combine)], is.null))) {
      p <- unlist(lapply(pdup[which(types %in% combine)], is.null))
      warning(paste("The following set specified in argument 'combine' is missing from 'pdup'",
                    "\n", paste("#", names(p[p]), collapse = "\n"),
                    "\nJoint disjoint of only the remaining sets are returned",
                    sep = ""))
    }
    # Rbind all in combine into pdup[4] and assign NULL to rest
    pdup[which(!types %in% combine)] <- list(NULL)
    pdup[[4]] <- rbindlist(pdup[1:3])
    pdup[which(types %in% combine)] <- list(NULL)
    pdup[[4]][, TYPE := "D"]
  }
  N <- length(seq_along(pdup))
  for (i in 1:N) {
    if (!is.null(pdup[[i]])) {
      # Get and cast the disjoint sets by PRIM_ID
      idlist <- strsplit(pdup[[i]]$ID, ", ")
      idcomb <- do.call("rbind", lapply(idlist, embed, 2))
      gg <- graph.edgelist(idcomb, directed = FALSE)
      disidlist <- split(V(gg)$name, clusters(gg)$membership)
      disidlist <- stack(lapply(disidlist, sort))
      #setDT(disidlist)
      disidlist <- as.data.table(disidlist)
      setkey(disidlist, ind)
      #disidlist <- data.table(stack(disidlist), key = "ind")
      disidlist[, ind := as.integer(ind)]
      setnames(disidlist, old = c("values", "ind"), new = c("ID", "SET_NO"))
      setkey(disidlist, ID)
      rm(gg, idlist, idcomb)
      #setDT(pdup[[i]])
      pdup[[i]] <- as.data.table(pdup[[i]])
      #pdup[[i]] <- data.table(pdup[[i]])
      pdup[[i]] <- pdup[[i]][, .(unlist(strsplit(IDKW, ", ", TRUE))), by = TYPE][,
                               .(IDKW = toString(sort(unique(unlist(strsplit(V1, ", ")))))),
                               .(TYPE, ID = gsub(":.*", "", V1))]
      pdup[[i]] <- unique(pdup[[i]])
      setkey(pdup[[i]], ID)
      disidlist <- merge(disidlist, pdup[[i]])
      disidlist <- disidlist[, list(ID = paste0(sort(unique(ID)),
                                                collapse = ", "),
                                    IDKW = paste0(sort(unique(IDKW)),
                                                  collapse = ", ")),
                             by = c("TYPE", "SET_NO")]
      disidlist[, COUNT := stri_count_fixed(ID, ",") + 1]
      setkey(disidlist, ID)
      disidlist[, SET_NO := seq(1, nrow(disidlist))]
      disidlist[, SET_NO := as.numeric(SET_NO)]
      setkey(disidlist, SET_NO)
      setcolorder(disidlist, c("SET_NO", "TYPE", "ID", "IDKW", "COUNT"))
      out[[i]] <- setDF(disidlist)
      #out[[i]] <- as.data.frame(disidlist)
    }
  }
  if (is.null(combine)) {
    out[[4]] <- NULL
  }
  class(out) <- "ProbDup"
  return(out)
}

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/



#' Merge keyword strings
#' 
#' These functions merge keyword strings separated by delimiters such as space, 
#' period or dash in a character vector into single keyword strings.
#' 
#' These functions aid in standardization of relevant data fields(columns) in 
#' PGR passport data for creation of a KWIC index with 
#' \code{\link[PGRdup]{KWIC}} function and subsequent identification of probable
#' duplicate accessions by the \code{\link[PGRdup]{ProbDup}} function.
#' 
#' It is recommended to run this function before using the 
#' \code{\link[PGRdup]{DataClean}} function on the relevant data fields(columns)
#' of PGR passport databases.
#' 
#' \code{MergeKW} merges together pairs of strings specified as a list in 
#' argument \code{y} wherever they exist in a character vector. The second 
#' string in the pair is merged even when it is followed by a number.
#' 
#' \code{MergePrefix} merges prefix strings specified as a character vector in 
#' argument \code{y} to the succeeding root word, wherever they exist in a 
#' character vector.
#' 
#' \code{MergeSuffix} merges suffix strings specified as a character vector in 
#' argument \code{y} to the preceding root word, wherever they exist in a 
#' character vector. The suffix strings which are followed by numbers are also 
#' merged.
#' 
#' @param x A character vector. If not, coerced to character by 
#'   \code{as.character}.
#' @param y A list of character vectors with pairs of strings that are to be 
#'   merged (for \code{MergeKW}) or a character vector of strings which are to 
#'   be merged to succeeding string (for \code{MergePrefix}) or the preceding 
#'   string (for \code{MergeSuffix}). If not of type character, coerced by 
#'   \code{as.character}.
#' @param delim Delimiting characters to be removed between keywords.
#' @return A character vector of the same length as \code{x} with the required 
#'   keyword strings merged.
#' @examples
#' names <- c("Punjab Bold", "Gujarat- Dwarf", "Nagpur.local", "SAM COL 144",
#'            "SAM COL--280", "NIZAMABAD-LOCAL", "Dark Green Mutant",
#'            "Dixie-Giant", "Georgia- Bunch", "Uganda-erect", "Small Japan",
#'            "Castle  Cary", "Punjab erect", "Improved small japan",
#'            "Dark Purple")
#'
#' # Merge pairs of strings
#' y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
#'            c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
#'            c("Mota", "Company"))
#' names <- MergeKW(names, y1, delim = c("space", "dash", "period"))
#'
#' # Merge prefix strings
#' y2 <- c("Light", "Small", "Improved", "Punjab", "SAM")
#' names <- MergePrefix(names, y2, delim = c("space", "dash", "period"))
#'
#' # Merge suffix strings
#' y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
#'         "Bunch", "Peanut")
#' names <- MergeSuffix(names, y3, delim = c("space", "dash", "period"))
#' @seealso \code{\link[PGRdup]{DataClean}}, \code{\link[PGRdup]{KWIC}}, 
#'   \code{\link[PGRdup]{ProbDup}}
#' @name MergeKW

#' @rdname MergeKW
#' @export
MergeKW <- function(x, y, delim = c("space", "dash", "period")) {
  if (!is.character(x)) {
    warning("x is not of type character; coerced to character")
    x <- as.character(x)
  }
  if (!is.list(y)) {
    stop("y is not a list")
  }
  if (is.element(FALSE, as.logical(lapply(y, function(x) length(x) == 2)))) {
    stop("list y is not in the appropriate format")
  }
  if (is.element(FALSE, as.logical(lapply(y, function(x) is.character(x))))) {
    warning("list y had non character vectors; coerced to character")
    y <-  as.logical(lapply(y, function(x) as.character(x)))
  }
  delim <- match.arg(delim, c("space", "dash", "period"), several.ok = TRUE)
  # replace all space characters with space (" ")
  x <- gsub(pattern = "[[:space:]]", replacement = " ", x)
  y <- unique(y)
  # Escape all Regex special characters in y
  y <- lapply(y, function(x) gsub(pattern = "([.|()\\^{}+$*?]|\\[|\\])",
                                  replacement = "\\\\\\1", x))
  options <- c("\\s", "\\-", ".")
  options2 <- logical(length = 3)
  if (is.element("space", delim)) {
    options2[1] <- TRUE
  }
  if (is.element("dash", delim)) {
    options2[2] <- TRUE
  }
  if (is.element("period", delim)) {
    options2[3] <- TRUE
  }
  p <-  paste0(options[options2], collapse = "")
  sapply(seq_len(length(y)), function(i){
    pat <- paste0("(?i)(?<=^", y[[i]][1], ")[", p, "]+(?=", y[[i]][2],
                  ")[[:digit:]]*\\b|\\b(?<=", y[[i]][1], ")[", p,
                  "]+(?=",  y[[i]][2], "[[:digit:]]*$)", collapse = "")
    x <<- gsub(pat, "", x, perl = TRUE)
    }
  )
  return(x)
}

#' @rdname MergeKW
#' @export
MergePrefix <- function(x, y, delim = c("space", "dash", "period")) {
  if (!is.character(x)) {
    warning("x is not of type character; coerced to character")
    x <- as.character(x)
  }
  if (!is.character(y)) {
    warning("x is not of type character; coerced to character")
    y <- as.character(y)
  }
  delim <- match.arg(delim, c("space", "dash", "period"), several.ok = TRUE)
  # replace all space characters with space (" ")
  x <- gsub(pattern = "[[:space:]]", replacement = " ", x)
  # Escape all Regex special characters in y
  y <- unique(toupper(y))
  y <- gsub(pattern = "([.|()\\^{}+$*?]|\\[|\\])",
            replacement = "\\\\\\1", y)
  options <- c("\\s", "\\-", ".")
  options2 <- logical(length = 3)
  if (is.element("space", delim)) {
    options2[1] <- TRUE
  }
  if (is.element("dash", delim)) {
    options2[2] <- TRUE
  }
  if (is.element("period", delim)) {
    options2[3] <- TRUE
  }
  p <-  paste0(options[options2], collapse = "")
  sapply(seq_len(length(y)), function(i){
    x <<- gsub(paste0("(?i)\\b(?<=^", y[i],
                      ")[", p, "]+"), "", x, perl = TRUE)
    }
  )
  return(x)
}

#' @rdname MergeKW
#' @export
MergeSuffix <- function(x, y, delim = c("space", "dash", "period")) {
  if (!is.character(x)) {
    warning("x is not of type character; coerced to character")
    x <- as.character(x)
  }
  if (!is.character(y)) {
    warning("x is not of type character; coerced to character")
    y <- as.character(y)
  }
  delim <- match.arg(delim, c("space", "dash", "period"), several.ok = TRUE)
  # replace all space characters with space (" ")
  x <- gsub(pattern = "[[:space:]]", replacement = " ", x)
  # Escape all Regex special characters in y
  y <- unique(toupper(y))
  y <- gsub(pattern = "([.|()\\^{}+$*?]|\\[|\\])",
            replacement = "\\\\\\1", y)
  options <- c("\\s", "\\-", ".")
  options2 <- logical(length = 3)
  if (is.element("space", delim)) {
    options2[1] <- TRUE
  }
  if (is.element("dash", delim)) {
    options2[2] <- TRUE
  }
  if (is.element("period", delim)) {
    options2[3] <- TRUE
  }
  p <-  paste0(options[options2], collapse = "")
  sapply(seq_len(length(y)), function(i){
      x <<- gsub(paste0("(?i)[", p, "]+(?=", y[i], "[[:digit:]]*",
                        "\\b)"), "", x, perl = TRUE)
      }
    )
  return(x)
}

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/



#' Sample groundnut PGR passport data
#'
#' Sample PGR passport data of 1000 groundnut accessions held in the Indian
#' National Genebank at National Bureau of Plant Genetic Resources (NBPGR), New
#' Delhi.
#'
#' @name GN1000
#' @usage data(GN1000)
#' @keywords datasets
#' @format A data frame having 1000 records with the following 10
#'   columns(fields): \describe{ \item{CommonName}{Common name} 
#'   \item{BotanicalName}{Botanical name} \item{NationalID}{NBPGR National 
#'   identifier} \item{CollNo}{Collector number} \item{DonorID}{Donor ID} 
#'   \item{OtherID1}{Other ID field 1} \item{OtherID2}{Other ID field 2} 
#'   \item{BioStatus}{Biological status} \item{SourceCountry}{Country of 
#'   origin} \item{TransferYear}{Year of transfer} }
#' @seealso \url{http://pgrportal.nbpgr.ernet.in/}
#' 
NULL

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/



if (getRversion() >= "2.15.1")  {
  utils::globalVariables(
    c(
      ".", "COMBINED", "COUNT", "DEL", "exep", "FACTOR",
      "FIELD", "fixed", "Freq", "FuzzydupID",
      "FuzzydupIDKW", "FuzzydupIDKW2", "genesysId", "ID",
      "IDKW", "ind", "iter", "K", "K1_PRIM_ID",
      "K2_PRIM_ID", "KEYWORD", "KIND", "KIND2", "KW1",
      "kwic", "label", "name", "Ndup", "NO.ACC", "NO.SETS",
      "PhoneticdupID", "PhoneticdupIDKW", "PRIM_ID",
      "primdup", "SemanticdupID", "SemanticdupIDKW", "Seq",
      "SET_NO", "shape", "shape2", "singles", "SPLIT",
      "TEMP", "Total", "TYPE", "V1", "value",
      "variable", "WITHIN", "ymax", "ymin",
      "digit", "enc", "idkw", "keyword", "synset"
    )
  )
}

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/

#' Prints summary of \code{ProbDup} object.
#' 
#' \code{print.ProbDup} prints to console the summary of an object of class 
#' \code{ProbDup} including the method used ("a", "b" or "c"), the database
#' fields(columns) considered, the number of probable duplicate sets of each
#' kind along with the corresponding number of records.
#' 
#' @param x An object of class \code{ProbDup}.
#' @param ... Unused
#' @seealso \code{\link[PGRdup]{ProbDup}}
#'   
#' @export
print.ProbDup <- function(x, ...) {
  attr <- attributes(x)
  x <- x[!sapply(x, is.null)]
  attributes(x) <- append(attributes(x), attr[2:4])
  m <- data.frame(`No. of Sets` = sapply(x, function(x) dim(x)[1]),
                  `No. of Records` = sapply(x,
                                            function(x) length(unique(unlist(strsplit(x$ID, split = ", "))))))
  Total <- as.character(colSums(m))
  Total[2] <- paste(Total[2], "(Distinct:",
                    length(unique(unlist(lapply(x,
                                                function(x) unlist(strsplit(x$ID, split = ", ")))))),
                    ")", sep = "")
  m <- rbind(m, Total = Total)
  cat(paste("Method : ", attributes(x)$method, "\n", sep = ""))
  cat(paste("\n", "KWIC1 fields : ", sep = ""))
  cat(paste(attributes(x)$fields$k1, sep = ""))
  if (attributes(x)$method != "a") {
    cat(paste("\n", "\n", "KWIC2 fields : ", sep = ""))
    cat(paste(attributes(x)$fields$k2, sep = ""))
  }
  cat("\n", "\n")
  print(m)
}

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/

#' Convert 'Darwin Core - Germplasm' zip archive to a flat file
#' 
#' \code{read.genesys} reads PGR data in a Darwin Core - germplasm zip 
#' archive downloaded from genesys database and creates a flat file
#' \code{data.frame} from it.
#' 
#' This function helps to import to R environment, the PGR data 
#' downloaded from genesys database \url{https://www.genesys-pgr.org/} as a 
#' Darwin Core - germplasm (DwC-germplasm) zip archive. The different csv files 
#' in the archive are merged as a flat file into a single \code{data.frame}.
#' 
#' All the space characters can be removed from the fields corresponding to 
#' accession names such as acceNumb, collNumb, ACCENAME, COLLNUMB, DONORNUMB and
#' OTHERNUMB using the argument \code{scrub.names.space} to facilitate creation 
#' of KWIC index with \code{\link[PGRdup]{KWIC}} function and subsequent
#' matching operations to identify probable duplicates with
#' \code{\link[PGRdup]{ProbDup}} function.
#' 
#' The argument \code{readme} can be used to print the readme file in the 
#' archive to console, if required.
#' 
#' @param zip.genesys A character vector giving the file path to the downloaded 
#'   zip file from Genesys.
#' @param scrub.names.space logical. If \code{TRUE}, all space characters are 
#'   removed from name field in names extension (see \strong{Details}).
#' @param readme logical. If \code{TRUE}, the genesys zip file readme is printed
#'   to console.
#' @return A data.frame with the flat file form of the genesys data.
#' @examples
#' 
#' \dontshow{
#' threads_dt <- data.table::getDTthreads()
#' threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
#' data.table::setDTthreads(2)
#' 
#' data.table::setDTthreads(2)
#' Sys.setenv(`OMP_THREAD_LIMIT` = 2)
#' }
#' 
#' \dontrun{
#' # Import the DwC-Germplasm zip archive "genesys-accessions-filtered.zip"
#' PGRgenesys <- read.genesys("genesys-accessions-filtered.zip",
#'                            scrub.names.space = TRUE, readme = TRUE)
#' }
#' 
#' \dontshow{
#' data.table::setDTthreads(threads_dt)
#' Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
#' }
#' 
#' @seealso \code{\link[data.table]{data.table}}
#' 
#' @import data.table
#' @importFrom utils unzip
#' @export
read.genesys <- function(zip.genesys, scrub.names.space = TRUE, readme = TRUE) {
  # Check whether archive is a Darwin Core - Germplasm Archive
  FileList <- unzip(zip.genesys, list = TRUE)$Name
  DwCList <- c("README.txt", "core.csv", "names.csv", "geo.csv", "coll.csv",
               "meta.xml")
  if (setequal(FileList, DwCList) == FALSE) {
    stop("The zip file is not a Genesys Darwin Core - Germplasm Archive")
  }
  # Create the temporary directory or flush CSVs if it exists already
  if (!file.exists(tempdir())) {
    dir.create(tempdir())
  } else {
    files <- c(paste(tempdir(), "\\names.csv", sep = ""),
               paste(tempdir(), "\\geo.csv", sep = ""),
               paste(tempdir(), "\\core.csv", sep = ""),
               paste(tempdir(), "\\coll.csv", sep = ""))
    for (i in 1:4) {
      if (file.exists(files[i])) file.remove(files[i])
    }
  }
  # Unzip the file into the dir
  unzip(zip.genesys, exdir = tempdir())
  # Import names.csv
  nam <- fread(input = list.files(tempdir(), pattern = "names.csv",
                                  full.names = T),
               header = TRUE, stringsAsFactors = FALSE,
               select = c("genesysId", "instCode", "name", "aliasType",
                          "lang", "version"),
               colClasses = rep("character", 6))
  setkey(nam, genesysId)
  if ( !length(setdiff(nam$aliasType,
                      c("ACCENAME", "COLLNUMB",
                        "DONORNUMB", "OTHERNUMB"))) == 0) {
    warning("Abnormal strings detected in 'aliasType' column of 'names.csv'")
  }

  # Remove space from acc name fields 1
  if (scrub.names.space) {
    nam[, name := gsub(pattern = "[[:space:]]", replacement = "", name)]
  }
  # Convert to wide form
  nam <- dcast.data.table(nam, genesysId ~ aliasType, value.var = "name",
                          fun.aggregate = function(x) paste(x, collapse = ":"))
  if (!length(setdiff(colnames(nam),
                     c("genesysId", "ACCENAME", "COLLNUMB",
                       "DONORNUMB", "OTHERNUMB"))) == 0) {
    nam[, setdiff(colnames(nam),
                  c("genesysId", "ACCENAME", "COLLNUMB",
                    "DONORNUMB", "OTHERNUMB")) := NULL]
  }

  # Import geo.csv
  geo <- fread(input = list.files(tempdir(), pattern = "geo.csv",
                                  full.names = T),
               header = TRUE, stringsAsFactors = FALSE,
               select = c("genesysId", "latitude", "longitude", "elevation",
                          "datum", "uncertainty", "method", "version"),
               colClasses = rep("character", 8))
  setkey(geo, genesysId)
  # Import core.csv
  cor <- fread(input = list.files(tempdir(), pattern = "core.csv",
                                  full.names = T),
               header = TRUE, stringsAsFactors = FALSE,
               select = c("genesysId", "uuid", "instCode", "acceNumb", "genus",
                          "species", "fullTaxa", "orgCty", "acqSrc", "acqDate",
                          "mlsStat", "available", "historic", "storage",
                          "sampStat", "duplSite", "createdBy", "createdDate",
                          "lastModifiedBy", "lastModifiedDate"),
               colClasses = rep("character", 21))
  setkey(cor, genesysId)
  # Import coll.csv
  col <- fread(input = list.files(tempdir(), pattern = "coll.csv",
                                  full.names = T),
               header = TRUE, stringsAsFactors = FALSE,
               select = c("genesysId", "collMissId", "collNumb", "collDate",
                          "collSrc", "collSite", "collCode", "collName",
                          "collInstAddress", "version"),
               colClasses = structure(rep("character", 10),
                                      .Names = c("genesysId", "collMissId",
                                                 "collNumb", "collDate",
                                                 "collSrc", "collSite",
                                                 "collCode", "collName",
                                                 "collInstAddress", "version")))
  setkey(col, genesysId)
  # Import and print the readme
  con <- unz(zip.genesys, "README.txt")
  rdm <- readLines(con)
  on.exit(close(con))
  if (readme) {
    cat(rdm, sep = "\n")
  }
  # Merge all csv files
  m <- merge(cor, col, by = "genesysId", all = TRUE)
  m <- merge(m, nam, by = "genesysId", all = TRUE)
  if (dim(geo)[1] != 0) {
    m <- merge(m, geo, by = "genesysId", all = TRUE,
               suffixes = c("_COLL", "_GEO"))
  }
  # Remove space from acc name fields 2
  if (scrub.names.space) {
    ts <-  c("acceNumb", "collNumb")
    m[, (ts) := lapply(.SD, function(x) gsub(pattern = "[[:space:]]",
                       replacement = "", x)), .SDcols = ts]
  }
  m <- as.data.frame(m)
  attr(m, "readme") <- rdm
  return(m)
  }

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/

#' Visualize the probable duplicate sets retrieved in a \code{ProbDup} object
#' 
#' \code{ViewProbDup} plots summary visualizations of accessions within the
#' probable duplicate sets retrieved in a \code{ProbDup} object according to a 
#' grouping factor field(column) in the original database(s).
#' 
#' @param pdup An object of class \code{ProbDup}.
#' @param db1 A data frame of the PGR passport database.
#' @param db2 A data frame of the PGR passport database. Required when 
#'   \code{pdup} was created using more than one KWIC Index.
#' @param factor.db1 The \code{db1} column to be considered for grouping the 
#'   accessions. Should be of class character or factor.
#' @param factor.db2 The \code{db2} column to be considered for grouping the 
#'   accessions. Should be of class character or factor. retrieved.
#' @param max.count The maximum count of probable duplicate sets whose 
#'   information is to be plotted (see \strong{Note}).
#' @param select A character vector of factor names in \code{factor.db1} and/or 
#'   \code{factor.db2} to be considered for grouping accessions (see 
#'   \strong{Note}).
#' @param order The order of the type of sets retrieved in the plot. The default
#'   is \code{"type"} (see \strong{Details}).
#' @param main The title of the plot.
#'   
#' @return A list containing the following objects: \tabular{ll}{ 
#'   \code{Summary1} \tab The summary \code{data.frame} of number of accessions 
#'   per factor level. \cr \code{Summary2} \tab The summary \code{data.frame} of
#'   number of accessions and sets per each type of sets classified according to
#'   factor levels. \cr \code{SummaryGrob} \tab A grid graphical object (Grob) 
#'   of the summary visualization plot. Can be plotted using the \code{grid.arrange} function \cr }
#'   
#' @note When any primary ID/key records in the fuzzy, phonetic or semantic 
#'   duplicate sets are found to be missing from the original databases 
#'   \code{db1} and \code{db2}, then they are ignored and only the matching 
#'   records are considered for visualization.
#'   
#'   This may be due to data standardization of the primary ID/key field using 
#'   the function \code{\link[PGRdup]{DataClean}} before creation of the KWIC 
#'   index and subsequent identification of probable duplicate sets. In such a 
#'   case, it is recommended to use an identical data standardization operation 
#'   on the databases \code{db1} and \code{db2} before running this function. 
#'   For summary and visualization of the set information in the object of class
#'   \code{ProbDup} by \code{ViewProbDup}, the disjoint of the retrieved sets 
#'   are made use of, as they are more meaningful than the raw sets retrieved. 
#'   So it is recommended that the disjoint of sets obtained using the 
#'   \code{DisProbDup} be used as the input \code{pdup}.
#'   
#'   All the accession records in sets with count > \code{max.count} will be 
#'   considered as being unique.
#'   
#'   The factor levels in the \code{factor.db1} and/or \code{factor.db2} columns
#'   corresponding to those mentioned in \code{select} argument alone will be 
#'   considered for visualization. All other factor levels will be grouped 
#'   together to a single level named "Others".
#'   
#'   The argument \code{order} can be used to specify the order in which the 
#'   type of sets retrieved are to be plotted in the visualization. The default 
#'   \code{"type"} will order according to the kind of sets, \code{"sets"} will 
#'   order according to the number of sets in each kind and \code{"acc"} will 
#'   order according to the number of accessions in each kind.
#'   
#'   The individual plots are made using \code{\link[ggplot2]{ggplot}} and then 
#'   grouped together using \code{\link[gridExtra]{gridExtra-package}}.
#'   
#' @examples
#' 
#' \dontshow{
#' threads_dt <- data.table::getDTthreads()
#' threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
#' data.table::setDTthreads(2)
#' 
#' data.table::setDTthreads(2)
#' Sys.setenv(`OMP_THREAD_LIMIT` = 2)
#' }
#' 
#' \dontrun{
#' 
#' # Method "b and c"
#' #=================
#' 
#' # Load PGR passport databases
#' GN1 <- GN1000[!grepl("^ICG", GN1000$DonorID), ]
#' GN1$DonorID <- NULL
#' GN2 <- GN1000[grepl("^ICG", GN1000$DonorID), ]
#' GN2 <- GN2[!grepl("S", GN2$DonorID), ]
#' GN2$NationalID <- NULL
#' 
#' GN1$SourceCountry <- toupper(GN1$SourceCountry)
#' GN2$SourceCountry <- toupper(GN2$SourceCountry)
#' 
#' GN1$SourceCountry <- gsub("UNITED STATES OF AMERICA", "USA", GN1$SourceCountry)
#' GN2$SourceCountry <- gsub("UNITED STATES OF AMERICA", "USA", GN2$SourceCountry)
#' 
#' # Specify as a vector the database fields to be used
#' GN1fields <- c("NationalID", "CollNo", "OtherID1", "OtherID2")
#' GN2fields <- c("DonorID", "CollNo", "OtherID1", "OtherID2")
#' 
#' # Clean the data
#' GN1[GN1fields] <- lapply(GN1[GN1fields], function(x) DataClean(x))
#' GN2[GN2fields] <- lapply(GN2[GN2fields], function(x) DataClean(x))
#' y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
#'            c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
#'            c("Mota", "Company"))
#' y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
#' y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
#'         "Bunch", "Peanut")
#' GN1[GN1fields] <- lapply(GN1[GN1fields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
#' GN1[GN1fields] <- lapply(GN1[GN1fields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
#' GN1[GN1fields] <- lapply(GN1[GN1fields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))
#' GN2[GN2fields] <- lapply(GN2[GN2fields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
#' GN2[GN2fields] <- lapply(GN2[GN2fields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
#' GN2[GN2fields] <- lapply(GN2[GN2fields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))
#' 
#' # Remove duplicated DonorID records in GN2
#' GN2 <- GN2[!duplicated(GN2$DonorID), ]
#' 
#' # Generate KWIC index
#' GN1KWIC <- KWIC(GN1, GN1fields)
#' GN2KWIC <- KWIC(GN2, GN2fields)
#' 
#' # Specify the exceptions as a vector
#' exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE",
#'           "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT",
#'           "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE",
#'           "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R",
#'           "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE",
#'           "U", "VALENCIA", "VIRGINIA", "WHITE")
#' 
#' # Specify the synsets as a list
#' syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))
#' 
#' GNdupc <- ProbDup(kwic1 = GN1KWIC, kwic2 = GN2KWIC, method = "c",
#'                   excep = exep, fuzzy = TRUE, phonetic = TRUE,
#'                   encoding = "primary", semantic = TRUE, syn = syn)
#' 
#' GNdupcView <- ViewProbDup(GNdupc, GN1, GN2, "SourceCountry", "SourceCountry",
#'                          max.count = 30, select = c("INDIA", "USA"), order = "type",
#'                          main = "Groundnut Probable Duplicates")
#'
#'library(gridExtra)                                                    
#'grid.arrange(GNdupcView$SummaryGrob)                          
#' 
#' }   
#' 
#' \dontshow{
#' data.table::setDTthreads(threads_dt)
#' Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
#' }
#'     
#' @seealso \code{\link[PGRdup]{ProbDup}}, \code{\link[PGRdup]{DisProbDup}}, 
#'   \code{\link[PGRdup]{DataClean}}, \code{\link[ggplot2]{ggplot}}, 
#'   \code{\link[gridExtra]{gridExtra-package}}
#'   
#' @import data.table
#' @importFrom ggplot2 ggplot
#' @importFrom ggplot2 theme
#' @importFrom ggplot2 element_rect
#' @importFrom ggplot2 element_blank
#' @importFrom ggplot2 element_line
#' @importFrom ggplot2 element_text
#' @importFrom ggplot2 geom_bar
#' @importFrom ggplot2 aes
#' @importFrom ggplot2 geom_rect
#' @importFrom ggplot2 ylab
#' @importFrom ggplot2 guides
#' @importFrom ggplot2 geom_point
#' @importFrom ggplot2 margin
#' @importFrom ggplot2 geom_tile
#' @importFrom ggplot2 geom_line
#' @importFrom ggplot2 guide_legend
#' @importFrom ggplot2 xlab
#' @importFrom ggplot2 ggplotGrob
#' @importFrom ggplot2 coord_flip
#' @importFrom ggplot2 scale_y_reverse
#' @importFrom ggplot2 coord_polar
#' @importFrom ggplot2 xlim
#' @importFrom ggplot2 geom_text
#' @importFrom ggplot2 labs
#' @importFrom grid textGrob
#' @importFrom grid unit
#' @importFrom gridExtra arrangeGrob
#' @importFrom methods is
#' @importFrom utils head
#' @export

ViewProbDup <- function(pdup, db1, db2 = NULL,
                             factor.db1, factor.db2 = NULL,
                             max.count = 30, select, order = "type",
                             main = NULL) {
  # Preliminary Checks
  if (!is(pdup, "ProbDup")) {
    stop("'pdup' is not of class ProbDup")
  }
  method <- attributes(pdup)$method
  fields <- attributes(pdup)$fields
  if (method == "c" | method == "b") {
    if (is.null(db2)) {
      stop(paste("argument 'db2' is missing, with no default.",
                 "\nSecond database is to be specified as method ",
                 method, " was used to generate 'pdup'", sep = ""))
    }
  }
  if (is.element(FALSE, fields[[1]] %in% colnames(db1))) {
    # Check if fields are present in db1 and stop if not
    stop("One or more kwic1 fields are missing in 'db1'")
  }
  if (!is.null(factor.db1) && !is.vector(factor.db1, mode = "character")) {
    stop("'factor.db1' is not a character vector")
  }
  if (is.element(FALSE, factor.db1 %in% colnames(db1))) {
    # Check if factor fields are present in db1 and stop if not
    stop("The 'factor.db1' field is missing in 'db1'.")
  }
  if (!(is.character(db1[, factor.db1]) || is.factor(db1[, factor.db1]))) {
    # check if factor column vector is of type character/vector
    stop("The 'factor.db1' column is not of type character or factor")
  }
  if (FALSE %in% (select %in% db1[, factor.db1])) {
    Pt <- data.frame(select, check = select %in% db1[, factor.db1], stringsAsFactors = F)
    if (!(TRUE %in% Pt$check)) {
      stop("None of the factor(s) specified in the 'select' argument are present in the 'factor.db1' column")
    }
    warning(paste("The following selected factor(s) is/are missing from 'factor.db1' column",
               paste(Pt[Pt$check == FALSE, 1], collapse = ", "), sep = "\n"))
  }
  fields[[1]] <- union(fields[[1]], factor.db1)
  #setDT(db1)
  db1 <- as.data.table(db1)
  if (is.factor(db1[, get(factor.db1)])) {
    db1[, (factor.db1) := as.character(get(factor.db1))]
  }
  # Convert whitespace to NA
  db1[grepl("^[[:space:]]*$", get(factor.db1)), (factor.db1) := NA]
  if (!identical(setdiff(colnames(db1), fields[[1]]), character(0))) {
    db1[, setdiff(colnames(db1), fields[[1]]) := NULL]
  }
  db1[, K1_PRIM_ID := get(fields[[1]][1])]
  setcolorder(db1, neworder = union("K1_PRIM_ID",
                                    setdiff(colnames(db1), "K1_PRIM_ID")))
  db1[, K1_PRIM_ID := as.character(K1_PRIM_ID)]
  setkey(db1, "K1_PRIM_ID")
  if (method == "c" | method == "b") {
    if (is.element(FALSE, fields[[2]] %in% colnames(db2)) == TRUE) {
      # Check if fields are present in db2 and stop if not
      stop('One or more kwic2 fields are missing in "db2"')
    }
    if (!is.null(factor.db2) && !is.vector(factor.db2, mode = "character")) {
      stop('"factor.db2" is not a character vector')
    }
    if (is.element(FALSE, factor.db2 %in% colnames(db2)) == TRUE) {
      # Check if factor fields are present in db2 and stop if not
      stop('The "factor.db2" field is missing in "db2"')
    }
    if (!(is.character(db2[, factor.db2]) || is.factor(db2[, factor.db2]))) {
      # check if factor column vector is of type character/vector
      stop("The 'factor.db2' column is not of type character or factor")
    }
    if (FALSE %in% (select %in% db2[, factor.db2])) {
      Pt <- data.frame(select, check = select %in% db2[, factor.db2],
                       stringsAsFactors = F)
      if (!(TRUE %in% Pt$check)) {
        stop("None of the factor(s) specified in the 'select' argument are present in the 'factor.db1' column")
      }
      warning(paste("The following selected factor(s) is/are missing from 'factor.db2' column",
                 paste(Pt[Pt$check == FALSE, 1], collapse = ", "), sep = "\n"))
    }

    fields[[2]] <- union(fields[[2]], factor.db2)
    #setDT(db2)
    db2 <- as.data.table(db2)
    if (is.factor(db2[, get(factor.db2)])) {
      db2[, (factor.db2) := as.character(get(factor.db2))]
    }
    # Convert whitespace to NA
    db2[grepl("^[[:space:]]*$", get(factor.db2)), (factor.db2) := NA]
    if (!identical(setdiff(colnames(db2), fields[[1]]), character(0))) {
      db2[, setdiff(colnames(db2), fields[[2]]) := NULL]
    }
    db2[, K2_PRIM_ID := get(fields[[2]][1])]
    setcolorder(db2, neworder = union("K2_PRIM_ID",
                                      setdiff(colnames(db2), "K2_PRIM_ID")))
    db2[, K2_PRIM_ID := as.character(K2_PRIM_ID)]
    setkey(db2, "K2_PRIM_ID")
  }
  order <- match.arg(order, c("type", "sets", "acc"), several.ok = FALSE)
  # Get disjoint sets
  if (!is.null(pdup[["DisjointDupicates"]])) {
    tryCatch(DisProbDup(pdup),
             warning = function(e) {
               message(gsub("returned", "displayed", e$message))
             })
    pdup <- suppressWarnings(DisProbDup(pdup))
  } else {
    tstr <- data.frame(names = c("F", "P", "S"),
                       `is.null` = c(is.null(pdup[[1]]), is.null(pdup[[2]]),
                                     is.null(pdup[[3]])),
                       stringsAsFactors = F)
    if (dim(tstr[tstr$is.null == FALSE, ])[1] == 1) {
      tryCatch(DisProbDup(pdup),
      warning = function(e) {
        message(gsub("returned", "displayed", e$message))
      })
      pdup <- suppressWarnings(DisProbDup(pdup))
    } else {
      pdup <- DisProbDup(pdup, combine = tstr[tstr$is.null == FALSE, ]$names)
    }
  }

  # Get Review data.frame
  tryCatch(ReviewProbDup(pdup, db1, db2,
                         extra.db1 = factor.db1, extra.db2 = factor.db2,
                         max.count, insert.blanks = FALSE),
           warning = function(e) {
             message(gsub("merged", "used", e$message))
           })
  sets <- suppressWarnings(ReviewProbDup(pdup, db1, db2,
                                         extra.db1 = factor.db1,
                                         extra.db2 = factor.db2,
                                         max.count, insert.blanks = FALSE))

  # Prepare the sets
  if (is.null(factor.db2)) {
    cols <- c("SET_NO", "PRIM_ID",
              paste("K1X", factor.db1, sep = "_"))
  } else {
    cols <- c("SET_NO", "PRIM_ID",
              paste("K1X", factor.db1, sep = "_"),
              paste("K2X", factor.db2, sep = "_"))
  }
  sets <- as.data.table(sets[, cols])
  #setDT(sets)
  if (is.null(factor.db2)) {
    sets[, FACTOR := get(cols[3])]
  } else {
    sets[, FACTOR := ifelse(is.na(get(cols[3])), get(cols[4]), get(cols[3]))]
  }
  sets[!FACTOR %in% select, FACTOR := "Others"]
  sets[, FACTOR := as.factor(FACTOR)]
  sets <- sets[order(SET_NO), as.list(table(FACTOR)),
               by = SET_NO] # Get summary table
  factorlevels <- setdiff(colnames(sets), "SET_NO")
  Nfactorlevels <- paste0("No.Acc. (", factorlevels, ")")
  setnames(sets, old = factorlevels, new = Nfactorlevels)
  sets[, (factorlevels) := data.frame(matrix(ifelse(sets[, Nfactorlevels, with = FALSE] != 0, 1, 0),
                                             ncol = length(Nfactorlevels)))]
  sets[, WITHIN := ifelse(rowSums(sets[, factorlevels, with = FALSE]) == 1,
                          1, 0)]
  sets[, `NO.ACC` := rowSums(sets[, Nfactorlevels, with = FALSE])]
  # Prepare orphans
  dups <- ParseProbDup(pdup, insert.blanks = F, max.count)$PRIM_ID
  orps <- db1[!K1_PRIM_ID %in% dups, .SD, .SDcols = c("K1_PRIM_ID",
                                                      factor.db1)]
  setnames(orps, colnames(orps), c("PRIM_ID", "FACTOR"))
  if (!is.null(factor.db2)) {
    orp2 <- db2[!K2_PRIM_ID %in% dups, .SD, .SDcols = c("K2_PRIM_ID",
                                                        factor.db2)]
    setnames(orp2, colnames(orp2), c("PRIM_ID", "FACTOR"))
    orps <- rbind(orps, orp2)
    rm(orp2)
  }
  ndupsets <- sum(unlist(sapply(pdup, function(x) dim(x)[1]))) # Add set numbers
  orps[, SET_NO := seq_len(nrow(orps)) + ndupsets]
  orps[!FACTOR %in% select, FACTOR := "Others"]
  orps[, FACTOR := as.factor(FACTOR)]
  orps <- orps[order(SET_NO), as.list(table(FACTOR)),
               by = SET_NO] # Get summary table
  factorlevels <- setdiff(colnames(orps), "SET_NO")
  Nfactorlevels <- paste0("No.Acc. (", factorlevels, ")")
  orps[, WITHIN := 0][, `NO.ACC` := 1]
  orps[, (Nfactorlevels) := orps[, factorlevels, with = FALSE]]
  # Final summary = sets + orphans
  bby <- append(factorlevels, "WITHIN")
  setcolorder(sets, colnames(orps))
  sets <- rbind(sets, orps)
  setkeyv(sets, append(factorlevels, "WITHIN"))
  summ1 <- sets[, lapply(.SD, sum), .SDcols = Nfactorlevels, by = bby]
  summ2 <- sets[, list(NO.ACC = sum(NO.ACC), NO.SETS = .N), by = bby]
  summ <- merge(summ1, summ2)
  summ <- summ[, lapply(.SD, as.numeric)]
  summ[, KIND := as.factor(rownames(summ))]
  rm(summ1, summ2)
  # Specify the shape for mat plot
  summ[, singles := ifelse(rowSums(summ[, factorlevels, with = FALSE]) == 1,
                           1, 0)]
  summ[, shape :=  singles + WITHIN]
  # Reorder summary table
  if (order == "type") {
    tord <- as.numeric(append(rev(as.character(summ[summ$shape == 0]$KIND)),
                              append(as.character(summ[summ$shape == 2]$KIND),
                                     as.character(summ[summ$shape == 1]$KIND))))
    summ[, KIND := factor(KIND, levels = summ[tord, KIND])]
  }
  if (order == "acc") {
    summ[, KIND := factor(KIND, levels = summ[order(NO.ACC), KIND])]
  }
  if (order == "sets") {
    summ[, KIND := factor(KIND,
                          levels = summ[order(NO.SETS, na.last = FALSE), KIND])]
  }
  # Set blanks in No.sets plot
  summ[which(rowSums(summ[, factorlevels, with = F]) == 1),
       NO.SETS := ifelse(WITHIN == 0, yes = NA, NO.SETS)]
  # Create the data.frame to specify shading
  shading <- data.frame(min = seq(from = 0.5,
                                  to = max(as.numeric(summ$KIND)),
                                  by = 1),
                        max = seq(from = 1.5,
                                  to = max(as.numeric(summ$KIND)) + 0.5,
                                  by = 1),
                        col = 0)
  shading[which(seq(seq_len(nrow(shading))) %% 2 == 1), ]$col <- 1
  shading$col <- ifelse(shading$col, "gray83", "gray93")
  #Stacked bar
  sbar <- melt(summ, id.vars = "KIND", measure.vars = Nfactorlevels)
  sbarg <- ggplot() +
    theme(panel.background = element_rect(fill = "transparent"),
          axis.title.x = element_blank(),
          axis.ticks.x = element_blank(),
          axis.text.x = element_blank(), #<>#
          axis.ticks.y = element_line(colour = "black"),
          axis.line = element_line(colour = NA),
          axis.line.y = element_line(colour = "black"),
          axis.text.y = element_text(colour = "black"),
          panel.grid.major.x = element_line(colour = "gray83"),
          panel.grid.minor.x = element_blank(),
		  panel.grid.major.y = element_line(colour = "gray33",
                                            linetype = "dotted"),
          plot.margin = unit(c(0, 0.5, 0.1, 0.5), "cm")) +
    geom_bar(sbar, mapping = aes(x = KIND, y = value, fill = variable),
             stat = "identity") +
    geom_rect(data = shading,
              aes(xmin = min, xmax = max, ymin = -Inf, ymax = Inf, alpha = 0.1),
              fill = shading$col, colour = "white") +
    geom_bar(sbar, mapping = aes(x = KIND, y = value, fill = variable),
             stat = "identity") +
    ylab("No. of accessions") +
    #coord_flip() +
    guides(fill = FALSE, colour = FALSE, alpha = FALSE)
  #dot plot
  dotp <- melt(summ, id.vars = "KIND", measure.vars = "NO.SETS")
  dotpg <- ggplot() +
    theme(panel.background = element_rect(fill = "transparent"),
          axis.title.x = element_blank(),
          axis.ticks.x = element_blank(),
          axis.text.x = element_blank(), #<>#
          axis.line = element_line(colour = NA),
          axis.line.y = element_line(colour = "black"),
          axis.ticks.y = element_line(colour = "black"),
          axis.text.y = element_text(colour = "black"),
          panel.grid.major.x = element_line(colour = "gray83"),
          panel.grid.minor.x = element_blank(),
          panel.grid.major.y = element_line(colour = "gray33",
                                            linetype = "dotted"),
          plot.margin = unit(c(1, 0.5, 0.3, 0.5), "cm")) +
    geom_point(dotp, mapping = aes(x = KIND, y = value), colour = "gray23",
               size = 3, pch = 18, na.rm = TRUE) +
    geom_rect(data = shading,
              aes(xmin = min, xmax = max, ymin = -Inf, ymax = Inf, alpha = 0.1),
              fill = shading$col, colour = "white") +
    geom_point(dotp, mapping = aes(x = KIND, y = value), colour = "gray23",
               size = 3, pch = 18, fill = "gray23", na.rm = TRUE) +
    ylab("No. of sets") +
    #coord_flip() +
    #scale_y_reverse() +
    guides(fill = FALSE, colour = FALSE, alpha = FALSE)
  # matrix
  mat <- melt(summ, id.vars = c("KIND", "WITHIN", "shape", "singles"),
              measure.vars = factorlevels)
  mat[, shape := ifelse(shape == 0, "Between",
                        ifelse(shape == 1, "Unique", "Within"))]
  mat[, shape2 := ifelse(value == 1, shape, NA)]
  mat[, KIND2 := as.numeric(ifelse(is.na(shape2), NA, KIND))]
  mat[is.na(mat$KIND2), KIND2 := seq(from = max(mat$KIND2, na.rm = T) + 1,
                                     length.out = length(mat[is.na(KIND2),]$KIND2))]
  matg <- ggplot() +
    theme(panel.background = element_rect(fill = "transparent"),
          axis.title.x =  element_text(colour = "black",
                                       margin = margin(20, 0, 20, 0)),
          axis.ticks.x = element_blank(),
          axis.text.x = element_blank(), #<>#
          axis.line = element_line(colour = NA),
          axis.line.y = element_line(colour = "black"),
          axis.ticks.y = element_line(colour = "black"),
          axis.text.y = element_text(colour = "black"),
          axis.title.y = element_blank(), #<>#
          #legend.position = "none", #<>#
          plot.margin = unit(c(0, 0.5, 0.5, 0.5), "cm"),
          #legend.title = element_blank(),
          panel.grid = element_blank()) +
    geom_tile(data = mat, mapping = aes(y = variable, x = as.factor(KIND)),
              colour = "white", fill = NA) +
    geom_rect(data = shading,
              aes(xmin = min, xmax = max,
                  ymin = 0.5, ymax = length(factorlevels) + 0.5, alpha = 0.1),
              fill = shading$col, colour = "white") +
    geom_tile(data = mat, mapping = aes(y = variable, x = as.factor(KIND)),
              colour = "white", fill = NA) +
    geom_point(data = mat, aes(y = variable, x = KIND, shape = mat$shape),
               colour = "gray73", alpha = 0.6, size = 5, na.rm = TRUE) +
    geom_point(data = mat, aes(y = variable, x = KIND, shape = mat$shape2),
               colour = "gray23",  size = 5,  na.rm = TRUE) +
    geom_line(data = mat, aes(y = variable, x = KIND,
                              group = KIND2),
              colour = "gray23",  linewidth = 1,  na.rm = TRUE) +
    #coord_flip() +
    guides(shape = guide_legend(title = "Accession\nType"),
           fill = FALSE, colour = FALSE, alpha = FALSE) +
    xlab("Type of sets retrieved")
  # Extract legend1
  grobs <- ggplotGrob(matg)$grobs
  legend1 <- grobs[[which(sapply(grobs, function(x) x$name) == "guide-box")]]
  rm(grobs)
  # Accession summary
  nacc <- colSums(summ[, Nfactorlevels, with = FALSE])
  nacc <- data.frame(nacc)
  nacc$factor <- rownames(nacc)
  nacc$factor <- factorlevels
  naccg <- ggplot(nacc, aes(x = factor, y = nacc, fill = factor)) +
    theme(panel.background = element_rect(fill = "transparent"),
          axis.ticks.y = element_blank(),
          axis.text.y = element_blank(),
          axis.line = element_line(colour = "black"),
          axis.line.y = element_blank(),
          axis.title.y = element_blank(),
          axis.ticks.x = element_line(colour = "black"),
          axis.text.x = element_text(colour = "black"),
          axis.title.x = element_text(colour = "black",
                                      margin = margin(20, 0, 20, 0)),
          panel.grid.major.y = element_line(colour = "gray93"),
          panel.grid.minor.y = element_blank(),
          plot.margin = unit(c(0, 0, 0.5, 0), "cm")) +
    geom_bar(stat = "identity", position = "identity") +
    ylab("No. of accessions") +
    coord_flip() +
    guides(fill  = guide_legend(title = "Accession\nGroup"),
           colour = FALSE, alpha = FALSE) +
    scale_y_reverse()
  # Extract legend1
  grobs <- ggplotGrob(naccg)$grobs
  legend2 <- grobs[[which(sapply(grobs, function(x) x$name) == "guide-box")]]
  rm(grobs)
  # Summary donut
  smry <- data.frame(Total = c("Sets", "Accessions", "  Redundant", "  Unique"),
                     Count = c(sum(dotp[!is.na(dotp$value), ]$value),
                               sum(nacc$nacc),
                               sum(sbar[sbar$KIND %in% dotp[!is.na(dotp$value), ]$KIND, ]$value),
                               sum(sbar[!sbar$KIND %in% dotp[!is.na(dotp$value), ]$KIND, ]$value)))
  smry1 <- smry[3:4, ]
  smry1$fraction <- smry1$Count / sum(smry1$Count)
  smry1$ymax <- cumsum(smry1$fraction)
  smry1$ymin <- c(0, head(smry1$ymax, n = -1))
  smry1$label <- paste(smry1$Total, "\n", round(smry1$fraction * 100, 2),
                       " %", sep = "")
  smryg <- ggplot(data = smry1, aes(fill = Total, ymax = ymax, ymin = ymin,
                                    xmax = 4, xmin = 3)) +
    geom_rect(show.legend = FALSE, fill = c("grey93", "grey73")) +
    coord_polar(theta = "y") +
    xlim(c(0, 4)) +
    theme(panel.background = element_blank(),
          panel.grid = element_blank(),
          axis.text = element_blank(),
          axis.ticks = element_blank(),
          plot.title = element_text(hjust = 0.5, size = 10)) +
    geom_text(aes(x = 3.5, y = ( (ymin + ymax) / 2), label = label)) +
    xlab("") +
    ylab("") +
    labs(title = paste("\n\nTotal no. of sets = ", smry[1, ]$Count, "\n",
                       "Total no. of accessions  = ", smry[2, ]$Count,
                       sep = ""))
  # Get and modify grobs
  naccg1 <- ggplotGrob(naccg + theme(legend.position = "none"))
  dotpg1 <- ggplotGrob(dotpg)
  matg1 <- ggplotGrob(matg + theme(legend.position = "none"))
  sbarg1 <- ggplotGrob(sbarg)
  smryg1 <- ggplotGrob(smryg)
  smryg1$layout$clip[smryg1$layout$name == "panel"] <- "off"
  tempgrob <- cbind(naccg1, matg1, size = "last")
  naccg1$heights <- tempgrob$heights
  matg1$heights <- tempgrob$heights
  dotpg1$widths <- matg1$widths
  sbarg1$widths <- matg1$widths
  legend <-  arrangeGrob(legend1, legend2, ncol = 2)
  SummaryGrob <- arrangeGrob(smryg1, dotpg1,
                             legend, sbarg1,
                             naccg1, matg1,
                             ncol = 2, widths = c(0.5, 1.5),
                             heights = c(0.8, 1, 0.6),
                             top = textGrob(main, just = "top"))
  Summary1 <- data.frame("Factor" = nacc$factor,
                         "No. of accessions" = nacc$nacc,
                        check.names = FALSE, stringsAsFactors = FALSE)
  Summary2 <- as.data.table(summ)
  Summary2[, c("singles", "shape") := NULL]
  setcolorder(Summary2, union("KIND", setdiff(colnames(Summary2), "KIND")))
  setnames(Summary2, old = "KIND", "Type of sets retrieved")
  setnames(Summary2, old = "NO.ACC", "Total no. of accessions")
  setnames(Summary2, old = "NO.SETS", "Total no. of sets")
  setDF(Summary2)
  out <- list(Summary1 = Summary1, Summary2 = Summary2,
              SummaryGrob = SummaryGrob)
  rm(Summary1, Summary2, SummaryGrob)
  return(out)
}

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/



#' Validate if a data frame column confirms to primary key/ID constraints
#' 
#' \code{ValidatePrimKey} checks if a column in a data frame confirms to the 
#' primary key/ID constraints of absence of duplicates and NULL values. Aberrant
#' records if encountered are returned in the output list.
#' 
#' The function checks whether a field(column) in a data frame of PGR passport 
#' database confirms to the primary key/ID constraints of absence of duplicates 
#' and NULL values. If records with nonconforming values in the column are 
#' encountered, they are returned in the output list for rectification.
#' 
#' If multiple fields(columns) are given as a character vector in 
#' \code{prim.key} field, only the first element will be considered as the 
#' primary key/ID field(column).
#' 
#' Cleaning of the data in the input field(column) using the 
#' \code{\link[PGRdup]{DataClean}} function with appropriate arguments is 
#' suggested before running this function.
#' 
#' It is recommended to run this function and rectify aberrant records in a PGR 
#' passport database before creating a KWIC index using the 
#' \code{\link[PGRdup]{KWIC}} function.
#' 
#' @param x A data frame.
#' @param prim.key A character vector indicating the name of the data frame 
#'   column to be validated for primary key/ID constraints (see 
#'   \strong{Details}).
#' @return A list with containing the following components: \tabular{ll}{ 
#'   \code{message1} \tab Indicates whether duplicated values were encountered 
#'   in \code{prim.key} field(column) of data frame \code{x} or not. \cr 
#'   \code{Duplicates} \tab A data frame of the records with duplicated prim.key
#'   values if they were encountered. \cr \code{message2} \tab Indicates whether
#'   NULL values were encountered in \code{prim.key} field(column) of data frame
#'   \code{x} or not. \cr \code{NullRecords} \tab A data frame of the records 
#'   with NULL prim.key values if they were encountered. \cr }
#' @seealso \code{\link[PGRdup]{DataClean}}, \code{\link[PGRdup]{KWIC}}
#' @examples
#' GN <- GN1000
#' ValidatePrimKey(x=GN, prim.key="NationalID")
#' \dontrun{
#' # Show error in case of duplicates and NULL values 
#' # in the primary key/ID field "NationalID"
#' GN[1001:1005,] <- GN[1:5,]
#' GN[1001,3] <- ""
#' ValidatePrimKey(x=GN, prim.key="NationalID")}
#' @export
ValidatePrimKey <- function(x, prim.key) {
  if (is.data.frame(x) == FALSE) {
    # Check if x is a data frame and stop if not
    stop("x is not a data frame")
  }
  if (is.vector(prim.key) == FALSE) {
    # Check if prim.key is a vector or not
    stop("prim.key is not a vector")
  }
  if (length(prim.key) > 1) {
    # Check if only one field is given as input and use first element if not
    prim.key <- prim.key[1]
    warning("prim.key length >1; Only the first element is used")
  }
  if (is.element("FALSE", prim.key %in% colnames(x)) == TRUE) {
    # Check if prim.key field is present in x and stop if not
    stop("prim.key field missing in x")
  }
  Result <- list(message1 = NULL, Duplicates = NULL, message2 = NULL,
                 NullRecords = NULL)
  # Convert NAs to empty strings
  x[prim.key][is.na(x[, prim.key]), ] <- ""
  if (is.element("TRUE", duplicated(x[prim.key]))) {
    # Check if duplicated records are there in prim.key
    Result$message1 <- "ERROR: Duplicated records found in prim.key field"
    message(Result$message1)
    x$primdup <- duplicated(x[prim.key]) | duplicated(x[prim.key],
                                                      fromLast = TRUE)
    Result$Duplicates <- subset(x, primdup == TRUE)
    Result$Duplicates[, "primdup"] <- NULL
    Result$Duplicates <- Result$Duplicates[order(Result$Duplicates[, prim.key]), ]
  } else {
    Result$message1 <- "OK: No duplicated records found in prim.key field"
    message(Result$message1)
  }
  if (is.element("", x[, prim.key]) == TRUE) {
    # Check if empty characters are present in prim.key field
    Result$message2 <- "ERROR: NULL records found in prim.key field"
    message(Result$message2)
    Result$NullRecords <- subset(x, get(prim.key) == "")
  } else {
    Result$message2 <- "OK: No NULL records found in prim.key field"
    message(Result$message2)
  }
  return(Result)
}

### This file is part of 'PGRdup' package for R.

### Copyright (C) 2014-2026, ICAR-NBPGR.
#
# PGRdup is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# PGRdup is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  https://www.r-project.org/Licenses/



#' Add probable duplicate sets fields to the PGR passport database
#' 
#' \code{AddProbDup} adds the fuzzy, phonetic and semantic probable duplicates 
#' sets data fields from an object of class \code{ProbDup} to the original PGR 
#' passport database.
#' 
#' This function helps to add information associated with identified fuzzy, 
#' phonetic and semantic probable duplicate sets using the 
#' \code{\link[PGRdup]{ProbDup}} function to the original PGR passport database.
#' Associated data fields such as \code{SET_NO}, \code{ID} and \code{IDKW} are 
#' added based on the \code{PRIM_ID} field(column).
#' 
#' In case more than one KWIC index was used to generate the object of class 
#' \code{ProbDup}, the argument \code{addto} can be used to specify to which 
#' database the data fields are to be added. The default \code{"I"} indicates 
#' the database from which the first KWIC index was created and \code{"II"} 
#' indicates the database from which the second index was created.
#' 
#' @param pdup An object of class \code{ProbDup}.
#' @param db A data frame of the PGR passport database.
#' @param addto Either \code{"I"} or \code{"II"} indicating the database to 
#'   which the data.fields are to be added (see \strong{Details}).
#' @param max.count The maximum count of probable duplicate sets whose 
#'   information is to be retrieved.
#' @return A data frame of the PGR passport database with the probable duplicate
#'   sets fields added.
#' @examples
#' 
#' \dontshow{
#' threads_dt <- data.table::getDTthreads()
#' threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
#' data.table::setDTthreads(2)
#' 
#' data.table::setDTthreads(2)
#' Sys.setenv(`OMP_THREAD_LIMIT` = 2)
#' }
#' 
#' \dontrun{
#' 
#' #' # Load PGR passport database
#' GN <- GN1000
#'
#' # Specify as a vector the database fields to be used
#' GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")
#'
#' # Clean the data
#' GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
#' y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
#' c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
#' c("Mota", "Company"))
#' y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
#' y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
#'         "Bunch", "Peanut")
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
#' GN[GNfields] <- lapply(GN[GNfields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))
#' 
#' # Generate KWIC index
#' GNKWIC <- KWIC(GN, GNfields)
#'
#' # Specify the exceptions as a vector
#' exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE", 
#'          "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT", 
#'          "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE", 
#'          "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R", 
#'          "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE", 
#'          "U", "VALENCIA", "VIRGINIA", "WHITE")
#'           
#' # Specify the synsets as a list
#' syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))
#'
#' # Fetch probable duplicate sets
#' GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, fuzzy = TRUE,
#'                  phonetic = TRUE, encoding = "primary", 
#'                  semantic = TRUE, syn = syn)
#' 
#' # Add the duplicates sets to the original database                 
#' GNwithdup <-  AddProbDup(pdup = GNdup, db = GN1000, addto = "I")                  
#' 
#' }
#' 
#' \dontshow{
#' data.table::setDTthreads(threads_dt)
#' Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
#' }
#' 
#' @seealso \code{\link[PGRdup]{DataClean}}, \code{\link[PGRdup]{KWIC}},
#'   \code{\link[PGRdup]{ProbDup}}
#' @note When any primary ID/key records in the fuzzy, phonetic or semantic 
#'   duplicate sets are found to be missing from the original database 
#'   \code{db}, then they are ignored and only the matching records are 
#'   considered for adding the information with a warning.
#'   
#'   This may be due to data standardization of the primary ID/key field using 
#'   the function \code{\link[PGRdup]{DataClean}} before creation of the KWIC 
#'   index and subsequent identification of probable duplicate sets. In such a 
#'   case, it is recommended to use an identical data standardization operation 
#'   on the database \code{db} before running this function.
#' @import data.table
#' @importFrom methods is
#' @export
AddProbDup <- function(pdup, db, addto = c("I", "II"), max.count = 30) {
  if (!is(pdup, "ProbDup")) {
    stop('"pdup" is not of class ProbDup')
  }
  addto <- match.arg(addto, several.ok = FALSE)
  method <- attributes(pdup)$method
  if (addto == "I") {
    p <- "[K1]"
    fields <- attributes(pdup)$fields[[1]]
  }
  if (addto == "II") {
    if (is.null(attributes(pdup)$fields[[2]])) {
      stop("argument 'addto = 'II', but pdup was not generated using a second KWIC index" )
    }
    p <- "[K2]"
    fields <- attributes(pdup)$fields[[2]]
  }
  if (is.data.frame(db) == FALSE) {
    # Check if db is a data frame and stop if not
    stop('"db" is not a data frame')
  }
  if (!fields[1] %in% colnames(db)) {
    stop(paste('Column(field) matching the primary ID/key "',
               fields[1], '" is not found in "db"', sep = ""))
  }
  if (is.element("FALSE", fields %in% colnames(db)) == TRUE) {
    # Check if fields are present in x and stop if not
    warning('One or more fields used for generating "pdup" are missing in "db"')
  }
  #setDT(db)
  db <- as.data.table(db)
  setkeyv(db, fields[[1]])
  # Retrieve the sets with "[K*]" according to argument addto
  pdup <- lapply(pdup, function(x) x[grepl(p, x$ID, fixed = TRUE), ])
  #invisible(lapply(seq_along(pdup), function(i) pdup[[i]]$PRIM_ID <<- gsub(p, "", pdup[[i]]$PRIM_ID, fixed = TRUE)))
  types <- c("F", "P", "S", "D")
  types2 <- c("Fuzzy", "Phonetic", "Semantic", "Disjoint")
  N <- length(pdup)
  for (i in 1:N) {
    if (!is.null(pdup[[i]])) {
      #setDT(pdup[[i]])
      pdup[[i]] <- as.data.table(pdup[[i]])
      pdup[[i]] <- subset(pdup[[i]], COUNT <= max.count)
      pdup[[i]][, TYPE := NULL]
      pdup[[i]] <- pdup[[i]][, list(PRIM_ID = unlist(strsplit(ID, ", " ))),
                             by = list(SET_NO, IDKW, ID)]
      if (method == "b" | method == "c") {
        if (addto == "I") {
          pdup[[i]] <- pdup[[i]][grepl("\\[K1\\].*", PRIM_ID, perl = TRUE)]
        }
        if (addto == "II") {
          pdup[[i]] <- pdup[[i]][grepl("\\[K2\\].*", PRIM_ID, perl = TRUE)]
        }
      }
      pdup[[i]][, PRIM_ID := gsub(p, "", pdup[[i]]$PRIM_ID, fixed = TRUE)]
      setkey(pdup[[i]], PRIM_ID)
      # Aggregate according to ID
      pdup[[i]] <- pdup[[i]][, list(SET_NO = paste0(sort(unique(SET_NO)),
                                                    collapse = ", "),
                                    IDKW = paste0(sort(unique(unlist(strsplit(IDKW, ", ")))),
                                                  collapse = ", "),
                                    ID = paste0(sort(unique(unlist(strsplit(ID, ", ")))),
                                                collapse = ", ")), by = PRIM_ID]
      setkey(pdup[[i]], PRIM_ID)
      setcolorder(pdup[[i]], c("PRIM_ID", "SET_NO", "ID", "IDKW"))
      setnames(pdup[[i]], old = colnames(pdup[[i]]),
               new = paste(types[i], colnames(pdup[[i]]), sep = "_"))
      setnames(pdup[[i]], old = paste(types[i], "PRIM_ID", sep = "_"),
               new = fields[1])
      if (length(setdiff(pdup[[i]][, get(fields[1])],
                         db[, get(fields[1])])) != 0) {
        warning(paste("Encountered primary ID records in",
                      types2[i], "probable duplicate sets missing from 'db'.",
                      "\nOnly matching records are merged"))
      }
      db <- merge(db, pdup[[i]], all.x = TRUE)
    }
  }
  rm(pdup)
  db[is.na(db)]   <- ""
  setDF(db)
  return(db)
}

---
title: "An Introduction to `PGRdup` Package"
author: "Aravind, J.^1^, Radhamani, J.^1^, Kalyani Srinivasan^1^, Ananda Subhash, B.^2^, and Tyagi, R. K.^1^"
date: '`r Sys.Date()`'
classoption: table, twoside
output:
  pdf_document:
    fig_caption: no
    toc: no
  html_document:
    df_print: paged
    toc: yes
header-includes:
- \usepackage{fancyhdr}
- \usepackage{wrapfig}
- \pagestyle{fancy}
- \fancyhead[LE,RO]{\slshape \rightmark}
- \fancyhead[LO,RE]{An Introduction to \texttt{PGRdup} Package}
- \fancyfoot[C]{\thepage}
- \usepackage{hyperref}
- \hypersetup{colorlinks=true}
- \hypersetup{linktoc=all}
- \hypersetup{linkcolor=blue}
bibliography: bibliography.bibtex
link-citations: yes
vignette: |
  %\VignetteIndexEntry{Introduction}
  %\usepackage[utf8]{inputenc}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---
```{r, include=FALSE, eval = FALSE}
options(tinytex.verbose = TRUE)
```

```{r, echo=FALSE}
out_type <- knitr::opts_knit$get("rmarkdown.pandoc.to")

r = getOption("repos")
r["CRAN"] = "https://cran.rstudio.com/"
#r["CRAN"] = "https://cloud.r-project.org/"
#r["CRAN"] = "https://ftp.iitm.ac.in/cran/"
options(repos = r)

# Workaround for missing pandoc in CRAN OSX build machines
out_type <- ifelse(out_type == "", "latex", out_type)

# Workaround for missing pandoc in Solaris build machines
out_type <- ifelse(identical (out_type, vector(mode = "logical", length = 0)),
                   "latex", out_type)
```

```{r, echo=FALSE}
# Restrict threads for avoid elapsed threshold error in CRAN check
threads_dt <- data.table::getDTthreads()
threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")

data.table::setDTthreads(2)
Sys.setenv(`OMP_THREAD_LIMIT` = 2)
```

```{r, results='asis', echo=FALSE}
switch(out_type,
    html = {cat("<p>1. ICAR-National Bureau of Plant Genetic Resources, New Delhi, India.</p>
	
<p>2. Centre for Development of Advanced Computing, Thiruvananthapuram, Kerala, India.</p>")},
    latex = cat("\\begin{center}
1. ICAR-National Bureau of Plant Genetic Resources, New Delhi, India.

2. Centre for Development of Advanced Computing, Thiruvananthapuram, Kerala, India.

\\end{center}" )
)
```


\begin{center}
\vspace{6pt}
\hrule
\end{center}

```{r, echo = FALSE}
knitr::opts_chunk$set(comment = "",
                      fig.cap= "")
```

\tableofcontents

\pagebreak

\begin{wrapfigure}{r}{0.35\textwidth}
  \vspace{1cm}
  \begin{center}
    \includegraphics[width=0.33\textwidth]{`r system.file("extdata", "PGRdup_v2.png", package = "PGRdup")`}
  \end{center}
  \vspace{-1.5cm}
\end{wrapfigure}\leavevmode

<img src="https://raw.githubusercontent.com/aravind-j/PGRdup/master/inst/extdata/PGRdup_v2.png" align="left" alt="logo" width="173" height = "200" style = "padding: 10px; border: none; float: right;">

## Introduction

**PGRdup** is an `R` package to facilitate the search for probable/possible duplicate accessions in Plant Genetic Resources (PGR) collections using passport databases. Primarily this package implements a workflow (Fig.&nbsp;1) designed to fetch groups or sets of germplasm accessions with similar passport data particularly in fields associated with accession names within or across PGR passport databases. It offers a suite of functions for data pre-processing, creation of a searchable Key Word in Context (KWIC) index of keywords associated with accession records and the identification of probable duplicate sets by fuzzy, phonetic and semantic matching of keywords. It also has functions to enable the user to review, modify and validate the probable duplicate sets retrieved.

The goal of this document is to introduce the users to these functions and familiarise them with the workflow intended to fetch probable duplicate sets. This document assumes a basic knowledge of `R` programming language.

The functions in this package are primarily built using the `R` packages [`data.table`](https://CRAN.R-project.org/package=data.table), [`igraph`](https://CRAN.R-project.org/package=igraph), [`stringdist`](https://CRAN.R-project.org/package=stringdist) and [`stringi`](https://CRAN.R-project.org/package=stringi).

\clearpage

\begin{center}
    \includegraphics{`r system.file("extdata", "PGRdup.png", package = "PGRdup")`}
\end{center}

<img src="https://raw.githubusercontent.com/aravind-j/PGRdup/master/inst/extdata/PGRdup.png" align="center" alt="logo" width="750" height = "85" style = "border: none;">


```{r, echo=FALSE, warning=FALSE, fig.cap = NULL}
if (requireNamespace("diagram", quietly = TRUE)) {
  suppressMessages(library(diagram))
  
  elpos <- coordinates(pos = c(2, 2, 2, 4, 2, 2, 1, 3))
  
  elpos[8,1] <- elpos[5,1]
  elpos[9,1] <- 0.375
  elpos[10,1] <- elpos[6,1]
  elpos[10,2] <- 0.5
  elpos[16,1] <- elpos[16,1] - 0.05
  
  elpos[4,2] <- elpos[4,2] + 0.0020
  elpos[15,2] <- elpos[15,2] + 0.0200
  elpos[16,2] <- elpos[16,2] + 0.0250
  elpos[17,2] <- elpos[17,2] + 0.0250
  elpos[18,2] <- elpos[18,2] + 0.0250

  t <- as.data.frame(elpos)
  t[t$V1<0.5,][,1] <- t[t$V1<0.5,][,1] + 0.1
  t[t$V1>0.5,][,1] <- t[t$V1>0.5,][,1] - 0.1
  elpos <- as.matrix(t)
  colnames(elpos) <- NULL
  
  
  elpos1 <- elpos[c(3,5,8,11,13),]
  elpos2 <- elpos[c(5,7,9),]
  elpos3 <- elpos[c(7,9,11),]
  
  par(mar = c(1, 1, 1, 1))
  openplotmat()
  
  #text(elpos, lab = as.character(c(1:18)), cex = 2)
  
  textrect(c(0.5,0.49), radx = 0.35, rady = 0.49, lab = "" , box.col = "ivory", shadow.size=0)
  textrect(c(0.5,0.9375), radx = 0.35, rady = 0.045, lab = "" , box.col = "tan2", shadow.size=0, lcol="transparent")
  textrect(c(0.5,0.2075), radx = 0.35, rady = 0.045, lab = "" , box.col = "tan2", shadow.size=0, lcol="transparent")
  textrect(c(0.5,0.6865), radx = 0.35, rady = 0.05, lab = "" , box.col = "wheat2", shadow.size=0, lcol="transparent")
  textrect(c(0.5,0.3115), radx = 0.35, rady = 0.065, lab = "" , box.col = "wheat2", shadow.size=0, lcol="transparent")
  textrect(c(0.5,0.49), radx = 0.35, rady = 0.49, lab = "" , box.col = "transparent", shadow.size=0)
  
  for (i in 1:dim(elpos1)[1]-1) {
    straightarrow(to = elpos1[i+1, ],
                  from = elpos1[i, ], arr.type = "curved", arr.lwd = 0.5,
                  lwd = 2, arr.pos = 0.5, arr.length = 0.2, arr.width = 0.15)
  }
  
  for (i in 2:dim(elpos2)[1]) {
    straightarrow(to = elpos2[i, ],
                  from = elpos2[1, ], arr.type = "curved", arr.lwd = 0.5,
                  lwd = 2, arr.pos = 0.5, arr.length = 0.2, arr.width = 0.15)
  }
  
  for (i in 1:dim(elpos3)[1]-1) {
    straightarrow(to = elpos3[3, ],
                  from = elpos3[i, ], arr.type = "curved", arr.lwd = 0.5,
                  lwd = 2, arr.pos = 0.5, arr.length = 0.2, arr.width = 0.15)
  }
  
  elpostitles <- elpos[c(1,2,15),]
  titles <- c("Workflow", "Core functions", "Helper functions")
  for (i in 1:dim(elpostitles)[1]){
    textellipse(elpostitles[i,], radx=nchar(titles[i])*0.01, rady=0.03, lab = titles[i], cex = 0.7, adj=c(0.5, 0.5),
                shadow.col = "darkolivegreen4", shadow.size = 0.003, box.col = "white")
  }
  
  elposflow1 <- elpos[c(3,5,11,13),]
  elposflow2 <- elpos[c(7,8,9),]
  flow1 <- c("Data pre-processing", "Generation of\nKWIC Index",
             "Probable duplicate\nset retrieval", "Set review, modification\n& validation")
  flow2 <- c("Fuzzy\nmatching", "Phonetic\nmatching", "Semantic\nmatching")
  for (i in 1:dim(elposflow1)[1]){
    textrect(elposflow1[i,], radx = 0.13, rady = 0.04, lab = flow1[i] , box.col = "white",
             shadow.col = "indianred1", shadow.size = 0.005, cex = 0.7)
  }
  for (i in 1:dim(elposflow2)[1]){
    textrect(elposflow2[i,], radx = 0.05, rady = 0.04, lab = flow2[i] , box.col = "white",
             shadow.col = "indianred1", shadow.size = 0.005, cex = 0.7)
  }
  
  elposfunc <- elpos[c(4,6,10,14,16,17, 18),]
  elposfunc[-c(5),][,1] <- elposfunc[-c(5),][,1] - 0.06
  func <- c("DataClean\nMergeKW\nMergePrefix\nMergeSuffix",
            "KWIC", "ProbDup", "DisProbDup\nReviewProbDup\nReconstructProbDup",
            "read.genesys\nValidatePrimKey\nDoubleMetaphone\nKWCounts",
            "ParseProbDup\nSplitProbDup\nMergeProbDup",
            "AddProbDup\nViewProbDup\nprint.KWIC\nprint.ProbDup")
  for (i in 1:dim(elposfunc)[1]){
    textplain(elposfunc[i,], lab = func[i], cex = 0.7, adj=c(0, 0.5), font=1, family = "sans", col="steelblue4")
  }
} else {
  print("package 'diagram' is required to generate this figure")
}
```

**Fig. 1.** PGRdup workflow and associated functions

## Version History
```{r, results='asis', echo=FALSE}
# Fetch release version
rver <- getNamespaceVersion("PGRdup")

# rver <- ifelse(test = gsub("(.\\.)(\\d+)(\\..)", "", getNamespaceVersion("PGRdup")) == "",
#                yes = getNamespaceVersion("PGRdup"),
#                no = as.vector(available.packages()["PGRdup",]["Version"]))
```
The current version of the package is `r rver`. The previous versions are as follows.

**Table 1.** Version history of `PGRdup` `R` package.

```{r, echo=FALSE, message=FALSE}
if (requireNamespace("RCurl", quietly = TRUE) & requireNamespace("httr", quietly = TRUE) & requireNamespace("XML", quietly = TRUE)) {
  pkg <- "PGRdup"
  link <- paste0("https://cran.r-project.org/src/contrib/Archive/", pkg, "/")
  
  if (RCurl::url.exists(link)) {
    # cafile <- system.file("CurlSSL", "cacert.pem", package = "RCurl")
    # page <- httr::GET(link, httr::config(cainfo = cafile))
    page <- httr::GET(link)
    page <- httr::content(page, as = 'text')
    # page <- RCurl::getURL(link)
    
    VerHistory <- XML::readHTMLTable(page)[[1]][,2:3]
    colnames(VerHistory) <- c("Version", "Date")
    VerHistory <- VerHistory[VerHistory$Version != "Parent Directory",]
    VerHistory <- VerHistory[!is.na(VerHistory$Version), ]
    VerHistory$Date <- as.Date(VerHistory$Date)
    VerHistory$Version <- gsub("PGRdup_", "", VerHistory$Version)
    VerHistory$Version <- gsub(".tar.gz", "", VerHistory$Version)
    
    VerHistory <- VerHistory[order(VerHistory$Date), c("Version", "Date")]
    rownames(VerHistory) <- NULL
    
    knitr::kable(VerHistory)
    
  } else {
    print("Access to CRAN page for 'PGRdup' is required to generate this table.'")
  }
  
} else {
  print("Packages 'RCurl', 'httr' and 'XML' are required to generate this table.")
}

```

To know detailed history of changes use `news(package='PGRdup')`.

## Installation
The package can be installed using the following functions:

```{r, eval=FALSE}
# Install from CRAN
install.packages('PGRdup', dependencies=TRUE)

```

Uninstalled dependencies (packages which `PGRdup` depends on *viz*- [`data.table`](https://CRAN.R-project.org/package=data.table), [`igraph`](https://CRAN.R-project.org/package=igraph), [`stringdist`](https://CRAN.R-project.org/package=stringdist) and [`stringi`](https://CRAN.R-project.org/package=stringi) are also installed because of the argument `dependencies=TRUE`.

Then the package can be loaded using the function
```{r, eval=FALSE}
library(PGRdup)
```

## Data Format
The package is essentially designed to operate on PGR passport data present in a [data frame object](https://www.google.com/#q=[R]+data.frame), with  each row holding one record and columns representing the attribute fields. For example, consider the dataset `GN1000` supplied along with the package.
```{r}
library(PGRdup)
# Load the dataset to the environment
data(GN1000)
# Show the class of the object
class(GN1000)
# View the first few records in the data frame
head(GN1000)
```

If the passport data exists as an excel sheet, it can be first converted to a comma-separated values (csv) file or tab delimited file and then easily imported into the `R` environment using the base functions `read.csv` and `read.table` respectively. Similarly `read_csv()` and `read_tsv()` from the [`readr`](https://CRAN.R-project.org/package=readr) package can also be used. Alternatively, the package [`readxl`](https://CRAN.R-project.org/package=readxl) can be used to directly read the data from excel. In case of large csv files, the function `fread` in the [`data.table`](https://CRAN.R-project.org/package=data.table) package can be used to rapidly load the data.

If the PGR passport data is in a database management system (DBMS), the required table can be imported as a data frame into `R`. using the appropriate [`R`-database interface package](https://www.burns-stat.com/r-database-interfaces/). For example [`dbConnect`](https://CRAN.R-project.org/package=dbConnect) for MySQL, [`ROracle`](https://CRAN.R-project.org/package=ROracle) for Oracle etc.

The PGR data downloaded from the [genesys](https://www.genesys-pgr.org/welcome) database as a [Darwin Core - Germplasm](https://github.com/dagendresen/darwincore-germplasm) zip archive can be imported into the `R` environment as a flat file `data.frame` using the `read.genesys` function.

```{r, eval=FALSE}
# Import the DwC-Germplasm zip archive "genesys-accessions-filtered.zip"
PGRgenesys <- read.genesys("genesys-accessions-filtered.zip",
                           scrub.names.space = TRUE, readme = TRUE)
```


## Data Pre-processing
Data pre-processing is a critical step which can affect the quality of the probable duplicate sets being retrieved. It involves data standardization as well as data cleaning which can be achieved using the functions `DataClean`, `MergeKW`, `MergePrefix` and `MergeSuffix`.

`DataClean` function can be used to clean the character strings in passport data fields(columns) specified as the input [character vector](https://www.google.com/#q=[R]+character+vector) `x` according to the conditions specified in the arguments.

Commas, semicolons and colons which are sometimes used to separate multiple strings or names within the same field can be replaced with a single space using the logical arguments `fix.comma`, `fix.semcol` and `fix.col` respectively.
```{r}
x <- c("A 14; EC 1697", "U 4-4-28; EC 21078; A 32", "PI 262801:CIAT 9075:GKP 9553/90",
       "NCAC 16049, PI 261987, RCM 493-3")
x
# Replace ',', ':' and ';' with space
DataClean(x, fix.comma=TRUE, fix.semcol=TRUE, fix.col=TRUE,
          fix.bracket=FALSE, fix.punct=FALSE, fix.space=FALSE, fix.sep=FALSE,
          fix.leadzero=FALSE)
```

Similarly the logical argument `fix.bracket` can be used to replace all brackets including parenthesis, square brackets and curly brackets with space.
```{r}
x <- c("(NRCG-1738)/(NFG649)", "26-5-1[NRCG-2528]", "Ah 1182 {NRCG-4340}")
x
# Replace parenthesis, square brackets and curly brackets with space
DataClean(x, fix.comma=FALSE, fix.semcol=FALSE, fix.col=FALSE,
          fix.bracket=TRUE,
          fix.punct=FALSE, fix.space=FALSE, fix.sep=FALSE, fix.leadzero=FALSE)
```

The logical argument `fix.punct` can be used to remove all punctuation from the data.
```{r}
x <- c("#26-6-3-1", "Culture No. 857", "U/4/47/13")
x
# Remove punctuation
DataClean(x, fix.comma=FALSE, fix.semcol=FALSE, fix.col=FALSE, fix.bracket=FALSE,
          fix.punct=TRUE,
          fix.space=FALSE, fix.sep=FALSE, fix.leadzero=FALSE)
```

`fix.space` can be used to convert all space characters such as tab, newline, vertical tab, form feed and carriage return to spaces and finally convert multiple spaces to single space.
```{r}
x <- c("RS   1", "GKSPScGb 208  PI 475855")
x
# Replace all space characters to space and convert multiple spaces to single space
DataClean(x, fix.comma=FALSE, fix.semcol=FALSE, fix.col=FALSE,
          fix.bracket=FALSE, fix.punct=FALSE,
          fix.space=TRUE,
          fix.sep=FALSE, fix.leadzero=FALSE)
```

`fix.sep` can be used to merge together accession identifiers composed of alphabetic characters separated from a series of digits by a space character.
```{r}
x <- c("NCAC 18078", "AH 6481", "ICG 2791")
x
# Merge alphabetic character separated from a series of digits by a space
DataClean(x, fix.comma=FALSE, fix.semcol=FALSE, fix.col=FALSE,
          fix.bracket=FALSE, fix.punct=FALSE, fix.space=FALSE,
          fix.sep=TRUE,
          fix.leadzero=FALSE)
```

`fix.leadzero` can be used to remove leading zeros from accession name fields to facilitate matching to identify probable duplicates.
```{r}
x <- c("EC 0016664", "EC0001690")
x
# Remove leading zeros
DataClean(x, fix.comma=FALSE, fix.semcol=FALSE, fix.col=FALSE,
          fix.bracket=FALSE, fix.punct=FALSE, fix.space=FALSE, fix.sep=FALSE,
          fix.leadzero=TRUE)
```

This function can hence be made use of in tidying up multiple forms of messy data existing in fields associated with accession names in PGR passport databases (Table&nbsp;1).
```{r}
names <- c("S7-12-6", "ICG-3505", "U 4-47-18;EC 21127", "AH 6481", "RS   1",
           "AK 12-24", "2-5 (NRCG-4053)", "T78, Mwitunde", "ICG 3410",
           "#648-4 (Gwalior)", "TG4;U/4/47/13", "EC0021003")
names
# Clean the data
DataClean(names)
```


```{r, eval=FALSE, echo = FALSE}
# DC <- data.frame(names = names, `DataClean(names)` = DataClean(names))
# knitr::kable(DC)
```

**Table 2.** Data pre-processing using `DataClean`.

|**names**          |**DataClean(names)** |
|:------------------|:--------------------|
|S7-12-6            |S7126                |
|ICG-3505           |ICG3505              |
|U 4-47-18;EC 21127 |U44718 EC21127       |
|AH 6481            |AH6481               |
|RS   1             |RS1                  |
|AK 12-24           |AK1224               |
|2-5 (NRCG-4053)    |25 NRCG4053          |
|T78, Mwitunde      |T78 MWITUNDE         |
|ICG 3410           |ICG3410              |
|#648-4 (Gwalior)   |6484 GWALIOR         |
|TG4;U/4/47/13      |TG4 U44713           |
|EC0021003          |EC21003              |

Several common keyword string pairs or keyword prefixes and suffixes exist in fields associated with accession names in PGR passport databases. They can be merged using the functions `MergeKW`, `MergePrefix` and `MergeSuffix` respectively. The keyword string pairs, prefixes and suffixes can be supplied as a [list](https://www.google.com/#q=[R]+list) or a [vector](https://www.google.com/#q=[R]+vector) to the argument `y` in these functions.

```{r}
names <- c("Punjab Bold", "Gujarat- Dwarf", "Nagpur.local", "SAM COL 144",
           "SAM COL--280", "NIZAMABAD-LOCAL", "Dark Green Mutant",
           "Dixie-Giant", "Georgia- Bunch", "Uganda-erect", "Small Japan",
           "Castle  Cary", "Punjab erect", "Improved small japan",
           "Dark Purple")
names

# Merge pairs of strings
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
           c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
           c("Mota", "Company"))
names <- MergeKW(names, y1, delim = c("space", "dash", "period"))

# Merge prefix strings
y2 <- c("Light", "Small", "Improved", "Punjab", "SAM", "Dark")
names <- MergePrefix(names, y2, delim = c("space", "dash", "period"))

# Merge suffix strings
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
        "Bunch", "Peanut")
names <- MergeSuffix(names, y3, delim = c("space", "dash", "period"))

names

```

These functions can be applied over multiple columns(fields) in a data frame using the [`lapply`](https://www.google.com/#q=[R]+lapply) function.

```{r}
# Load example dataset
GN <- GN1000

# Specify as a vector the database fields to be used
GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")
head(GN[GNfields])

# Clean the data
GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
c("Mota", "Company"))
y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
        "Bunch", "Peanut")
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergeSuffix(x, y3, delim = c("space", "dash")))
head(GN[GNfields])
```


## Generation of KWIC Index
The function `KWIC` generates a Key Word in Context index [@knupffer1988european; @kfj97] from the data frame of a PGR passport database based on the fields(columns) specified in the argument `fields` along with the keyword frequencies and gives the output as a list of class `KWIC`. The first element of the vector specified in `fields` is considered as the primary key or identifier which uniquely identifies all rows in the data frame.

This function fetches keywords from different fields specified, which can be subsequently used for matching to identify probable duplicates. The frequencies of the keywords retrieved can help in determining if further data pre-processing is required and also to decide whether any common keywords can be exempted from matching (Fig.&nbsp;2).

```{r}
# Load example dataset
GN <- GN1000

# Specify as a vector the database fields to be used
GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")

# Clean the data
GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
c("Mota", "Company"))
y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
        "Bunch", "Peanut")
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergeSuffix(x, y3, delim = c("space", "dash")))

# Generate the KWIC index
GNKWIC <- KWIC(GN, GNfields, min.freq = 1)
class(GNKWIC)
GNKWIC

# Retrieve the KWIC index from the KWIC object
KWIC <- GNKWIC[[1]]
KWIC <- KWIC[order(KWIC$KEYWORD, decreasing = TRUE),]
head(KWIC[,c("PRIM_ID", "KWIC_L", "KWIC_KW", "KWIC_R")], n = 10)

# Retrieve the keyword frequencies from the KWIC object
KeywordFreq <- GNKWIC[[2]]
head(KeywordFreq)
```


```{r, echo=FALSE, warning=FALSE, fig.height = 4, eval=TRUE}
# Plot wordcloud of keyword frequencies
if (requireNamespace("wordcloud", quietly = TRUE)) {
  suppressMessages(library(wordcloud))
  par(mar = c(0,0,0,0))
  # pal <- brewer.pal(8,"Dark2")
  pal <- c("#1B9E77", "#D95F02", "#7570B3", "#E7298A", "#66A61E", "#E6AB02",
           "#A6761D", "#666666")
  wordcloud(words = KeywordFreq[,1], freq = KeywordFreq[,2], min.freq = 4,
            colors = pal, random.order = FALSE, rot.per = 0,
            fixed.asp = FALSE)
} else {
  print("package 'wordcloud' is required to generate this figure")
}

```

**Fig. 2.** Word cloud of keywords retrieved

The function will throw an error in case of duplicates or NULL values in the primary key/ID field mentioned.

```{r}
GN <- GN1000
GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
# Generate dummy duplicates for illustration
GN[1001:1005,] <- GN[1:5,]
# Generate dummy NULL values for illustration
GN[1001,3] <- ""
GN[1002,3] <- ""
GN[1001:1005,]
```


```{r, eval = FALSE}
GNKWIC <- KWIC(GN, GNfields, min.freq=1)
```


```{r, echo = FALSE}
message(paste("Error in KWIC(GN, GNfields, min.freq = 1) :", 
              "  Primary key/ID field should be unique and not NULL",
              " Use PGRdup::ValidatePrimKey() to identify and rectify the aberrant records first",
              sep = "\n"))
```


The erroneous records can be identified using the helper function `ValidatePrimKey`.

```{r, message=FALSE}
# Validate the primary key/ID field for duplication or existence of NULL values
ValidatePrimKey(x = GN, prim.key = "NationalID")
# Remove the offending records
GN <- GN[-c(1001:1005), ]
# Validate again
ValidatePrimKey(x = GN, prim.key = "NationalID")
```


## Retrieval of Probable Duplicate Sets
Once KWIC indexes are generated, probable duplicates of germplasm accessions can be identified by fuzzy, phonetic and semantic matching of the associated keywords using the function `ProbDup`. The sets are retrieved as a list of data frames of class `ProbDup`.

Keywords that are not to be used for matching can be specified as a vector in the `excep` argument.

### Methods
The function can execute matching according to either one of the following three methods as specified by the `method` argument.

1. **Method `"a"`** : Performs string matching of keywords in a single KWIC index to identify probable duplicates of accessions in a single PGR passport database.

```{r}
# Load example dataset
GN <- GN1000

# Specify as a vector the database fields to be used
GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")

# Clean the data
GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
c("Mota", "Company"))
y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
        "Bunch", "Peanut")
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergeSuffix(x, y3, delim = c("space", "dash")))

# Generate the KWIC index
GNKWIC <- KWIC(GN, GNfields)
```

```{r}
# Specify the exceptions as a vector
exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE",
         "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT",
         "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE",
         "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R",
         "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE",
         "U", "VALENCIA", "VIRGINIA", "WHITE")

# Fetch fuzzy duplicates by method 'a'
GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, fuzzy = TRUE,
                 phonetic = FALSE, semantic = FALSE)
class(GNdup)
GNdup
```

```{r}
# Fetch phonetic duplicates by method 'a'
GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, fuzzy = FALSE,
                 phonetic = TRUE, semantic = FALSE)
class(GNdup)
GNdup
```

2. **Method `"b"`** : Performs string matching of keywords in the first KWIC index (query) with that of the keywords in the second index (source) to identify probable duplicates of accessions of the first PGR passport database among the accessions in the second database.

3. **Method `"c"`** : Performs string matching of keywords in two different KWIC indexes jointly to identify probable duplicates of accessions from among two PGR passport databases.

```{r}
# Load PGR passport databases
GN1 <- GN1000[!grepl("^ICG", GN1000$DonorID), ]
GN1$DonorID <- NULL
GN2 <- GN1000[grepl("^ICG", GN1000$DonorID), ]
GN2$NationalID <- NULL

# Specify database fields to use
GN1fields <- c("NationalID", "CollNo", "OtherID1", "OtherID2")
GN2fields <- c("DonorID", "CollNo", "OtherID1", "OtherID2")

# Clean the data
GN1[GN1fields] <- lapply(GN1[GN1fields], function(x) DataClean(x))
GN2[GN2fields] <- lapply(GN2[GN2fields], function(x) DataClean(x))
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
c("Mota", "Company"))
y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
        "Bunch", "Peanut")
GN1[GN1fields] <- lapply(GN1[GN1fields],
                         function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN1[GN1fields] <- lapply(GN1[GN1fields],
                         function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN1[GN1fields] <- lapply(GN1[GN1fields],
                         function(x) MergeSuffix(x, y3, delim = c("space", "dash")))
GN2[GN2fields] <- lapply(GN2[GN2fields],
                         function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN2[GN2fields] <- lapply(GN2[GN2fields],
                         function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN2[GN2fields] <- lapply(GN2[GN2fields],
                         function(x) MergeSuffix(x, y3, delim = c("space", "dash")))

# Remove duplicated DonorID records in GN2
GN2 <- GN2[!duplicated(GN2$DonorID), ]

# Generate KWIC index
GN1KWIC <- KWIC(GN1, GN1fields)
GN2KWIC <- KWIC(GN2, GN2fields)

# Specify the exceptions as a vector
exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE",
         "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT",
         "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE",
         "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R",
         "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE",
         "U", "VALENCIA", "VIRGINIA", "WHITE")

# Fetch fuzzy and phonetic duplicate sets by method b
GNdupb <- ProbDup(kwic1 = GN1KWIC, kwic2 = GN2KWIC, method = "b",
                  excep = exep, fuzzy = TRUE, phonetic = TRUE,
                  encoding = "primary", semantic = FALSE)
class(GNdupb)
GNdupb
```

```{r}
# Fetch fuzzy and phonetic duplicate sets by method c
GNdupc <- ProbDup(kwic1 = GN1KWIC, kwic2 = GN2KWIC, method = "c",
                  excep = exep, fuzzy = TRUE, phonetic = TRUE,
                  encoding = "primary", semantic = FALSE)
class(GNdupc)
GNdupc
```

### Matching Strategies

1. **Fuzzy matching** or approximate string matching of keywords is carried out by computing the [generalized levenshtein (edit) distance](https://en.wikipedia.org/wiki/Levenshtein_distance) between them. This distance measure counts the number of deletions, insertions and substitutions necessary to turn one string to another.

```{r}
# Load example dataset
GN <- GN1000

# Specify as a vector the database fields to be used
GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")

# Clean the data
GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
c("Mota", "Company"))
y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
        "Bunch", "Peanut")
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergeSuffix(x, y3, delim = c("space", "dash")))

# Generate the KWIC index
GNKWIC <- KWIC(GN, GNfields)

# Specify the exceptions as a vector
exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE",
         "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT",
         "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE",
         "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R",
         "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE",
         "U", "VALENCIA", "VIRGINIA", "WHITE")

# Fetch fuzzy duplicates
GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, 
                 fuzzy = TRUE, max.dist = 3,
                 phonetic = FALSE, semantic = FALSE)
GNdup
```

The maximum distance to be considered for a match can be specified by `max.dist` argument.

```{r}
GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, 
                 fuzzy = TRUE, max.dist = 1,
                 phonetic = FALSE, semantic = FALSE)
GNdup
```

Exact matching can be enforced with the argument `force.exact` set as TRUE. It can be used to avoid fuzzy matching when the number of alphabet characters in keywords is lesser than a critical value (`max.alpha`). Similarly, the value of `max.digit` can also be set according to the requirements to enforce exact matching. The default value of `Inf` avoids fuzzy matching and enforces exact matching for all keywords having any numerical characters. If `max.digit` and `max.alpha` are both set to `Inf`, exact matching will be enforced for all the keywords.
    
When exact matching is enforced, for keywords having both alphabet and numeric characters and with the number of alphabet characters greater than `max.digit`, matching will be carried out separately for alphabet and numeric characters present.
    
```{r}
GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, 
                 fuzzy = TRUE, force.exact = TRUE, max.alpha = 4, max.digit = Inf,
                 phonetic = FALSE, semantic = FALSE)
GNdup
```

2. **Phonetic matching** of keywords is carried out using the Double Metaphone phonetic algorithm which is implemented as the helper function `DoubleMetaphone`, [@p00], to identify keywords that have the similar pronunciation.

```{r}
GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, 
                 fuzzy = FALSE,
                 phonetic = TRUE,
                 semantic = FALSE)
GNdup
```

Either the primary or alternate encodings can be used by specifying the `encoding` argument.
    
```{r}
GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, 
                 fuzzy = FALSE,
                 phonetic = TRUE, encoding = "alternate",
                 semantic = FALSE)
GNdup
```

The argument `phon.min.alpha` sets the limits for the number of alphabet characters to be present in a string for executing phonetic matching.
    
```{r}
GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, 
                 fuzzy = FALSE,
                 phonetic = TRUE, encoding = "alternate", phon.min.alpha = 4,
                 semantic = FALSE)
GNdup
```

Similarly `min.enc` sets the limits for the number of characters to be present in the encoding of a keyword for phonetic matching.
    
```{r}
GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, 
                 fuzzy = FALSE,
                 phonetic = TRUE, encoding = "alternate", min.enc = 4,
                 semantic = FALSE)
GNdup
```

3. **Semantic matching** matches keywords based on a list of accession name synonyms supplied as list with character vectors of synonym sets (synsets) to the `syn` argument. Synonyms in this context refer to interchangeable identifiers or names by which an accession is recognized. Multiple keywords specified as members of the same synset in `syn` are matched. To facilitate accurate identification of synonyms from the KWIC index, identical data standardization operations using the `Merge*` and `DataClean` functions for both the original database fields and the synset list are recommended.

```{r}
# Specify the synsets as a list
syn <- list(c("CHANDRA", "AH 114"), c("TG-1", "VIKRAM"))

# Clean the data in the synsets
syn <- lapply(syn, DataClean)

GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, 
                 fuzzy = FALSE, phonetic = FALSE,
                 semantic = TRUE, syn = syn)
GNdup
```

### Memory and Speed Constraints
As the number of keywords in the KWIC indexes increases, the memory consumption by the function also increases proportionally. This is due to the reason that for string matching, this function relies upon creation of a _n_$\times$_m_ matrix of all possible keyword pairs for comparison, where _n_ and _m_ are the number of keywords in the query and source indexes respectively. This can lead to `cannot allocate vector of size...` errors in case of large KWIC indexes where the comparison matrix is too large to reside in memory. In such a case, the `chunksize` argument can be reduced from the default 1000 to get the appropriate size of the KWIC index keyword block to be used for searching for matches at a time. However a smaller `chunksize` may lead to longer computation time due to the memory-time trade-off.

The progress of matching is displayed in the console as number of keyword blocks completed out of the total number of blocks, the percentage of achievement and a text-based progress bar.

In case of multi-byte characters in keywords, the speed of keyword matching is further dependent upon the `useBytes` argument as described in `help("stringdist-encoding")` for the `stringdist` function in the namesake [package](https://CRAN.R-project.org/package=stringdist) [@van2014stringdist], which is made use of here for string matching.

The CPU time taken for retrieval of probable duplicate sets under different options for the arguments `chunksize` and `useBytes` can be visualized using the [`microbenchmark`](https://CRAN.R-project.org/package=microbenchmark) package (Fig.&nbsp;3).

```{r, eval = FALSE}
# Load example dataset
GN <- GN1000

# Specify as a vector the database fields to be used
GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")

# Clean the data
GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
           c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
           c("Mota", "Company"))
y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.", "Bunch", "Peanut")
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergeSuffix(x, y3, delim = c("space", "dash")))

# Generate the KWIC index
GNKWIC <- KWIC(GN, GNfields)

# Specify the exceptions as a vector
exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE", "DARK",
          "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT", "GUTHUKAI",
          "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE", "LIGHT", "LOCAL",
          "OF", "OVERO", "P", "PEANUT", "PURPLE", "R", "RED", "RUNNER", "S1", "SAM",
          "SMALL", "SPANISH", "TAN", "TYPE", "U", "VALENCIA", "VIRGINIA", "WHITE")

# Specify the synsets as a list
syn <- list(c("CHANDRA", "AH 114"), c("TG-1", "VIKRAM"))
syn <- lapply(syn, DataClean)

```

```{r, eval=TRUE, echo=FALSE}
if (requireNamespace("microbenchmark", quietly = TRUE)) {
  mbm <- TRUE
} else {
  mbm <- FALSE
}
```

```{r, eval=FALSE, echo=TRUE}
timings <- microbenchmark::microbenchmark(
  # Fetch duplicate sets with default chunk.size
  t1 = ProbDup(kwic1 = GNKWIC, method = "a", excep = exep,
                                     chunksize = 1000, useBytes = TRUE,
                                     fuzzy = TRUE, phonetic = TRUE,
                                     semantic = TRUE, syn = syn),
  # Fetch duplicate sets chunk.size 2000
  t2 = ProbDup(kwic1 = GNKWIC, method = "a", excep = exep,
                                     chunksize = 2000, useBytes = TRUE,
                                     fuzzy = TRUE, phonetic = TRUE,
                                     semantic = TRUE, syn = syn),
  # Fetch duplicate sets chunk.size 100
  t3 = ProbDup(kwic1 = GNKWIC, method = "a", excep = exep,
                                     chunksize = 100, useBytes = TRUE,
                                     fuzzy = TRUE, phonetic = TRUE,
                                     semantic = TRUE, syn = syn),
  # Fetch duplicate sets useBytes = FALSE
  t4 = ProbDup(kwic1 = GNKWIC, method = "a", excep = exep,
                                     chunksize = 1000, useBytes = FALSE,
                                     fuzzy = TRUE, phonetic = TRUE,
                                     semantic = TRUE, syn = syn), times = 10)
```

```{r, eval = FALSE, echo=TRUE}
plot(timings, col = c("#1B9E77", "#D95F02", "#7570B3", "#E7298A"),
     xlab = "Expression", ylab = "Time")
legend("topright", c("t1 : chunksize = 1000,\n     useBytes = T (default)\n",
         "t2 : chunksize = 2000,\n     useBytes = T\n",
         "t3 : chunksize = 500,\n     useBytes = T\n",
         "t4 : chunksize = 1000,\n     useBytes = F\n"),
       bty = "n", cex = 0.6)
```

```{r, eval=TRUE, echo=FALSE}
timings <- data.frame(expr = structure(c(2L, 1L, 1L, 3L, 1L, 2L, 3L, 
                                   4L, 3L, 3L, 4L, 2L, 2L, 1L, 2L, 4L, 4L, 4L, 1L, 4L, 3L, 2L, 1L, 
                                   1L, 3L, 2L, 1L, 4L, 2L, 3L, 1L, 2L, 4L, 3L, 4L, 3L, 2L, 3L, 4L, 
                                   1L), .Label = c("t1", "t2", "t3", "t4"), class = "factor"),
                time = c(1408220600,
                         1363532801, 1400142541, 3000887552, 1399254608, 1224892145, 3013227764, 
                         1328487587, 2916351114, 2950914731, 1353476659, 1509574164, 1278119453, 
                         1413596470, 1303064433, 1350284131, 1467990710, 1444774083, 1375139405, 
                         1396005444, 3105281223, 1224878080, 1360539070, 1421216118, 3245590555, 
                         1367002429, 1877008017, 1622301327, 1251792020, 2949414824, 1362527416, 
                         1263918992, 1415981648, 2955363516, 1341522638, 2957571945, 1402744001, 
                         3177392306, 1349590813, 1565604772))

boxplot(time ~ expr, timings, xlab = "Expression", ylab = "Time",
        col = c("#1B9E77", "#D95F02", "#7570B3", "#E7298A"))
legend("topright", c("t1 : chunksize = 1000,\n     useBytes = T (default)\n",
         "t2 : chunksize = 2000,\n     useBytes = T\n",
         "t3 : chunksize = 500,\n     useBytes = T\n",
         "t4 : chunksize = 1000,\n     useBytes = F\n"),
       bty = "n", cex = 0.6)
```

**Fig. 3.** CPU time with different `ProbDup` arguments estimated using the `microbenchmark` package.

## Set Review, Modification and Validation
The initially retrieved sets may be intersecting with each other because there might be accessions which occur in more than duplicate set. Disjoint sets can be generated by merging such overlapping sets using the function `DisProbDup`.

Disjoint sets are retrieved either individually for each type of probable duplicate sets or considering all type of sets simultaneously. In case of the latter, the disjoint of all the type of sets alone are returned in the output as an additional data frame `DisjointDupicates` in an object of class `ProbDup`.

```{r, results="hide", message = FALSE}
# Load example dataset
GN <- GN1000

# Specify as a vector the database fields to be used
GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")

# Clean the data
GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
c("Mota", "Company"))
y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
        "Bunch", "Peanut")
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergeSuffix(x, y3, delim = c("space", "dash")))

# Generate KWIC index
GNKWIC <- KWIC(GN, GNfields)

# Specify the exceptions as a vector
exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE",
         "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT",
         "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE",
         "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R",
         "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE",
         "U", "VALENCIA", "VIRGINIA", "WHITE")

# Specify the synsets as a list
syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))

# Fetch probable duplicate sets
GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, fuzzy = TRUE,
                 phonetic = TRUE, encoding = "primary",
                 semantic = TRUE, syn = syn)
```


```{r}
# Initial number of sets
GNdup

# Get disjoint probable duplicate sets of each kind
disGNdup1 <- DisProbDup(GNdup, combine = NULL)
# # Number of sets after combining intersecting sets
disGNdup1

# Get disjoint probable duplicate sets combining all the kinds of sets
disGNdup2 <- DisProbDup(GNdup, combine = c("F", "P", "S"))
# Number of sets after combining intersecting sets
disGNdup2

```


Once duplicate sets are retrieved they can be validated by manual clerical review by comparing with original PGR passport database(s) using the `ReviewProbDup` function. This function helps to retrieve PGR passport information associated with fuzzy, phonetic or semantic probable duplicate sets in an object of class `ProbDup` from the original databases(s) from which they were identified. The original information of accessions comprising a set, which have not been subjected to data standardization can be compared under manual clerical review for the validation of the set. By default only the fields(columns) which were used initially for creation of the KWIC indexes using the KWIC function are retrieved. Additional fields(columns) if necessary can be specified using the `extra.db1` and `extra.db2` arguments.

When any primary ID/key records in the fuzzy, phonetic or semantic duplicate sets are found to be missing from the original databases specified in `db1` and `db2`, then they are ignored and only the matching records are considered for retrieving the information with a warning.

This may be due to data standardization of the primary ID/key field using the function `DataClean` before creation of the KWIC index and subsequent identification of probable duplicate sets. In such a case, it is recommended to use an identical data standardization operation on the  primary ID/key field of databases specified in `db1` and `db2` before running this function.

With `R` <= v3.0.2, due to copying of named objects by `list()`, `Invalid .internal.selfref detected and fixed...` warning can appear, which may be safely ignored.

The output data frame can be subjected to clerical review either after exporting into an external spreadsheet using `write.csv` function or by using the `edit` function.

The column `DEL` can be used to indicate whether a record has to be deleted from a set or not. `Y` indicates "Yes", and the default `N` indicates "No".

The column `SPLIT` similarly can be used to indicate whether a record in a set has to be branched into a new set. A set of identical integers in this column other than the default `0` can be used to indicate that they are to be removed and assembled into a new set.


```{r}
# Load the original database and clean the Primary ID/key field
GN1000 <- GN1000
GN1000$NationalID <- DataClean(GN1000$NationalID)

# Get the data frame for reviewing the duplicate sets identified
RevGNdup <- ReviewProbDup(pdup = disGNdup1, db1 = GN1000,
                          extra.db1 = c("SourceCountry", "TransferYear"),
                          max.count = 30, insert.blanks = TRUE)
```

```{r}
head(RevGNdup)
```


```{r, eval=FALSE}
# Examine and review the duplicate sets using edit function
RevGNdup <- edit(RevGNdup)

# OR examine and review the duplicate sets after exporting them as a csv file
write.csv(file="Duplicate sets for review.csv", x=RevGNdup)
```

After clerical review, the data frame created using the function `ReviewProbDup` from an object of class `ProbDup` can be reconstituted back to the same object after the review using the function `ReconstructProbDup`.

The instructions for modifying the sets entered in the appropriate format in the columns `DEL` and `SPLIT` during clerical review are taken into account for reconstituting the probable duplicate sets. Any records with `Y` in column `DEL` are deleted and records with identical integers in the column `SPLIT` other than the default `0` are reassembled into a new set.

```{r}
# The original set data
subset(RevGNdup, SET_NO==13 & TYPE=="P", select= c(IDKW, DEL, SPLIT))
# Make dummy changes to the set for illustration
RevGNdup[c(113, 116), 6] <- "Y"
RevGNdup[c(111, 114), 7] <- 1
RevGNdup[c(112, 115, 117), 7] <- 2
# The instruction for modification in columns DEL and SPLIT
subset(RevGNdup, SET_NO==13 & TYPE=="P", select= c(IDKW, DEL, SPLIT))

# Reconstruct ProDup object
GNdup2 <- ReconstructProbDup(RevGNdup)
# Initial no. of sets
disGNdup1
# No. of sets after modifications
GNdup2
```

## Other Functions
The `ProbDup` object is a list of data frames of different kinds of probable duplicate sets _viz_- `FuzzyDuplicates`, `PhoneticDuplicates`, `SemanticDuplicates` and `DisjointDuplicates`. Each row of the component data frame will have information of a set, the type of set, the set members as well as the keywords based on which the set was formed. This data can be reshaped into long form using the function `ParseProbDup`. This function which will transform a `ProbDup` object into a single data frame.

```{r}
# Convert 'ProbDup' object to a long form data frame of sets
GNdupParsed <- ParseProbDup(GNdup)
head(GNdupParsed)
```

The prefix `K*` here indicates the KWIC index of origin. This is useful in ascertaining the database of origin of the accessions when method `"b"` or `"c"` was used to create the input `ProbDup` object.

Once the sets are reviewed and modified, the validated set data fields from the `ProbDup` object can be added to the original PGR passport database using the function `AddProbDup`. The  associated data fields such as `SET_NO`, `ID` and `IDKW` are added based on the `PRIM_ID` field(column).

```{r}
# Loading original database
GN2 <- GN1000

# Add the duplicates set data to the original database
GNwithdup <-  AddProbDup(pdup = GNdup, db = GN2, addto = "I")
```

In case more than one KWIC index was used to generate the object of class `ProbDup`, the argument `addto` can be used to specify to which database the data fields are to be added. The default `"I"` indicates the database from which the first KWIC index was created and `"II"` indicates the database from which the second index was created.

The function `SplitProbDup` can be used to split an object of class `ProbDup` into two on the basis of set counts. This is useful for reviewing separately the sets with larger set counts.

```{r, results="hide", message = FALSE}
# Load PGR passport database
GN <- GN1000

# Specify as a vector the database fields to be used
GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")

# Clean the data
GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
c("Mota", "Company"))
y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
        "Bunch", "Peanut")
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields],
                       function(x) MergeSuffix(x, y3, delim = c("space", "dash")))

# Generate KWIC index
GNKWIC <- KWIC(GN, GNfields)

# Specify the exceptions as a vector
exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE",
         "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT",
         "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE",
         "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R",
         "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE",
         "U", "VALENCIA", "VIRGINIA", "WHITE")

# Specify the synsets as a list
syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))

# Fetch probable duplicate sets
GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, fuzzy = TRUE,
                 phonetic = TRUE, encoding = "primary",
                 semantic = TRUE, syn = syn)
```


```{r}
# Split the probable duplicate sets
GNdupSplit <- SplitProbDup(GNdup, splitat = c(10, 10, 10))
GNdupSplit[[1]]
GNdupSplit[[3]]
```

Alternatively, two different `ProbDup` objects can be merged together using  the function `MergeProbDup`.

```{r}
GNdupMerged <- MergeProbDup(GNdupSplit[[1]], GNdupSplit[[3]])
GNdupMerged
```

The summary of accessions according to a grouping factor field(column) in the original database(s) within the probable duplicate sets retrieved in a `ProbDup` object can be visualized by the `ViewProbDup` function. The resulting plot can be used to examine the extent of probable duplication within and between groups of accessions records.

```{r}
# Load PGR passport databases
GN1 <- GN1000[!grepl("^ICG", GN1000$DonorID), ]
GN1$DonorID <- NULL
GN2 <- GN1000[grepl("^ICG", GN1000$DonorID), ]
GN2 <- GN2[!grepl("S", GN2$DonorID), ]
GN2$NationalID <- NULL

GN1$SourceCountry <- toupper(GN1$SourceCountry)
GN2$SourceCountry <- toupper(GN2$SourceCountry)

GN1$SourceCountry <- gsub("UNITED STATES OF AMERICA", "USA", GN1$SourceCountry)
GN2$SourceCountry <- gsub("UNITED STATES OF AMERICA", "USA", GN2$SourceCountry)

# Specify as a vector the database fields to be used
GN1fields <- c("NationalID", "CollNo", "OtherID1", "OtherID2")
GN2fields <- c("DonorID", "CollNo", "OtherID1", "OtherID2")

# Clean the data
GN1[GN1fields] <- lapply(GN1[GN1fields], function(x) DataClean(x))
GN2[GN2fields] <- lapply(GN2[GN2fields], function(x) DataClean(x))
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
           c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
           c("Mota", "Company"))
y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
        "Bunch", "Peanut")
GN1[GN1fields] <- lapply(GN1[GN1fields],
                         function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN1[GN1fields] <- lapply(GN1[GN1fields],
                         function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN1[GN1fields] <- lapply(GN1[GN1fields],
                         function(x) MergeSuffix(x, y3, delim = c("space", "dash")))
GN2[GN2fields] <- lapply(GN2[GN2fields],
                         function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN2[GN2fields] <- lapply(GN2[GN2fields],
                         function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN2[GN2fields] <- lapply(GN2[GN2fields],
                         function(x) MergeSuffix(x, y3, delim = c("space", "dash")))

# Remove duplicated DonorID records in GN2
GN2 <- GN2[!duplicated(GN2$DonorID), ]

# Generate KWIC index
GN1KWIC <- KWIC(GN1, GN1fields)
GN2KWIC <- KWIC(GN2, GN2fields)

# Specify the exceptions as a vector
exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE",
          "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT",
          "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE",
          "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R",
          "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE",
          "U", "VALENCIA", "VIRGINIA", "WHITE")

# Specify the synsets as a list
syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))
```

```{r}
GNdupc <- ProbDup(kwic1 = GN1KWIC, kwic2 = GN2KWIC, method = "c",
                  excep = exep, fuzzy = TRUE, phonetic = TRUE,
                  encoding = "primary", semantic = TRUE, syn = syn)

# Get the summary data.frames and Grob
GNdupcView <- ViewProbDup(GNdupc, GN1, GN2, "SourceCountry", "SourceCountry",
                         max.count = 30, select = c("INDIA", "USA"), order = "type",
                         main = "Groundnut Probable Duplicates")
```

```{r, eval = FALSE}
# View the summary data.frames
GNdupcView[[1]]
GNdupcView[[2]]
```

```{r, fig.height = 7, fig.width=10}
# Plot the summary visualization
library(gridExtra)
grid.arrange(GNdupcView[[3]])
```

**Fig. 5.** Summary visualization of groundnut probable duplicate sets retrieved according to `SourceCountry` field.

The function `KWCounts` can be used to compute the keyword counts from PGR passport database fields(columns) which are considered for identification of probable duplicates. These keyword counts can give a rough indication of the completeness of the data in such fields (Fig.&nbsp;3).

```{r, fig.height = 8}
# Compute the keyword counts for the whole data
GNKWCouts <- KWCounts(GN, GNfields, exep)

# Compute the keyword counts for 'duplicated' records
GND <- ParseProbDup(disGNdup2, Inf, F)$PRIM_ID

GNDKWCouts <- KWCounts(GN[GN$NationalID %in% GND, ],
                       GNfields, exep)

# Compute the keyword counts for 'unique' records
GNUKWCouts <- KWCounts(GN[!GN$NationalID %in% GND, ],
                       GNfields, exep)

# Plot the counts as barplot
par(mfrow = c(3,1))

bp1 <- barplot(table(GNKWCouts$COUNT),
               xlab = "Word count", ylab = "Frequency",
               main = "A", col = "#1B9E77")
text(bp1, 0, table(GNKWCouts$COUNT),cex = 1, pos = 3)
legend("topright", paste("No. of records =",
                         nrow(GN)),
       bty = "n")

bp2 <- barplot(table(GNDKWCouts$COUNT),
               xlab = "Word count", ylab = "Frequency",
               main = "B", col = "#D95F02")
text(bp2, 0, table(GNDKWCouts$COUNT),cex = 1, pos = 3)
legend("topright", paste("No. of records =",
                   nrow(GN[GN$NationalID %in% GND, ])),
       bty = "n")

bp3 <- barplot(table(GNUKWCouts$COUNT),
               xlab = "Word count", ylab = "Frequency",
               main = "C", col = "#7570B3")
text(bp3, 0, table(GNUKWCouts$COUNT),cex = 1, pos = 3)
legend("topright", paste("No. of records =",
                   nrow(GN[!GN$NationalID %in% GND, ])),
       bty = "n")

```

**Fig. 6.** The keyword counts in the database fields considered for identification of probable duplicates for **A.**&nbsp;the entire `GN1000` dataset, **B.**&nbsp;the probable duplicate records alone and **C.**&nbsp;the unique records alone.


## Citing `PGRdup`
```{r, eval = FALSE}
citation("PGRdup")
```

```{r, echo = FALSE, collapse = TRUE}
detach("package:PGRdup", unload=TRUE)
suppressPackageStartupMessages(library(PGRdup))
cit <- citation("PGRdup")
# yr <- format(Sys.Date(), "%Y")
# cit[1]$year <- yr
# oc <- class(cit)
# 
# cit <- unclass(cit)
# attr(cit[[1]],"textVersion") <- gsub("\\(\\)",
#                                      paste("\\(", yr, "\\)", sep = ""),
#                                      attr(cit[[1]],"textVersion"))
# class(cit) <- oc
if (is.null(cit[1]$year)) {
  cit[1]$year <- format(Sys.Date(), "%Y")
  cit$textVersion <- gsub("()", 
                          paste("(", format(Sys.Date(), "%Y"), ")", sep =""), 
                          cit$textVersion, fixed = TRUE)
}
cit
```

## Session Info

```{r}
sessionInfo()
```

## References

```{r, echo=FALSE}
data.table::setDTthreads(threads_dt)
Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DisProbDup.R
\name{DisProbDup}
\alias{DisProbDup}
\title{Get disjoint probable duplicate sets}
\usage{
DisProbDup(pdup, combine = c("F", "P", "S"))
}
\arguments{
\item{pdup}{An object of class \code{ProbDup}.}

\item{combine}{A character vector indicating the type of sets to be
considered together for retrieving disjoint sets. If \code{NULL}, then
disjoint sets within each type are retrieved (see \strong{Details}).}
}
\value{
Returns an object of class \code{ProbDup} with either the disjoint
  sets within each type - \code{FuzzyDuplicates}, \code{PhoneticDuplicates}
  and \code{SemanticDuplicates} when \code{combine = NULL} or the combined
  disjoint duplicate sets as an additional element \code{DisjointDupicates}
  according to the choice specified in the argument \code{combine}.
}
\description{
\code{DisProbDup} finds and joins intersecting sets in an object of class
\code{ProbDup} to get disjoint probable duplicate sets.
}
\details{
This function considers the accession primary keys/IDs for finding
intersecting sets and subsequently joins them to retrieve disjoint sets.
These operations are implemented utilizing the
\code{\link[igraph:aaa-igraph-package]{igraph}} package functions.

Disjoint sets are retrieved either individually for each type of probable
duplicate sets or considering all type of sets simultaneously. In case of the
latter, the disjoint of all the type of sets alone are returned in the output
as an additional data frame \code{DisjointDuplicates} in an object of class
\code{ProbDup}
}
\examples{

\dontshow{
threads_dt <- data.table::getDTthreads()
threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
data.table::setDTthreads(2)

data.table::setDTthreads(2)
Sys.setenv(`OMP_THREAD_LIMIT` = 2)
}

\dontrun{

# Load PGR passport database
GN <- GN1000

# Specify as a vector the database fields to be used
GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")

# Clean the data
GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
c("Mota", "Company"))
y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
        "Bunch", "Peanut")
GN[GNfields] <- lapply(GN[GNfields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))

# Generate KWIC index
GNKWIC <- KWIC(GN, GNfields)

# Specify the exceptions as a vector
exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE",
         "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT",
         "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE",
         "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R",
         "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE",
         "U", "VALENCIA", "VIRGINIA", "WHITE")

# Specify the synsets as a list
syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))

# Fetch probable duplicate sets
GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, fuzzy = TRUE,
                 phonetic = TRUE, encoding = "primary",
                 semantic = TRUE, syn = syn)
lapply(GNdup, dim)

# Get disjoint probable duplicate sets of each kind
disGNdup1 <- DisProbDup(GNdup, combine = NULL)
lapply(disGNdup1, nrow)

# Get disjoint probable duplicate sets combining all the kinds of sets
disGNdup2 <- DisProbDup(GNdup, combine = c("F", "P", "S"))
lapply(disGNdup2, nrow)

}

\dontshow{
data.table::setDTthreads(threads_dt)
Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
}

}
\seealso{
\code{\link[PGRdup]{ProbDup}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ParseProbDup.R
\name{ParseProbDup}
\alias{ParseProbDup}
\title{Parse an object of class \code{ProbDup} to a data frame.}
\usage{
ParseProbDup(pdup, max.count = 30, insert.blanks = TRUE)
}
\arguments{
\item{pdup}{An object of class \code{ProbDup}.}

\item{max.count}{The maximum count of probable duplicate sets which are to be parsed to a data frame.}

\item{insert.blanks}{logical. If \code{TRUE}, inserts a row of \code{NAs} 
after each set.}
}
\value{
A data frame of the long/narrow form of the probable duplicate sets 
  data with the following core columns: \tabular{ll}{ 
  \code{SET_NO} \tab The set number. \cr \code{TYPE} \tab The type of 
  probable duplicate set. 'F' for fuzzy, 'P' for phonetic and 'S' for 
  semantic matching sets. \cr \code{K} \tab The KWIC index or database of 
  origin of the record. The \code{method} is specified within the square 
  brackets in the column name.  \cr \code{PRIM_ID} \tab The primary ID of the
  accession record from which the set could be identified. \cr \code{IDKW} 
  \tab The 'matching' keywords along with the IDs. \cr \code{COUNT} \tab The number of elements in a set. \cr } For the 
  retrieved columns(fields) the prefix \code{K*} indicates the KWIC index of 
  origin.
}
\description{
\code{ParseProbDup} converts an object of class \code{ProbDup} to a data frame for export.
}
\examples{

\dontshow{
threads_dt <- data.table::getDTthreads()
threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
data.table::setDTthreads(2)

data.table::setDTthreads(2)
Sys.setenv(`OMP_THREAD_LIMIT` = 2)
}

\dontrun{

#' # Load PGR passport database
GN <- GN1000

# Specify as a vector the database fields to be used
GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")

# Clean the data
GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
c("Mota", "Company"))
y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
        "Bunch", "Peanut")
GN[GNfields] <- lapply(GN[GNfields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))

# Generate KWIC index
GNKWIC <- KWIC(GN, GNfields)

# Specify the exceptions as a vector
exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE", 
         "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT", 
         "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE", 
         "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R", 
         "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE", 
         "U", "VALENCIA", "VIRGINIA", "WHITE")
          
# Specify the synsets as a list
syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))

# Fetch probable duplicate sets
GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, fuzzy = TRUE,
                 phonetic = TRUE, encoding = "primary", 
                 semantic = TRUE, syn = syn)
                 
# Convert to data frame of sets               
GNdupParsed <- ParseProbDup(GNdup)

}

\dontshow{
data.table::setDTthreads(threads_dt)
Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
}

}
\seealso{
\code{\link[PGRdup]{ProbDup}},
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/KWIC.R
\encoding{UTF-8}
\name{KWIC}
\alias{KWIC}
\title{Create a KWIC index}
\usage{
KWIC(x, fields, min.freq = 10)
}
\arguments{
\item{x}{A data frame from which KWIC index is to be generated.}

\item{fields}{A character vector with the names of fields(columns) of the
data frame from which KWIC index is to be generated. The first field is
considered as the primary key or identifier (see \strong{Details}).}

\item{min.freq}{Frequency of keywords are not computed if below
\code{min.freq}. Default is 10.}
}
\value{
A list of class \code{KWIC} containing the following components:
  \tabular{ll}{ \code{KWIC} \tab The KWIC index in the form of a data frame.
  \cr \code{KeywordFreq} \tab A data frame of the keywords detected with
  frequency greater than \code{min.freq}. \cr \code{Fields} \tab A character
  vector with the names of the PGR database fields from which the keywords
  were extracted. \cr }
}
\description{
\code{KWIC} creates a Keyword in Context index from PGR passport database
fields.
}
\details{
The function generates a Keyword in Context index from a data frame of a PGR
passport database based on the fields(columns) stated in the arguments, using
\code{\link[data.table]{data.table}} package.

The first element of vector \code{fields} is considered as the primary key or
identifier which uniquely identifies all rows in the data frame.

Cleaning of the data the input fields(columns) using the
\code{\link[PGRdup]{DataClean}} function with appropriate arguments is
suggested before running this function.
}
\examples{

\dontshow{
threads_dt <- data.table::getDTthreads()
threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
data.table::setDTthreads(2)

data.table::setDTthreads(2)
Sys.setenv(`OMP_THREAD_LIMIT` = 2)
}

# Load PGR passport database
GN <- GN1000

# Specify as a vector the database fields to be used
GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")

# Clean the data
GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))

\dontrun{

# Generate KWIC index
GNKWIC <- KWIC(GN, GNfields)
GNKWIC

# Retrieve the KWIC index from the KWIC object
KWIC <- GNKWIC[[1]]

# Retrieve the keyword frequencies from the KWIC object
KeywordFreq <- GNKWIC[[2]]

# Show error in case of duplicates and NULL values
# in the primary key/ID field "NationalID"
GN[1001:1005,] <- GN[1:5,]
GN[1001,3] <- ""
GNKWIC <- KWIC(GN, GNfields)
}

\dontshow{
data.table::setDTthreads(threads_dt)
Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
}

}
\references{
Kn\enc{ü}{ue}pffer, H. 1988. "The European Barley Database of the ECP/GR:
  An Introduction." \emph{Die Kulturpflanze} 36 (1): 135-62.
  \doi{10.1007/BF01986957}.

  Kn\enc{ü}{ue}pffer, H., L. Frese, and M. W. M. Jongen. 1997. "Using Central Crop
  Databases: Searching for Duplicates and Gaps." In \emph{Central Crop
  Databases: Tools for Plant Genetic Resources Management. Report of a
  Workshop, Budapest, Hungary, 13-16 October 1996}, edited by E. Lipman, M.
  W. M. Jongen, T. J. L. van Hintum, T. Gass, and L. Maggioni, 67-77. Rome,
  Italy and Wageningen, The Netherlands: International Plant Genetic
  Resources Institute and Centre for Genetic Resources.
}
\seealso{
\code{\link[data.table]{data.table}},
  \code{\link[PGRdup]{print.KWIC}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PGRdup-package.R
\docType{package}
\name{PGRdup-package}
\alias{PGRdup-package}
\alias{PGRdup}
\title{The PGRdup Package}
\description{
Functions to facilitate genebank managers in the identification of probable 
duplicate accessions from plant genetic resources (PGR) passport databases.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://cran.r-project.org/package=PGRdup}
  \item \url{https://github.com/aravind-j/PGRdup}
  \item \doi{10.5281/zenodo.841963}
  \item \url{https://aravind-j.github.io/PGRdup/}
  \item \url{https://www.rdocumentation.org/packages/PGRdup}
  \item Report bugs at \url{https://github.com/aravind-j/PGRdup/issues}
}

}
\author{
J Aravind \email{j.aravind@icar.org.in} \cr J Radhamani 
  \email{jalli.radhamani@icar.gov.in} \cr Kalyani Srinivasan 
  \email{kalyani.srinivasan@icar.gov.in} \cr B Ananda Subhash 
  \email{anandasubhash@gmail.com} \cr RK Tyagi
  \email{rishi.tyagi@icar.gov.in}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.genesys.R
\name{read.genesys}
\alias{read.genesys}
\title{Convert 'Darwin Core - Germplasm' zip archive to a flat file}
\usage{
read.genesys(zip.genesys, scrub.names.space = TRUE, readme = TRUE)
}
\arguments{
\item{zip.genesys}{A character vector giving the file path to the downloaded 
zip file from Genesys.}

\item{scrub.names.space}{logical. If \code{TRUE}, all space characters are 
removed from name field in names extension (see \strong{Details}).}

\item{readme}{logical. If \code{TRUE}, the genesys zip file readme is printed
to console.}
}
\value{
A data.frame with the flat file form of the genesys data.
}
\description{
\code{read.genesys} reads PGR data in a Darwin Core - germplasm zip 
archive downloaded from genesys database and creates a flat file
\code{data.frame} from it.
}
\details{
This function helps to import to R environment, the PGR data 
downloaded from genesys database \url{https://www.genesys-pgr.org/} as a 
Darwin Core - germplasm (DwC-germplasm) zip archive. The different csv files 
in the archive are merged as a flat file into a single \code{data.frame}.

All the space characters can be removed from the fields corresponding to 
accession names such as acceNumb, collNumb, ACCENAME, COLLNUMB, DONORNUMB and
OTHERNUMB using the argument \code{scrub.names.space} to facilitate creation 
of KWIC index with \code{\link[PGRdup]{KWIC}} function and subsequent
matching operations to identify probable duplicates with
\code{\link[PGRdup]{ProbDup}} function.

The argument \code{readme} can be used to print the readme file in the 
archive to console, if required.
}
\examples{

\dontshow{
threads_dt <- data.table::getDTthreads()
threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
data.table::setDTthreads(2)

data.table::setDTthreads(2)
Sys.setenv(`OMP_THREAD_LIMIT` = 2)
}

\dontrun{
# Import the DwC-Germplasm zip archive "genesys-accessions-filtered.zip"
PGRgenesys <- read.genesys("genesys-accessions-filtered.zip",
                           scrub.names.space = TRUE, readme = TRUE)
}

\dontshow{
data.table::setDTthreads(threads_dt)
Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
}

}
\seealso{
\code{\link[data.table]{data.table}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MergeKW.R
\name{MergeKW}
\alias{MergeKW}
\alias{MergePrefix}
\alias{MergeSuffix}
\title{Merge keyword strings}
\usage{
MergeKW(x, y, delim = c("space", "dash", "period"))

MergePrefix(x, y, delim = c("space", "dash", "period"))

MergeSuffix(x, y, delim = c("space", "dash", "period"))
}
\arguments{
\item{x}{A character vector. If not, coerced to character by 
\code{as.character}.}

\item{y}{A list of character vectors with pairs of strings that are to be 
merged (for \code{MergeKW}) or a character vector of strings which are to 
be merged to succeeding string (for \code{MergePrefix}) or the preceding 
string (for \code{MergeSuffix}). If not of type character, coerced by 
\code{as.character}.}

\item{delim}{Delimiting characters to be removed between keywords.}
}
\value{
A character vector of the same length as \code{x} with the required 
  keyword strings merged.
}
\description{
These functions merge keyword strings separated by delimiters such as space, 
period or dash in a character vector into single keyword strings.
}
\details{
These functions aid in standardization of relevant data fields(columns) in 
PGR passport data for creation of a KWIC index with 
\code{\link[PGRdup]{KWIC}} function and subsequent identification of probable
duplicate accessions by the \code{\link[PGRdup]{ProbDup}} function.

It is recommended to run this function before using the 
\code{\link[PGRdup]{DataClean}} function on the relevant data fields(columns)
of PGR passport databases.

\code{MergeKW} merges together pairs of strings specified as a list in 
argument \code{y} wherever they exist in a character vector. The second 
string in the pair is merged even when it is followed by a number.

\code{MergePrefix} merges prefix strings specified as a character vector in 
argument \code{y} to the succeeding root word, wherever they exist in a 
character vector.

\code{MergeSuffix} merges suffix strings specified as a character vector in 
argument \code{y} to the preceding root word, wherever they exist in a 
character vector. The suffix strings which are followed by numbers are also 
merged.
}
\examples{
names <- c("Punjab Bold", "Gujarat- Dwarf", "Nagpur.local", "SAM COL 144",
           "SAM COL--280", "NIZAMABAD-LOCAL", "Dark Green Mutant",
           "Dixie-Giant", "Georgia- Bunch", "Uganda-erect", "Small Japan",
           "Castle  Cary", "Punjab erect", "Improved small japan",
           "Dark Purple")

# Merge pairs of strings
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
           c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
           c("Mota", "Company"))
names <- MergeKW(names, y1, delim = c("space", "dash", "period"))

# Merge prefix strings
y2 <- c("Light", "Small", "Improved", "Punjab", "SAM")
names <- MergePrefix(names, y2, delim = c("space", "dash", "period"))

# Merge suffix strings
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
        "Bunch", "Peanut")
names <- MergeSuffix(names, y3, delim = c("space", "dash", "period"))
}
\seealso{
\code{\link[PGRdup]{DataClean}}, \code{\link[PGRdup]{KWIC}}, 
  \code{\link[PGRdup]{ProbDup}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GN1000.R
\name{GN1000}
\alias{GN1000}
\title{Sample groundnut PGR passport data}
\format{
A data frame having 1000 records with the following 10
  columns(fields): \describe{ \item{CommonName}{Common name} 
  \item{BotanicalName}{Botanical name} \item{NationalID}{NBPGR National 
  identifier} \item{CollNo}{Collector number} \item{DonorID}{Donor ID} 
  \item{OtherID1}{Other ID field 1} \item{OtherID2}{Other ID field 2} 
  \item{BioStatus}{Biological status} \item{SourceCountry}{Country of 
  origin} \item{TransferYear}{Year of transfer} }
}
\usage{
data(GN1000)
}
\description{
Sample PGR passport data of 1000 groundnut accessions held in the Indian
National Genebank at National Bureau of Plant Genetic Resources (NBPGR), New
Delhi.
}
\seealso{
\url{http://pgrportal.nbpgr.ernet.in/}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AddProbDup.R
\name{AddProbDup}
\alias{AddProbDup}
\title{Add probable duplicate sets fields to the PGR passport database}
\usage{
AddProbDup(pdup, db, addto = c("I", "II"), max.count = 30)
}
\arguments{
\item{pdup}{An object of class \code{ProbDup}.}

\item{db}{A data frame of the PGR passport database.}

\item{addto}{Either \code{"I"} or \code{"II"} indicating the database to 
which the data.fields are to be added (see \strong{Details}).}

\item{max.count}{The maximum count of probable duplicate sets whose 
information is to be retrieved.}
}
\value{
A data frame of the PGR passport database with the probable duplicate
  sets fields added.
}
\description{
\code{AddProbDup} adds the fuzzy, phonetic and semantic probable duplicates 
sets data fields from an object of class \code{ProbDup} to the original PGR 
passport database.
}
\details{
This function helps to add information associated with identified fuzzy, 
phonetic and semantic probable duplicate sets using the 
\code{\link[PGRdup]{ProbDup}} function to the original PGR passport database.
Associated data fields such as \code{SET_NO}, \code{ID} and \code{IDKW} are 
added based on the \code{PRIM_ID} field(column).

In case more than one KWIC index was used to generate the object of class 
\code{ProbDup}, the argument \code{addto} can be used to specify to which 
database the data fields are to be added. The default \code{"I"} indicates 
the database from which the first KWIC index was created and \code{"II"} 
indicates the database from which the second index was created.
}
\note{
When any primary ID/key records in the fuzzy, phonetic or semantic 
  duplicate sets are found to be missing from the original database 
  \code{db}, then they are ignored and only the matching records are 
  considered for adding the information with a warning.
  
  This may be due to data standardization of the primary ID/key field using 
  the function \code{\link[PGRdup]{DataClean}} before creation of the KWIC 
  index and subsequent identification of probable duplicate sets. In such a 
  case, it is recommended to use an identical data standardization operation 
  on the database \code{db} before running this function.
}
\examples{

\dontshow{
threads_dt <- data.table::getDTthreads()
threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
data.table::setDTthreads(2)

data.table::setDTthreads(2)
Sys.setenv(`OMP_THREAD_LIMIT` = 2)
}

\dontrun{

#' # Load PGR passport database
GN <- GN1000

# Specify as a vector the database fields to be used
GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")

# Clean the data
GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
c("Mota", "Company"))
y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
        "Bunch", "Peanut")
GN[GNfields] <- lapply(GN[GNfields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))

# Generate KWIC index
GNKWIC <- KWIC(GN, GNfields)

# Specify the exceptions as a vector
exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE", 
         "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT", 
         "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE", 
         "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R", 
         "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE", 
         "U", "VALENCIA", "VIRGINIA", "WHITE")
          
# Specify the synsets as a list
syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))

# Fetch probable duplicate sets
GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, fuzzy = TRUE,
                 phonetic = TRUE, encoding = "primary", 
                 semantic = TRUE, syn = syn)

# Add the duplicates sets to the original database                 
GNwithdup <-  AddProbDup(pdup = GNdup, db = GN1000, addto = "I")                  

}

\dontshow{
data.table::setDTthreads(threads_dt)
Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
}

}
\seealso{
\code{\link[PGRdup]{DataClean}}, \code{\link[PGRdup]{KWIC}},
  \code{\link[PGRdup]{ProbDup}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.KWIC.R
\name{print.KWIC}
\alias{print.KWIC}
\title{Prints summary of \code{KWIC} object.}
\usage{
\method{print}{KWIC}(x, ...)
}
\arguments{
\item{x}{An object of class \code{KWIC}.}

\item{...}{Unused}
}
\description{
\code{print.KWIC} prints to console the summary of an object of class 
\code{KWIC} including the database fields(columns) used, the total number of
keywords and the number of distinct keywords in the index.
}
\seealso{
\code{\link[PGRdup]{KWIC}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataClean.R
\name{DataClean}
\alias{DataClean}
\title{Clean PGR passport data}
\usage{
DataClean(
  x,
  fix.comma = TRUE,
  fix.semcol = TRUE,
  fix.col = TRUE,
  fix.bracket = TRUE,
  fix.punct = TRUE,
  fix.space = TRUE,
  fix.sep = TRUE,
  fix.leadzero = TRUE
)
}
\arguments{
\item{x}{A character vector. If not, coerced to character by 
\code{as.character}.}

\item{fix.comma}{logical. If \code{TRUE}, all the commas are replaced by 
space (see \strong{Details}).}

\item{fix.semcol}{logical. If \code{TRUE}, all the semicolons are replaced by
space (see \strong{Details}).}

\item{fix.col}{logical. If \code{TRUE}, all the colons are replaced by space 
(see \strong{Details}).}

\item{fix.bracket}{logical. If \code{TRUE}, all the brackets are replaced by 
space (see \strong{Details}).}

\item{fix.punct}{logical. If \code{TRUE}, all punctuation characters are 
removed (see \strong{Details}).}

\item{fix.space}{logical. If \code{TRUE}, all space characters are replaced 
by space and multiple spaces are converted to single space (see 
\strong{Details}).}

\item{fix.sep}{logical. If \code{TRUE}, space between alphabetic characters 
followed by digits is removed (see \strong{Details}).}

\item{fix.leadzero}{logical. If \code{TRUE}, leading zeros are removed (see 
\strong{Details}).}
}
\value{
A character vector with the cleaned data converted to upper case. 
  \code{NAs} if any are converted to blank strings.
}
\description{
\code{DataClean} cleans the data in a character vector according to the 
conditions in the arguments.
}
\details{
This function aids in standardization and preparation of the PGR passport 
data for creation of a KWIC index with \code{\link[PGRdup]{KWIC}} function 
and the identification of probable duplicate accessions by the 
\code{\link[PGRdup]{ProbDup}} function. It cleans the character strings in 
passport data fields(columns) specified as the input character vector 
\code{x} according to the conditions in the arguments in the same order. If 
the input vector \code{x} is not of type character, it is coerced to a 
character vector.

This function is designed particularly for use with fields corresponding to 
accession names such as accession ids, collection numbers, accession names 
etc. It is essentially a wrapper around the \code{\link[base:grep]{gsub}} base 
function with \code{\link[base]{regex}} arguments. It also converts all 
strings to upper case and removes leading and trailing spaces.

Commas, semicolons and colons which are sometimes used to separate multiple 
strings or names within the same field can be replaced with a single space 
using the logical arguments \code{fix.comma}, \code{fix.semcol} and 
\code{fix.col} respectively.

Similarly the logical argument \code{fix.bracket} can be used to replace all 
brackets including parenthesis, square brackets and curly brackets with
space.

The logical argument \code{fix.punct} can be used to remove all punctuation 
from the data.

\code{fix.space} can be used to convert all space characters such as tab, 
newline, vertical tab, form feed and carriage return to spaces and finally 
convert multiple spaces to single space.

\code{fix.sep} can be used to merge together accession identifiers 
composed of alphabetic characters separated from as series of digits by a 
space character. For example IR 64, PUSA 256 etc.

\code{fix.leadzero} can be used to remove leading zeros from accession name 
fields to facilitate matching to identify probable duplicates. e.g. IR0064 -> 
IR64
}
\examples{
names <- c("S7-12-6", "ICG-3505", "U 4-47-18;EC 21127", "AH 6481", "RS   1",
           "AK 12-24", "2-5 (NRCG-4053)", "T78, Mwitunde", "ICG 3410",
           "#648-4 (Gwalior)", "TG4;U/4/47/13", "EC0021003")
DataClean(names)
}
\seealso{
\code{\link[base:grep]{gsub}}, \code{\link[base]{regex}}, 
  \code{\link[PGRdup]{MergeKW}}, \code{\link[PGRdup]{KWIC}}, 
  \code{\link[PGRdup]{ProbDup}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.ProbDup.R
\name{print.ProbDup}
\alias{print.ProbDup}
\title{Prints summary of \code{ProbDup} object.}
\usage{
\method{print}{ProbDup}(x, ...)
}
\arguments{
\item{x}{An object of class \code{ProbDup}.}

\item{...}{Unused}
}
\description{
\code{print.ProbDup} prints to console the summary of an object of class 
\code{ProbDup} including the method used ("a", "b" or "c"), the database
fields(columns) considered, the number of probable duplicate sets of each
kind along with the corresponding number of records.
}
\seealso{
\code{\link[PGRdup]{ProbDup}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ValidatePrimKey.R
\name{ValidatePrimKey}
\alias{ValidatePrimKey}
\title{Validate if a data frame column confirms to primary key/ID constraints}
\usage{
ValidatePrimKey(x, prim.key)
}
\arguments{
\item{x}{A data frame.}

\item{prim.key}{A character vector indicating the name of the data frame 
column to be validated for primary key/ID constraints (see 
\strong{Details}).}
}
\value{
A list with containing the following components: \tabular{ll}{ 
  \code{message1} \tab Indicates whether duplicated values were encountered 
  in \code{prim.key} field(column) of data frame \code{x} or not. \cr 
  \code{Duplicates} \tab A data frame of the records with duplicated prim.key
  values if they were encountered. \cr \code{message2} \tab Indicates whether
  NULL values were encountered in \code{prim.key} field(column) of data frame
  \code{x} or not. \cr \code{NullRecords} \tab A data frame of the records 
  with NULL prim.key values if they were encountered. \cr }
}
\description{
\code{ValidatePrimKey} checks if a column in a data frame confirms to the 
primary key/ID constraints of absence of duplicates and NULL values. Aberrant
records if encountered are returned in the output list.
}
\details{
The function checks whether a field(column) in a data frame of PGR passport 
database confirms to the primary key/ID constraints of absence of duplicates 
and NULL values. If records with nonconforming values in the column are 
encountered, they are returned in the output list for rectification.

If multiple fields(columns) are given as a character vector in 
\code{prim.key} field, only the first element will be considered as the 
primary key/ID field(column).

Cleaning of the data in the input field(column) using the 
\code{\link[PGRdup]{DataClean}} function with appropriate arguments is 
suggested before running this function.

It is recommended to run this function and rectify aberrant records in a PGR 
passport database before creating a KWIC index using the 
\code{\link[PGRdup]{KWIC}} function.
}
\examples{
GN <- GN1000
ValidatePrimKey(x=GN, prim.key="NationalID")
\dontrun{
# Show error in case of duplicates and NULL values 
# in the primary key/ID field "NationalID"
GN[1001:1005,] <- GN[1:5,]
GN[1001,3] <- ""
ValidatePrimKey(x=GN, prim.key="NationalID")}
}
\seealso{
\code{\link[PGRdup]{DataClean}}, \code{\link[PGRdup]{KWIC}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/KWCounts.R
\name{KWCounts}
\alias{KWCounts}
\title{Generate keyword counts}
\usage{
KWCounts(x, fields, excep)
}
\arguments{
\item{x}{A data frame.}

\item{fields}{A character vector with the names of fields(columns) of the 
data frame from which KWIC index is to be generated. The first field is 
considered as the primary key or identifier (see \strong{Details}).}

\item{excep}{A vector of the keywords not to be considered for the counts
(see \strong{Details}).}
}
\value{
A data frame with the keyword counts for each record.
}
\description{
\code{KWCounts} generates keyword counts from PGR passport database 
fields(columns).
}
\details{
This function computes the keyword counts from PGR passport database 
fields(columns) specified in the \code{fields} argument. The first field is 
considered as the primary key or identifier and is not used for counting the 
keywords. Any strings given in the \code{excep} argument are ignored for 
generating the counts.

The keyword counts can give a rough indication of the completeness of the 
data in the database fields being used for identification of probable 
duplicates.
}
\note{
For large number of exceptions and/or large data.frame computation of
keyword counts may take some time.
}
\examples{

\dontshow{
threads_dt <- data.table::getDTthreads()
threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
data.table::setDTthreads(2)

data.table::setDTthreads(2)
Sys.setenv(`OMP_THREAD_LIMIT` = 2)
}

# Load PGR passport database
GN <- GN1000

# Specify database fields to use as a vector
GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")

# Specify the exceptions as a vector
exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE", 
         "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT", 
         "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE", 
         "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R", 
         "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE", 
         "U", "VALENCIA", "VIRGINIA", "WHITE")
         
# Compute the keyword counts
GNKWCouts <- KWCounts(GN, GNfields, exep)

# Plot the keyword counts
bp <- barplot(table(GNKWCouts$COUNT),
              xlab = "Word count", ylab = "Frequency", col = "#CD5555")
text(bp, 0, table(GNKWCouts$COUNT),cex=1,pos=3)

\dontshow{
data.table::setDTthreads(threads_dt)
Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
}

}
\seealso{
\code{\link[stringi]{stri_count}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DoubleMetaphone.R
\encoding{UTF-8}
\name{DoubleMetaphone}
\alias{DoubleMetaphone}
\title{'Double Metaphone' phonetic algorithm}
\usage{
DoubleMetaphone(str)
}
\arguments{
\item{str}{A character vector whose strings are to be encoded by double
metaphone algorithm.}
}
\value{
Returns a list with two character vectors of the same length as the
  input vector. The first character vector contains the primary double
  metaphone encodings, while the second character vector contains the
  alternate encodings.
}
\description{
\code{DoubleMetaphone} converts strings to double metaphone phonetic codes.
}
\details{
An implementation of the Double Metaphone phonetic algorithm in \code{R}. If
non-ASCII characters encountered in the input character vector \code{str}, a
warning is issued and they are transliterated so that the accented characters
are converted to their ASCII unaccented versions.
}
\note{
In case of non-ASCII characters in strings, a warning is issued and
  accented characters are converted to their ASCII unaccented versions.
}
\section{Acknowledgement}{
 The \code{C} code for the double metaphone
  algorithm was adapted from Maurice Aubrey's perl module hosted at the
  \strong{gitpan/Text-DoubleMetaphone}
  \href{https://github.com/gitpan/Text-DoubleMetaphone/blob/master/double_metaphone.c}{public
   github library} along with the corresponding
  \href{https://github.com/gitpan/Text-DoubleMetaphone/blob/master/README}{license
   information}.
}

\examples{
# Return the primary and secondary Double Metaphone encodings for a character vector.
str1 <- c("Jyothi", "Jyoti")
str2 <- c("POLLACHI", "BOLLACHI")
DoubleMetaphone(str1)
DoubleMetaphone(str2)
\dontrun{
# Issue a warning in case of non-ASCII characters.
str3 <- c("J\xf5geva", "Jogeva")
DoubleMetaphone(str3) }
}
\references{
Philips, Lawrence. 2000. "The Double Metaphone Search Algorithm."
  \emph{C/C++ Users Journal} 18 (6): 38-43.
  \url{https://dl.acm.org/doi/10.5555/349124.349132}.
}
\seealso{
\code{\link[stringdist]{phonetic}},
  \href{https://cran.r-project.org/package=RecordLinkage}{\code{phonetics}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ProbDup.R
\encoding{UTF-8}
\name{ProbDup}
\alias{ProbDup}
\title{Identify probable duplicates of accessions}
\usage{
ProbDup(
  kwic1,
  kwic2 = NULL,
  method = c("a", "b", "c"),
  excep = NULL,
  chunksize = 1000,
  useBytes = TRUE,
  fuzzy = TRUE,
  max.dist = 3,
  force.exact = TRUE,
  max.alpha = 4,
  max.digit = Inf,
  phonetic = TRUE,
  encoding = c("primary", "alternate"),
  phon.min.alpha = 5,
  min.enc = 3,
  semantic = FALSE,
  syn = NULL
)
}
\arguments{
\item{kwic1}{An object of class \code{KWIC}.}

\item{kwic2}{An object of class \code{KWIC}. Required for \code{method}
\code{"b"} and \code{"c"} only (see \strong{Details}).}

\item{method}{The method to be followed for identification of probable
duplicates. Either \code{"a"}, \code{"b"} or \code{"c"}. (see
\strong{Details}).}

\item{excep}{A vector of the keywords in KWIC not to be used for probable
duplicate search (see \strong{Details}).}

\item{chunksize}{A value indicating the size of KWIC index keyword block to
be used for searching for matches at a time in case of large number of
keywords(see \strong{Note}).}

\item{useBytes}{logical. If \code{TRUE}, performs byte-wise comparison
instead of character-wise comparison (see \strong{Note}).}

\item{fuzzy}{logical. If \code{TRUE} identifies probable duplicates based on
fuzzy matching.}

\item{max.dist}{The maximum levenshtein distance between keyword strings
allowed for a match. Default is 3 (see \strong{Details}).}

\item{force.exact}{logical. If \code{TRUE}, enforces exact matching instead
of fuzzy matching for keyword strings which match the criteria specified in
arguments \code{max.alpha} and \code{max.digit} (see \strong{Details}).}

\item{max.alpha}{Maximum number of alphabet characters present in a keyword
string up to which exact matching is enforced rather than fuzzy matching.
Default is 4 (see \strong{Details}).}

\item{max.digit}{Maximum number of numeric characters present in a keyword
string up to which exact matching is enforced rather than fuzzy matching.
Default is Inf (see \strong{Details}).}

\item{phonetic}{logical. If \code{TRUE} identifies probable duplicates based
on phonetic matching.}

\item{encoding}{Double metaphone encoding for phonetic matching. The default
is \code{"primary"} (see \strong{Details}).}

\item{phon.min.alpha}{Minimum number of alphabet characters to be present in
a keyword string for phonetic matching (see \strong{Details}).}

\item{min.enc}{Minimum number of characters to be be present in double
metaphone encoding of a keyword string for phonetic matching (see
\strong{Details}).}

\item{semantic}{logical. If \code{TRUE} identifies probable duplicates based
on semantic matching.}

\item{syn}{A list with character vectors of synsets (see \strong{Details}).}
}
\value{
A list of class \code{ProbDup} containing the following data frames
  of probable duplicate sets identified along with the corresponding keywords
  and set counts: \enumerate{ \item \code{FuzzyDuplicates} \item
  \code{PhoneticDuplicates} \item \code{SemanticDuplicates} } Each data frame
  has the following columns: \tabular{ll}{ \code{SET_NO} \tab The set number.
  \cr \code{TYPE} \tab The type of probable duplicate set. 'F' for fuzzy, 'P'
  for phonetic and 'S' for semantic matching sets. \cr \code{ID} \tab The
  primary IDs of records of accessions comprising a set. \cr \code{ID:KW}
  \tab The 'matching' keywords along with the IDs. \cr \code{COUNT} \tab The
  number of elements in a set. \cr }

  The prefix \code{[K*]} indicates the KWIC index of origin of the KEYWORD or
  PRIM_ID.
}
\description{
\code{ProbDup} identifies probable duplicates of germplasm accessions in KWIC
indexes created from PGR passport databases using fuzzy, phonetic and
semantic matching strategies.
}
\details{
This function performs fuzzy, phonetic and semantic matching of keywords in
KWIC indexes of PGR passport databases (created using
\code{\link[PGRdup]{KWIC}} function) to identify probable duplicates of
germplasm accessions. The function can execute matching according to either
of the following three methods as specified by the \code{method} argument.

\describe{ \item{Method \code{a}:}{Perform string matching of keywords in a
single KWIC index to identify probable duplicates of accessions in a single
PGR passport database.} \item{Method \code{b}:}{Perform string matching of
keywords in the first KWIC index (query) with that of the keywords in the
second index (source) to identify probable duplicates of accessions of the
first PGR passport database among the accessions in the second database.}
\item{Method \code{c}:}{Perform string matching of keywords in two different
KWIC indexes jointly to identify probable duplicates of accessions from among
two PGR passport databases.}}

\strong{Fuzzy matching} or approximate string matching of keywords is carried
out by computing the generalized levenshtein (edit) distance between them.
This distance measure  counts the number of deletions, insertions and
substitutions necessary to turn one string to the another. A distance of up
to \code{max.dist} are considered for a match.

Exact matching will be enforced when the argument \code{force.exact} is
\code{TRUE}. It can be used to avoid fuzzy matching when the number of
alphabet characters in keywords is lesser than a critical value
(\code{max.alpha}). Similarly, the value of \code{max.digit} can also be set
according to the requirements. The default value of \code{Inf} avoids fuzzy
matching and enforces exact matching for all keywords having any numerical
characters. If \code{max.digit} and \code{max.alpha} are both set to
\code{Inf}, exact matching will be enforced for all the keywords.

When exact matching is enforced, for keywords having both alphabet and
numeric characters and with the number of alphabet characters greater than
\code{max.digit}, matching will be carried out separately for alphabet and
numeric characters present.

\strong{Phonetic matching} of keywords is carried out using the Double
Metaphone phonetic algorithm (\code{\link[PGRdup]{DoubleMetaphone}}) to
identify keywords that have the similar pronunciation. Either the
\code{primary} or \code{alternate} encodings can be used by specifying the
\code{encoding} argument. The argument \code{phon.min.alpha} sets the limits
for the number of alphabet characters to be present in a string for executing
phonetic matching. Similarly \code{min.enc} sets the limits for the number of
characters to be present in the encoding of a keyword for phonetic matching.

\strong{Semantic matching} matches keywords based on a list of accession name
synonyms supplied as list with character vectors of synonym sets (synsets) to
the \code{syn} argument. Synonyms in this context refers to interchangeable
identifiers or names by which an accession is recognized. Multiple keywords
specified as members of the same synset in \code{syn} are merged together. To
facilitate accurate identification of synonyms from the KWIC index, identical
data standardization operations using the \code{\link[PGRdup]{MergeKW}} and
\code{\link[PGRdup]{DataClean}} functions for both the original database
fields and the synset list are recommended.

The probable duplicate sets identified initially here may be intersecting
with other sets. To get the disjoint sets after the union of all the
intersecting sets use the \code{\link[PGRdup]{DisProbDup}} function.

The function \code{\link[PGRdup]{AddProbDup}} can be used to add the
information associated with the identified sets in an object of class
\code{ProbDup} as fields(columns) to the original PGR passport database.

All of the string matching operations here are executed through the
\code{\link[stringdist]{stringdist-package}} functions.
}
\note{
As the number of keywords in the KWIC indexes increases, the memory
  consumption by the function also increases. For string matching, this
  function relies upon creation of a \eqn{n}*\eqn{m} matrix of all possible
  keyword pairs for comparison, where \eqn{n} and \eqn{m} are the number of
  keywords in the query and source indexes respectively. This can lead to
  \code{cannot allocate vector of size} errors in case very large KWIC
  indexes where the comparison matrix is too large to reside in memory. In
  such a case, try to adjust the \code{chunksize} argument to get the
  appropriate size of the KWIC index keyword block to be used for searching
  for matches at a time. However a smaller chunksize may lead to longer
  computation time due to the memory-time trade-off.

  The progress of matching is displayed in the console as number of blocks
  completed out of total (e.g. 6 / 30), the percentage of achievement (e.g.
  30\%) and a text-based progress bar.

  In case of multi-byte characters in keywords, the matching speed is further
  dependent upon the \code{useBytes} argument as described in
  \strong{Encoding issues} for the \code{\link[stringdist]{stringdist}}
  function, which is made use of here for string matching.
}
\examples{

\dontshow{
threads_dt <- data.table::getDTthreads()
threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
data.table::setDTthreads(2)

data.table::setDTthreads(2)
Sys.setenv(`OMP_THREAD_LIMIT` = 2)
}

\dontrun{

# Method "a"
#===========

# Load PGR passport database
GN <- GN1000

# Specify as a vector the database fields to be used
GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")

# Clean the data
GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
c("Mota", "Company"))
y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
        "Bunch", "Peanut")
GN[GNfields] <- lapply(GN[GNfields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))

# Generate KWIC index
GNKWIC <- KWIC(GN, GNfields)

# Specify the exceptions as a vector
exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE",
         "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT",
         "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE",
         "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R",
         "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE",
         "U", "VALENCIA", "VIRGINIA", "WHITE")

# Specify the synsets as a list
syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))

# Fetch probable duplicate sets
GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, fuzzy = TRUE,
                 phonetic = TRUE, encoding = "primary",
                 semantic = TRUE, syn = syn)
GNdup

# Method "b and c"
#=================

# Load PGR passport databases
GN1 <- GN1000[!grepl("^ICG", GN1000$DonorID), ]
GN1$DonorID <- NULL
GN2 <- GN1000[grepl("^ICG", GN1000$DonorID), ]
GN2 <- GN2[!grepl("S", GN2$DonorID), ]
GN2$NationalID <- NULL

# Specify as a vector the database fields to be used
GN1fields <- c("NationalID", "CollNo", "OtherID1", "OtherID2")
GN2fields <- c("DonorID", "CollNo", "OtherID1", "OtherID2")

# Clean the data
GN1[GN1fields] <- lapply(GN1[GN1fields], function(x) DataClean(x))
GN2[GN2fields] <- lapply(GN2[GN2fields], function(x) DataClean(x))
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
c("Mota", "Company"))
y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
        "Bunch", "Peanut")
GN1[GN1fields] <- lapply(GN1[GN1fields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN1[GN1fields] <- lapply(GN1[GN1fields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN1[GN1fields] <- lapply(GN1[GN1fields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))
GN2[GN2fields] <- lapply(GN2[GN2fields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN2[GN2fields] <- lapply(GN2[GN2fields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN2[GN2fields] <- lapply(GN2[GN2fields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))

# Remove duplicated DonorID records in GN2
GN2 <- GN2[!duplicated(GN2$DonorID), ]

# Generate KWIC index
GN1KWIC <- KWIC(GN1, GN1fields)
GN2KWIC <- KWIC(GN2, GN2fields)

# Specify the exceptions as a vector
exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE",
         "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT",
         "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE",
         "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R",
         "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE",
         "U", "VALENCIA", "VIRGINIA", "WHITE")

# Specify the synsets as a list
syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))

# Fetch probable duplicate sets
GNdupb <- ProbDup(kwic1 = GN1KWIC, kwic2 = GN2KWIC, method = "b",
                  excep = exep, fuzzy = TRUE, phonetic = TRUE,
                  encoding = "primary", semantic = TRUE, syn = syn)
GNdupb

GNdupc <- ProbDup(kwic1 = GN1KWIC, kwic2 = GN2KWIC, method = "c",
                  excep = exep, fuzzy = TRUE, phonetic = TRUE,
                  encoding = "primary", semantic = TRUE, syn = syn)
GNdupc

}

\dontshow{
data.table::setDTthreads(threads_dt)
Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
}

}
\references{
van der Loo, M. P. J. 2014. "The Stringdist Package for
  Approximate String Matching." \emph{R Journal} 6 (1):111-22.
  \url{https://journal.r-project.org/articles/RJ-2014-011/index.html}.
}
\seealso{
\code{\link[PGRdup]{KWIC}}, \code{\link[PGRdup]{DoubleMetaphone}}
  \code{\link[stringdist:stringdist]{stringdistmatrix}},
  \code{\link[utils]{adist}}, \code{\link[PGRdup]{print.ProbDup}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ReconstructProbDup.R
\name{ReconstructProbDup}
\alias{ReconstructProbDup}
\title{Reconstruct an object of class ProbDup}
\usage{
ReconstructProbDup(rev)
}
\arguments{
\item{rev}{A data frame with the the core columns(fields) \code{SET_NO}, 
\code{TYPE}, \code{K}, \code{PRIM_ID}, \code{DEL}, \code{SPLIT}, 
\code{COUNT} and \code{IDKW}}
}
\value{
An object of class \code{ProbDup} with the modified fuzzy,
  phonetic and semantic probable duplicate sets according to the instructions
  specified under clerical review.
}
\description{
\code{ReconstructProbDup} reconstructs a data frame of probable duplicate 
sets created using the function \code{ReviewProbDup} and subjected to manual 
clerical review, back into an object of class \code{ProbDup}.
}
\details{
A data frame created using the function \code{\link[PGRdup]{ReviewProbDup}} 
from an object of class \code{ProbDup} for manual clerical review of 
identified probable duplicate sets can be reconstituted back to the same 
object after the review using this function. The instructions for modifying 
the sets entered in the appropriate format in the columns \code{DEL} and 
\code{SPLIT} during clerical review are taken into account for reconstituting
the probable duplicate sets.

Any records with \code{Y} in column \code{DEL} are deleted and records with 
identical integers in the column \code{SPLIT} other than the default \code{0}
are reassembled into a new set.
}
\examples{

\dontshow{
threads_dt <- data.table::getDTthreads()
threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
data.table::setDTthreads(2)

data.table::setDTthreads(2)
Sys.setenv(`OMP_THREAD_LIMIT` = 2)
}

\dontrun{

# Load PGR passport database
GN <- GN1000

# Specify as a vector the database fields to be used
GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")

# Clean the data
GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
c("Mota", "Company"))
y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
        "Bunch", "Peanut")
GN[GNfields] <- lapply(GN[GNfields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))

# Generate KWIC index
GNKWIC <- KWIC(GN, GNfields)

# Specify the exceptions as a vector
exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE", 
         "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT", 
         "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE", 
         "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R", 
         "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE", 
         "U", "VALENCIA", "VIRGINIA", "WHITE")
          
# Specify the synsets as a list
syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))

# Fetch probable duplicate sets
GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, fuzzy = TRUE,
                 phonetic = TRUE, encoding = "primary", 
                 semantic = TRUE, syn = syn)

# Get disjoint probable duplicate sets of each kind
disGNdup <- DisProbDup(GNdup, combine = NULL)

# Get the data frame for reviewing the duplicate sets identified
RevGNdup <- ReviewProbDup(pdup = disGNdup, db1 = GN1000,
                          extra.db1 = c("SourceCountry", "TransferYear"), 
                          max.count = 30, insert.blanks = TRUE)
# Examine and review the duplicate sets using edit function
RevGNdup <- edit(RevGNdup)

# Examine and make changes to a set
subset(RevGNdup, SET_NO==12 & TYPE=="P", select= c(IDKW, DEL, SPLIT))
RevGNdup[c(110, 112, 114, 118, 121, 122, 124), 6] <- "Y"
RevGNdup[c(111, 115, 128), 7] <- 1
RevGNdup[c(113, 117, 120), 7] <- 2
RevGNdup[c(116, 119), 7] <- 3
RevGNdup[c(123, 125), 7] <- 4
RevGNdup[c(126, 127), 7] <- 5
subset(RevGNdup, SET_NO==12 & TYPE=="P", select= c(IDKW, DEL, SPLIT))

# Reconstruct ProDup object
GNdup2 <- ReconstructProbDup(RevGNdup)
lapply(disGNdup, nrow)
lapply(GNdup2, nrow)

}

\dontshow{
data.table::setDTthreads(threads_dt)
Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
}

}
\seealso{
\code{\link[PGRdup]{ProbDup}}, \code{\link[PGRdup]{ReviewProbDup}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ReviewProbDup.R
\name{ReviewProbDup}
\alias{ReviewProbDup}
\title{Retrieve probable duplicate set information from PGR passport database for
review}
\usage{
ReviewProbDup(
  pdup,
  db1,
  db2 = NULL,
  extra.db1 = NULL,
  extra.db2 = NULL,
  max.count = 30,
  insert.blanks = TRUE
)
}
\arguments{
\item{pdup}{An object of class \code{ProbDup}.}

\item{db1}{A data frame of the PGR passport database.}

\item{db2}{A data frame of the PGR passport database. Required when
\code{pdup} was created using more than one KWIC Index.}

\item{extra.db1}{A character vector of extra \code{db1} column names to be
retrieved.}

\item{extra.db2}{A character vector of extra \code{db2} column names to be
retrieved.}

\item{max.count}{The maximum count of probable duplicate sets whose
information is to be retrieved.}

\item{insert.blanks}{logical. If \code{TRUE}, inserts a row of \code{NAs}
after each set.}
}
\value{
A data frame of the long/narrow form of the probable duplicate sets
  data along with associated fields from the original database(s). The core
  columns in the resulting data frame are as follows: \tabular{ll}{
  \code{SET_NO} \tab The set number. \cr \code{TYPE} \tab The type of
  probable duplicate set. 'F' for fuzzy, 'P' for phonetic and 'S' for
  semantic matching sets. \cr \code{K[*]} \tab The KWIC index or database of
  origin of the record. The \code{method} is specified within the square
  brackets in the column name.  \cr \code{PRIM_ID} \tab The primary ID of the
  accession record from which the set could be identified. \cr \code{IDKW}
  \tab The 'matching' keywords along with the IDs. \cr \code{DEL} \tab Column
  to indicate whether record has to be deleted or not. \cr \code{SPLIT} \tab
  Column to indicate whether record has to be branched and assembled into new
  set. \cr \code{COUNT} \tab The number of elements in a set. \cr } For the
  retrieved columns(fields) the prefix \code{K*} indicates the KWIC index of
  origin.
}
\description{
\code{ReviewProbDup} retrieves information associated with the probable
duplicate sets from the original PGR passport database(s) from which they
were identified in order to facilitate manual clerical review.
}
\details{
This function helps to retrieve PGR passport information associated with
fuzzy, phonetic or semantic probable duplicate sets in an object of class
\code{ProbDup} from the original databases(s) from which they were
identified. The original information of accessions comprising a set, which
have not been subjected to data standardization can be compared under manual
clerical review for the validation of the set.

By default only the fields(columns) which were used initially for creation of
the KWIC indexes using the \code{\link[PGRdup]{KWIC}} function are retrieved.
Additional fields(columns) if necessary can be specified using the
\code{extra.db1} and \code{extra.db2} arguments.

The output data frame can be subjected to clerical review either after
exporting into an external spreadsheet using
\code{\link[utils:write.table]{write.csv}} function or by using the
\code{\link[utils]{edit}} function.

The column \code{DEL} can be used to indicate whether a record has to be
deleted from a set or not. \code{Y} indicates "Yes", and the default \code{N}
indicates "No".

The column \code{SPLIT} similarly can be used to indicate whether a record in
a set has to be branched into a new set. A set of identical integers in this
column other than the default \code{0} can be used to indicate that they are
to be removed and assembled into a new set.
}
\note{
When any primary ID/key records in the fuzzy, phonetic or semantic
  duplicate sets are found to be missing from the original databases
  \code{db1} and \code{db2}, then they are ignored and only the matching
  records are considered for retrieving the information with a warning.

  This may be due to data standardization of the primary ID/key field using
  the function \code{\link[PGRdup]{DataClean}} before creation of the KWIC
  index and subsequent identification of probable duplicate sets. In such a
  case, it is recommended to use an identical data standardization operation
  on the databases \code{db1} and \code{db2} before running this function.

  With \code{R} <= v3.0.2, due to copying of named objects by \code{list()},
  \code{Invalid .internal.selfref detected and fixed...} warning can appear,
  which may be safely ignored.
}
\examples{

\dontshow{
threads_dt <- data.table::getDTthreads()
threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
data.table::setDTthreads(2)

data.table::setDTthreads(2)
Sys.setenv(`OMP_THREAD_LIMIT` = 2)
}

\dontrun{

# Load PGR passport database
GN <- GN1000

# Specify as a vector the database fields to be used
GNfields <- c("NationalID", "CollNo", "DonorID", "OtherID1", "OtherID2")

# Clean the data
GN[GNfields] <- lapply(GN[GNfields], function(x) DataClean(x))
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
c("Mota", "Company"))
y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
        "Bunch", "Peanut")
GN[GNfields] <- lapply(GN[GNfields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN[GNfields] <- lapply(GN[GNfields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))

# Generate KWIC index
GNKWIC <- KWIC(GN, GNfields)

# Specify the exceptions as a vector
exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE",
         "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT",
         "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE",
         "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R",
         "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE",
         "U", "VALENCIA", "VIRGINIA", "WHITE")

# Specify the synsets as a list
syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))

# Fetch probable duplicate sets
GNdup <- ProbDup(kwic1 = GNKWIC, method = "a", excep = exep, fuzzy = TRUE,
                 phonetic = TRUE, encoding = "primary",
                 semantic = TRUE, syn = syn)

# Get disjoint probable duplicate sets of each kind
disGNdup <- DisProbDup(GNdup, combine = NULL)

# Get the data frame for reviewing the duplicate sets identified
RevGNdup <- ReviewProbDup(pdup = disGNdup, db1 = GN1000,
                          extra.db1 = c("SourceCountry", "TransferYear"),
                          max.count = 30, insert.blanks = TRUE)
# Examine and review the duplicate sets using edit function
RevGNdup <- edit(RevGNdup)

# OR examine and review the duplicate sets after exporting them as a csv file
write.csv(file="Duplicate sets for review.csv", x=RevGNdup)

}

\dontshow{
data.table::setDTthreads(threads_dt)
Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
}

}
\seealso{
\code{\link[PGRdup]{DataClean}}, \code{\link[PGRdup]{KWIC}},
  \code{\link[PGRdup]{ProbDup}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ViewProbDup.R
\name{ViewProbDup}
\alias{ViewProbDup}
\title{Visualize the probable duplicate sets retrieved in a \code{ProbDup} object}
\usage{
ViewProbDup(
  pdup,
  db1,
  db2 = NULL,
  factor.db1,
  factor.db2 = NULL,
  max.count = 30,
  select,
  order = "type",
  main = NULL
)
}
\arguments{
\item{pdup}{An object of class \code{ProbDup}.}

\item{db1}{A data frame of the PGR passport database.}

\item{db2}{A data frame of the PGR passport database. Required when 
\code{pdup} was created using more than one KWIC Index.}

\item{factor.db1}{The \code{db1} column to be considered for grouping the 
accessions. Should be of class character or factor.}

\item{factor.db2}{The \code{db2} column to be considered for grouping the 
accessions. Should be of class character or factor. retrieved.}

\item{max.count}{The maximum count of probable duplicate sets whose 
information is to be plotted (see \strong{Note}).}

\item{select}{A character vector of factor names in \code{factor.db1} and/or 
\code{factor.db2} to be considered for grouping accessions (see 
\strong{Note}).}

\item{order}{The order of the type of sets retrieved in the plot. The default
is \code{"type"} (see \strong{Details}).}

\item{main}{The title of the plot.}
}
\value{
A list containing the following objects: \tabular{ll}{ 
  \code{Summary1} \tab The summary \code{data.frame} of number of accessions 
  per factor level. \cr \code{Summary2} \tab The summary \code{data.frame} of
  number of accessions and sets per each type of sets classified according to
  factor levels. \cr \code{SummaryGrob} \tab A grid graphical object (Grob) 
  of the summary visualization plot. Can be plotted using the \code{grid.arrange} function \cr }
}
\description{
\code{ViewProbDup} plots summary visualizations of accessions within the
probable duplicate sets retrieved in a \code{ProbDup} object according to a 
grouping factor field(column) in the original database(s).
}
\note{
When any primary ID/key records in the fuzzy, phonetic or semantic 
  duplicate sets are found to be missing from the original databases 
  \code{db1} and \code{db2}, then they are ignored and only the matching 
  records are considered for visualization.
  
  This may be due to data standardization of the primary ID/key field using 
  the function \code{\link[PGRdup]{DataClean}} before creation of the KWIC 
  index and subsequent identification of probable duplicate sets. In such a 
  case, it is recommended to use an identical data standardization operation 
  on the databases \code{db1} and \code{db2} before running this function. 
  For summary and visualization of the set information in the object of class
  \code{ProbDup} by \code{ViewProbDup}, the disjoint of the retrieved sets 
  are made use of, as they are more meaningful than the raw sets retrieved. 
  So it is recommended that the disjoint of sets obtained using the 
  \code{DisProbDup} be used as the input \code{pdup}.
  
  All the accession records in sets with count > \code{max.count} will be 
  considered as being unique.
  
  The factor levels in the \code{factor.db1} and/or \code{factor.db2} columns
  corresponding to those mentioned in \code{select} argument alone will be 
  considered for visualization. All other factor levels will be grouped 
  together to a single level named "Others".
  
  The argument \code{order} can be used to specify the order in which the 
  type of sets retrieved are to be plotted in the visualization. The default 
  \code{"type"} will order according to the kind of sets, \code{"sets"} will 
  order according to the number of sets in each kind and \code{"acc"} will 
  order according to the number of accessions in each kind.
  
  The individual plots are made using \code{\link[ggplot2]{ggplot}} and then 
  grouped together using \code{\link[gridExtra]{gridExtra-package}}.
}
\examples{

\dontshow{
threads_dt <- data.table::getDTthreads()
threads_OMP <- Sys.getenv("OMP_THREAD_LIMIT")
data.table::setDTthreads(2)

data.table::setDTthreads(2)
Sys.setenv(`OMP_THREAD_LIMIT` = 2)
}

\dontrun{

# Method "b and c"
#=================

# Load PGR passport databases
GN1 <- GN1000[!grepl("^ICG", GN1000$DonorID), ]
GN1$DonorID <- NULL
GN2 <- GN1000[grepl("^ICG", GN1000$DonorID), ]
GN2 <- GN2[!grepl("S", GN2$DonorID), ]
GN2$NationalID <- NULL

GN1$SourceCountry <- toupper(GN1$SourceCountry)
GN2$SourceCountry <- toupper(GN2$SourceCountry)

GN1$SourceCountry <- gsub("UNITED STATES OF AMERICA", "USA", GN1$SourceCountry)
GN2$SourceCountry <- gsub("UNITED STATES OF AMERICA", "USA", GN2$SourceCountry)

# Specify as a vector the database fields to be used
GN1fields <- c("NationalID", "CollNo", "OtherID1", "OtherID2")
GN2fields <- c("DonorID", "CollNo", "OtherID1", "OtherID2")

# Clean the data
GN1[GN1fields] <- lapply(GN1[GN1fields], function(x) DataClean(x))
GN2[GN2fields] <- lapply(GN2[GN2fields], function(x) DataClean(x))
y1 <- list(c("Gujarat", "Dwarf"), c("Castle", "Cary"), c("Small", "Japan"),
           c("Big", "Japan"), c("Mani", "Blanco"), c("Uganda", "Erect"),
           c("Mota", "Company"))
y2 <- c("Dark", "Light", "Small", "Improved", "Punjab", "SAM")
y3 <- c("Local", "Bold", "Cary", "Mutant", "Runner", "Giant", "No.",
        "Bunch", "Peanut")
GN1[GN1fields] <- lapply(GN1[GN1fields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN1[GN1fields] <- lapply(GN1[GN1fields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN1[GN1fields] <- lapply(GN1[GN1fields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))
GN2[GN2fields] <- lapply(GN2[GN2fields], function(x) MergeKW(x, y1, delim = c("space", "dash")))
GN2[GN2fields] <- lapply(GN2[GN2fields], function(x) MergePrefix(x, y2, delim = c("space", "dash")))
GN2[GN2fields] <- lapply(GN2[GN2fields], function(x) MergeSuffix(x, y3, delim = c("space", "dash")))

# Remove duplicated DonorID records in GN2
GN2 <- GN2[!duplicated(GN2$DonorID), ]

# Generate KWIC index
GN1KWIC <- KWIC(GN1, GN1fields)
GN2KWIC <- KWIC(GN2, GN2fields)

# Specify the exceptions as a vector
exep <- c("A", "B", "BIG", "BOLD", "BUNCH", "C", "COMPANY", "CULTURE",
          "DARK", "E", "EARLY", "EC", "ERECT", "EXOTIC", "FLESH", "GROUNDNUT",
          "GUTHUKAI", "IMPROVED", "K", "KUTHUKADAL", "KUTHUKAI", "LARGE",
          "LIGHT", "LOCAL", "OF", "OVERO", "P", "PEANUT", "PURPLE", "R",
          "RED", "RUNNER", "S1", "SAM", "SMALL", "SPANISH", "TAN", "TYPE",
          "U", "VALENCIA", "VIRGINIA", "WHITE")

# Specify the synsets as a list
syn <- list(c("CHANDRA", "AH114"), c("TG1", "VIKRAM"))

GNdupc <- ProbDup(kwic1 = GN1KWIC, kwic2 = GN2KWIC, method = "c",
                  excep = exep, fuzzy = TRUE, phonetic = TRUE,
                  encoding = "primary", semantic = TRUE, syn = syn)

GNdupcView <- ViewProbDup(GNdupc, GN1, GN2, "SourceCountry", "SourceCountry",
                         max.count = 30, select = c("INDIA", "USA"), order = "type",
                         main = "Groundnut Probable Duplicates")

library(gridExtra)                                                    
grid.arrange(GNdupcView$SummaryGrob)                          

}   

\dontshow{
data.table::setDTthreads(threads_dt)
Sys.setenv(`OMP_THREAD_LIMIT` = threads_OMP)
}
    
}
\seealso{
\code{\link[PGRdup]{ProbDup}}, \code{\link[PGRdup]{DisProbDup}}, 
  \code{\link[PGRdup]{DataClean}}, \code{\link[ggplot2]{ggplot}}, 
  \code{\link[gridExtra]{gridExtra-package}}
}
